Package: SeqVarTools
Version: 1.51.2
Type: Package
Title: Tools for variant data
Description: An interface to the fast-access storage format for VCF
        data provided in SeqArray, with tools for common operations and
        analysis.
Author: Stephanie M. Gogarten, Xiuwen Zheng, Adrienne Stilp
Maintainer: Stephanie M. Gogarten <sdmorris@uw.edu>
Depends: SeqArray
Imports: grDevices, graphics, stats, methods, Biobase, BiocGenerics,
        gdsfmt, GenomicRanges, IRanges, S4Vectors, GWASExactHW,
        logistf, Matrix, data.table,
Suggests: BiocStyle, RUnit, stringr, knitr, rmarkdown
License: GPL-3
URL: https://github.com/smgogarten/SeqVarTools
LazyData: yes
VignetteBuilder: knitr
biocViews: SNP, GeneticVariability, Sequencing, Genetics
Collate: AllClasses.R AllGenerics.R AllUtilities.R
        Methods-SeqVarGDSClass.R Methods-SeqVarData.R
        Methods-Iterator.R chromWithPAR.R duplicateDiscordance.R hwe.R
        inbreedCoeff.R mendelErr.R pca.R setVariantID.R
        alternateAlleleDetection.R refFrac.R regression.R
git_url: https://git.bioconductor.org/packages/SeqVarTools
git_branch: devel
git_last_commit: a22ff2d
git_last_commit_date: 2026-07-09
Repository: Bioconductor 3.24
Date/Publication: 2026-07-09
NeedsCompilation: no
Packaged: 2026-07-10 00:35:57 UTC; biocbuild

importFrom(grDevices, adjustcolor)
importFrom(graphics, abline, plot, text)
importFrom(stats, as.formula, binom.test, binomial, coef,
           complete.cases, glm, lm, median, model.matrix,
           na.omit, pchisq, qchisq, setNames, vcov)

import(methods)
import(gdsfmt)
import(SeqArray)
importFrom(BiocGenerics, end)
importFrom(S4Vectors, elementMetadata, mcols, "mcols<-",
           queryHits, subjectHits)
importFrom(IRanges, IRanges, findOverlaps, mergeByOverlaps,
           subsetByOverlaps)
importFrom(GenomicRanges, GRanges, GRangesList, seqnames)
import(Biobase)
import(logistf)
import(data.table)
importFrom(GWASExactHW, HWExact)
importFrom(Matrix, Matrix)

exportClasses(SeqVarData,
              SeqVarIterator,
              SeqVarBlockIterator,
              SeqVarRangeIterator,
              SeqVarWindowIterator,
              SeqVarListIterator)

exportMethods(show,
              sampleData, "sampleData<-",
              variantData, "variantData<-",
              validateSex,
              granges,
              variantFilter,
              lastFilter, "lastFilter<-",
              variantBlock,
              variantRanges,
              iterateFilter,
              resetIterator,
              currentRanges,
              currentVariants,
              applyMethod,
              refChar,
              altChar,
              nAlleles,
              isVariant,
              isSNV,
              getGenotype,
              getGenotypeAlleles,
              refDosage,
              altDosage,
              alleleDosage,
              expandedAltDosage,
              expandedVariantIndex,
              variantInfo,
              getVariableLengthData,
              imputedDosage,
              titv,
              alleleFrequency,
              alleleCount,
              minorAlleleCount,
              missingGenotypeRate,
              heterozygosity,
              hethom,
              homozygosity,
              meanBySample,
              countSingletons,
              pca,
              hwe,
              inbreedCoeff,
              mendelErr,
              duplicateDiscordance,
              alternateAlleleDetection,
              refFrac,
              refFracOverHets,
              refFracPlot,
              regression,
              chromWithPAR)

export(SeqVarData,
       SeqVarIterator,
       SeqVarBlockIterator,
       SeqVarRangeIterator,
       SeqVarWindowIterator,
       SeqVarListIterator,
       setVariantID)

# SeqVarTools
An interface to the fast-access storage format for VCF
data provided in [SeqArray](https://github.com/zhengxwen/SeqArray), with tools for common operations and
analysis.

## Bioconductor

[http://www.bioconductor.org/packages/release/bioc/html/SeqVarTools.html](http://www.bioconductor.org/packages/release/bioc/html/SeqVarTools.html)

* Tutorials
  * [Introduction to SeqVarTools](https://bioconductor.org/packages/release/bioc/vignettes/SeqVarTools/inst/doc/SeqVarTools.pdf)
  * [Iterators in SeqVarTools](https://bioconductor.org/packages/devel/bioc/vignettes/SeqVarTools/inst/doc/Iterators.pdf)
* [Reference manual](http://www.bioconductor.org/packages/release/bioc/manuals/SeqVarTools/man/SeqVarTools.pdf)
* [News](http://bioconductor.org/packages/release/bioc/news/SeqVarTools/NEWS)

## Installation

```R
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("SeqVarTools")
```

setMethod("pca",
          "SeqVarGDSClass",
          function(gdsobj, eigen.cnt=32) {
            nsamp <- .nSamp(gdsobj)
            if (eigen.cnt > nsamp) eigen.cnt <- nsamp

            ## genetic correlation matrix
            genmat <- local({

              ## need a global variable (only available in the bracket of "local")
              s <- matrix(0.0, nrow=nsamp, ncol=nsamp)

              ## apply the function variant by variant
              seqApply(gdsobj, "genotype", function(x) {
                g <- (x==0L)                  # indicator of reference allele
                p <- mean(g, na.rm=TRUE)      # allele frequency
                g2 <- colSums(g) - 2*p        # genotypes adjusted by allele frequency
                m <- (g2 %o% g2) / (p*(1-p))  # genetic correlation matrix
                m[!is.finite(m)] <- 0         # correct missing values
                s <<- s + m                   # output to the global variable "s"
              }, margin="by.variant", as.is="none")

              ## output, scaled by the trace of matrix "s" over the number of samples
              s / (sum(diag(s)) / nsamp)
            })

            ## eigen-decomposition
            eig <- eigen(genmat)

            eig$values <- eig$values[1:eigen.cnt]
            eig$vectors <- eig$vectors[,1:eigen.cnt]
            
            rownames(eig$vectors) <- seqGetData(gdsobj, "sample.id")
            
            names(eig) <- c("eigenval", "eigenvect")
            eig
          })

SeqVarIterator <- function(seqData, variantFilter, verbose=TRUE) {
    class(seqData) <- "SeqVarIterator"
    seqData@variantFilter <- variantFilter

    ## set filter to first element
    seqSetFilter(seqData, variant.sel=variantFilter[[1]], verbose=verbose)
    
    ## pass-by-reference slot for lastFilter
    seqData@lastFilter <- new.env()
    lastFilter(seqData) <- 1
    
    seqData
}


SeqVarBlockIterator <- function(seqData, variantBlock=10000, verbose=TRUE) {

    if (variantBlock > .nVar(seqData)) {
        variantBlock <- .nVar(seqData)
    }
    
    ## original filter
    current.sel <- which(seqGetFilter(seqData)$variant.sel)

    ## divide selected variants into blocks
    variantFilter <- split(current.sel, ceiling(seq_along(current.sel)/variantBlock))
    
    seqData <- SeqVarIterator(seqData, variantFilter, verbose=verbose)
    class(seqData) <- "SeqVarBlockIterator"
    seqData@variantBlock <- as.integer(variantBlock)

    seqData
}


SeqVarRangeIterator <- function(seqData, variantRanges, verbose=TRUE) {
    ## original filter
    current.sel <- which(seqGetFilter(seqData)$variant.sel)

    ## filter for iterator ranges
    variants <- granges(seqData)
    new.sel <- .subjectByQuery(variantRanges, variants, hits.only=FALSE)$subjectHits
    variantFilter <- lapply(new.sel, function(x) current.sel[x])
    
    seqData <- SeqVarIterator(seqData, variantFilter, verbose=verbose)
    class(seqData) <- "SeqVarRangeIterator"
    seqData@variantRanges <- variantRanges

    seqData
}


SeqVarWindowIterator <- function(seqData, windowSize=10000, windowShift=5000, verbose=TRUE) {
    ## original filter
    current.sel <- which(seqGetFilter(seqData)$variant.sel)
    
    # identify windows
    variants <- granges(seqData)
    windows <- do.call("c", lapply(unique(seqnames(variants)), function(chr) {
        start <- seq(1, max(end(variants[seqnames(variants) == chr])), windowShift)
        GRanges(seqnames=chr, ranges=IRanges(start=start, width=windowSize))
    }))

    ## variants in each window
    sbq <- .subjectByQuery(windows, variants, hits.only=TRUE)
    new.sel <- sbq$subjectHits

    ## only keep unique windows
    keep <- !duplicated(new.sel)
    new.sel <- new.sel[keep]
    windows <- windows[sbq$queryHits[keep]]

    variantFilter <- lapply(new.sel, function(x) current.sel[x])
    
    seqData <- SeqVarIterator(seqData, variantFilter, verbose=verbose)
    class(seqData) <- "SeqVarWindowIterator"
    seqData@variantRanges <- windows
    seqData@windowSize <- as.integer(windowSize)
    seqData@windowShift <- as.integer(windowShift)
        
    seqData
}


SeqVarListIterator <- function(seqData, variantRanges, verbose=TRUE) {
    ## original filter
    current.sel <- which(seqGetFilter(seqData)$variant.sel)
    
    ## filter for iterator ranges
    variants <- granges(seqData)
    new.sel <- .subjectByQuery(variantRanges, variants, hits.only=FALSE)$subjectHits
    variantFilter <- lapply(new.sel, function(x) current.sel[x])
    
    seqData <- SeqVarIterator(seqData, variantFilter, verbose=verbose)
    class(seqData) <- "SeqVarListIterator"
    seqData@variantRanges <- variantRanges

    seqData
}


setMethod("variantFilter",
          "SeqVarIterator",
          function(x) {
              x@variantFilter
          })

setMethod("lastFilter",
          "SeqVarIterator",
          function(x) {
              x@lastFilter$i
          })

setReplaceMethod("lastFilter",
                 c("SeqVarIterator", "numeric"),
                 function(x, value) {
                     x@lastFilter$i <- as.integer(value)
                     x
                 })

setMethod("iterateFilter",
          "SeqVarIterator",
          function(x, verbose=TRUE) {
              ## set filter for next element
              if (lastFilter(x) < length(variantFilter(x))) {
                  i <- lastFilter(x) + 1
                  seqSetFilter(x, variant.sel=variantFilter(x)[[i]], verbose=verbose)
                  lastFilter(x) <- i
                  return(TRUE)
              } else {
                  .emptyVarFilter(x, verbose=verbose)
                  return(FALSE)
              }
          })

setMethod("resetIterator",
          "SeqVarIterator",
          function(x, verbose=TRUE) {
              ## set filter for first element
              seqSetFilter(x, variant.sel=variantFilter(x)[[1]], verbose=verbose)
              lastFilter(x) <- 1
          })


setMethod("variantBlock",
          "SeqVarBlockIterator",
          function(x) {
              x@variantBlock
          })


setMethod("variantRanges",
          "SeqVarRangeIterator",
          function(x) {
              x@variantRanges
          })

setMethod("variantRanges",
          "SeqVarListIterator",
          function(x) {
              x@variantRanges
          })


setMethod("currentRanges",
          "SeqVarBlockIterator",
          function(x) {
              granges(x)
          })

setMethod("currentRanges",
          "SeqVarRangeIterator",
          function(x) {
              variantRanges(x)[lastFilter(x)]
          })

setMethod("currentRanges",
          "SeqVarListIterator",
          function(x) {
              variantRanges(x)[[lastFilter(x)]]
          })

setMethod("currentVariants",
          "SeqVarIterator",
          function(x) {
              variant <- granges(x)
              range <- currentRanges(x)
              mergeByOverlaps(variant, range)
          })

setMethod("currentVariants",
          "SeqVarBlockIterator",
          function(x) {
              df <- as(granges(x), "DataFrame")
              names(df)[1] <- "variant"
              df
          })


setMethod("show",
          "SeqVarIterator",
          function(object) {
              cat(class(object), "object; on iteration",
                  lastFilter(object), "of", length(variantFilter(object)), "\n")
              cat(" | GDS:\n")
              print(object)
              cat(" | sampleData:\n")
              show(sampleData(object))
              cat(" | variantData:\n")
              show(variantData(object))
          })

setMethod("show",
          "SeqVarRangeIterator",
          function(object) {
              callNextMethod(object)
              cat(" | variantRanges:\n")
              show(variantRanges(object))
          })

setMethod("show",
          "SeqVarListIterator",
          function(object) {
              callNextMethod(object)
              cat(" | variantRanges:\n")
              show(variantRanges(object))
          })

## -----------------------------------------------------------------------------
library(SeqVarTools)
gds <- seqOpen(seqExampleFileName("gds"))
seqData <- SeqVarData(gds)
iterator <- SeqVarBlockIterator(seqData, variantBlock=500)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)

## -----------------------------------------------------------------------------
seqSetFilter(seqData, variant.sel=1:100)
iterator <- SeqVarBlockIterator(seqData, variantBlock=500)
var.info <- variantInfo(iterator)
nrow(var.info)
iterateFilter(iterator)
seqResetFilter(seqData)

## -----------------------------------------------------------------------------
library(GenomicRanges)
gr <- GRanges(seqnames=rep(1,3), 
              ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
iterator <- SeqVarRangeIterator(seqData, variantRanges=gr)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)

## -----------------------------------------------------------------------------
seqSetFilterChrom(seqData, include="22")
iterator <- SeqVarWindowIterator(seqData, windowSize=10000, 
                                 windowShift=5000)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)

## -----------------------------------------------------------------------------
gr <- GRangesList(
  GRanges(seqnames=rep(22,2), 
          ranges=IRanges(start=c(16e6, 17e6), width=1e6)),
  GRanges(seqnames=rep(22,2), 
          ranges=IRanges(start=c(18e6, 20e6), width=1e6)))
iterator <- SeqVarListIterator(seqData, variantRanges=gr)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)

## -----------------------------------------------------------------------------
variantInfo(iterator)
resetIterator(iterator)
variantInfo(iterator)

## -----------------------------------------------------------------------------
seqClose(gds)

setMethod("inbreedCoeff",
          "SeqVarGDSClass",
          function(gdsobj, margin=c("by.variant", "by.sample"), use.names=FALSE, parallel=FALSE) {           
            margin <- match.arg(margin)
            if (margin == "by.variant") {
              counts <- .countGenotypes(gdsobj, parallel=parallel)
              f <- .f(counts)
              if (use.names)
                names(f) <- seqGetData(gdsobj, "variant.id")
              f
            } else {
              nsamp <- .nSamp(gdsobj)

              ## need global variables
              n <- integer(nsamp)
              s <- double(nsamp)

              ## apply the function variant by variant
              seqApply(gdsobj, "genotype", function(x) {
                p <- mean(x==0L, na.rm=TRUE)     # allele frequency
                g <- colSums(x==0L)              # genotypes: # of reference allele
                d <- (g*g - g*(1 + 2*p) + 2*p*p) / (2*p*(1-p))
                n <<- n + is.finite(d)           # output to the global variable "n"
                d[!is.finite(d)] <- 0
                s <<- s + d                      # output to the global variable "s"
              }, margin="by.variant", as.is="none")

              ## output
              ic <- s / n
              if (use.names)
                names(ic) <- seqGetData(gdsobj, "sample.id")
              ic
            }
          })

test_parseRefAllele <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T")
  checkIdentical(c("A","AA","A","A"), SeqVarTools:::.parseRefAllele(x))
}

test_parseAltAllele_n0 <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T")
  checkIdentical(c("G","G","GG","G,T"), SeqVarTools:::.parseAltAllele(x, n=0))
}

test_parseAltAllele_n1 <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T", "A,G,T,C", "A,G,TT,C")
  checkIdentical(c("G","G","GG","G","G","G"), SeqVarTools:::.parseAltAllele(x, n=1))
}

test_parseAltAllele_n2 <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T", "A,G,T,C", "A,G,TT,C")
  checkIdentical(c(NA,NA,NA,"T","T","TT"), SeqVarTools:::.parseAltAllele(x, n=2))
}

test_parseAltAllele_n3 <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T", "A,G,T,C", "A,G,TT,C")
  checkIdentical(c(NA,NA,NA,NA,"C","C"), SeqVarTools:::.parseAltAllele(x, n=3))
}

## test_parseNumAlleles <- function() {
##   x <- c("A", "A,A", "A,A,A", "A,A,A,A")
##   checkIdentical(1:4, SeqVarTools:::.parseNumAlleles(x))
## }

test_maxAlleleLength <- function() {
  x <- c("A,G", "AA,G", "A,GG", "A,G,T")
  checkIdentical(as.integer(c(1,2,2,1)), SeqVarTools:::.maxAlleleLength(x))
}

## applyMethod
setMethod("applyMethod",
          c(gdsobj="SeqVarGDSClass", FUN="function", variant="character"),
          function(gdsobj, FUN, variant, sample=NULL, ...) {
            .applyMethod(gdsobj, FUN, variant, sample, ...)
          })

setMethod("applyMethod",
          c(gdsobj="SeqVarGDSClass", FUN="function", variant="numeric"),
          function(gdsobj, FUN, variant, sample=NULL, ...) {
            .applyMethod(gdsobj, FUN, variant, sample, ...)
          })

setMethod("applyMethod",
          c(gdsobj="SeqVarGDSClass", FUN="function", variant="GRanges"),
          function(gdsobj, FUN, variant, sample=NULL, ...) {
            .applyMethod(gdsobj, FUN, .rangesToID(gdsobj, variant), sample, ...)
          })

setMethod("applyMethod",
          c(gdsobj="SeqVarGDSClass", FUN="function", variant="missing"),
          function(gdsobj, FUN, variant, sample=NULL, ...) {
            .applyMethod(gdsobj, FUN, variant.id=NULL, sample, ...)
          })


## allele methods
setMethod("refChar",
          "SeqVarGDSClass",
          function(gdsobj) {
            .parseRefAllele(seqGetData(gdsobj, "allele"))
          })

setMethod("altChar",
          "SeqVarGDSClass",
          function(gdsobj, n=0) {
            .parseAltAllele(seqGetData(gdsobj, "allele"), n=n)
          })

setMethod("nAlleles",
          "SeqVarGDSClass",
          function(gdsobj) {
            #.parseNumAlleles(seqGetData(gdsobj, "allele"))
            seqNumAllele(gdsobj)
          })


## descriptive methods
setMethod("isVariant",
          "SeqVarGDSClass",
          function(gdsobj, use.names=FALSE, parallel=FALSE) {
            var <- seqApply(gdsobj, "genotype",
                            function(x) {colSums(x, na.rm=TRUE) > 0},
                            margin="by.variant", as.is="list",
                            parallel=parallel)
            var <- matrix(unlist(var, use.names=FALSE), ncol=length(var))
            if (use.names) {
                dimnames(var) <- list(sample=NULL, variant=NULL)
                .applyNames(gdsobj, var)
            }else var
          })
 
setMethod("isSNV",
          "SeqVarGDSClass",
          function(gdsobj, biallelic=TRUE) {
            a <- seqGetData(gdsobj, "allele")
            if (biallelic) {
              nchar(a) == 3
            } else {
              .maxAlleleLength(a) == 1
            }
          })


## data retrieval and formatting
setMethod("getGenotype",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, parallel=FALSE) {
              if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
              if (.ploidy(gdsobj) == 1) {
                  gc <- seqGetData(gdsobj, "genotype")[1,,]
              } else {
                  gc <- seqApply(gdsobj, c(geno="genotype", phase="phase"),
                                 function(x) {sep=ifelse(x$phase, "|", "/")
                                     paste0(x$geno[1,], sep, x$geno[2,])},
                                 margin="by.variant", as.is="list",
                                 parallel=parallel)
                  gc <- matrix(unlist(gc, use.names=FALSE), ncol=length(gc))
                  ## gc <- seqBlockApply(gdsobj, c(geno="genotype", phase="phase"),
                  ##                     function(x) {
                  ##                         sep <- ifelse(x$phase, "|", "/")
                  ##                         matrix(paste0(x$geno[1,,], sep, x$geno[2,,]), nrow=ncol(x$geno))
                  ##                     },
                  ##                margin="by.variant", as.is="list", ...)
                  ## gc <- do.call(cbind, gc)
                  gc[gc == "NA/NA"] <- NA
              }
              if (use.names) {
                  dimnames(gc) <- list(sample=NULL, variant=NULL)
                  .applyNames(gdsobj, gc)
              } else gc
          })

setMethod("getGenotypeAlleles",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, sort=FALSE, parallel=FALSE) {
              if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
              if (.ploidy(gdsobj) == 1) {
                  gc <- seqApply(gdsobj,
                                 c(geno="genotype", allele="allele"),
                                 function(x) {
                                     alleles <- unlist(strsplit(x$allele, ",", fixed=TRUE),
                                                       use.names=FALSE)
                                     names(alleles) <- 0:(length(alleles) - 1)
                                     alleles[as.character(x$geno[1,])]
                                 }, margin="by.variant", as.is="list",
                                 parallel=parallel)
                  gc <- matrix(unlist(gc, use.names=FALSE), ncol=length(gc))
              } else {
                  gc <- seqApply(gdsobj,
                                 c(geno="genotype", phase="phase", allele="allele"),
                                 function(x, sort) {
                                     alleles <- unlist(strsplit(x$allele, ",", fixed=TRUE),
                                                       use.names=FALSE)
                                     names(alleles) <- 0:(length(alleles) - 1)
                                     a <- alleles[as.character(x$geno[1,])]
                                     b <- alleles[as.character(x$geno[2,])]
                                     if (sort) {
                                         paste(pmin(a,b), pmax(a,b), sep="/")
                                     } else  {
                                         sep=ifelse(x$phase, "|", "/")
                                         paste0(a, sep, b)
                                     }
                                 }, margin="by.variant", as.is="list", sort=sort,
                                 parallel=parallel)
                  gc <- matrix(unlist(gc, use.names=FALSE), ncol=length(gc))
                  ## gc <- seqBlockApply(gdsobj,
                  ##                c(geno="genotype", phase="phase", allele="allele"),
                  ##                function(x, sort) {
                  ##                    alleles <- strsplit(x$allele, ",", fixed=TRUE)
                  ##                    allele.map <- function(g, allele) {
                  ##                        names(allele) <- 0:(length(allele) - 1)
                  ##                        unname(allele[as.character(g)])
                  ##                    }
                  ##                    a <- mapply(allele.map, as.data.frame(x$geno[1,,]), alleles, USE.NAMES=FALSE)
                  ##                    b <- mapply(allele.map, as.data.frame(x$geno[2,,]), alleles, USE.NAMES=FALSE)
                  
                  ##                    if (sort) {
                  ##                        g <- paste(pmin(a,b), pmax(a,b), sep="/")
                  ##                    } else  {
                  ##                        sep=ifelse(x$phase, "|", "/")
                  ##                        g <- paste0(a, sep, b)
                  ##                    }
                  ##                    matrix(g, nrow=ncol(x$geno))
                  ##                }, margin="by.variant", as.is="list", sort=sort, ...)
                  ## gc <- do.call(cbind, gc)
                  gc[gc == "NA/NA"] <- NA
              }
              if (use.names) {
                  dimnames(gc) <- list(sample=NULL, variant=NULL)
                  .applyNames(gdsobj, gc)
              } else gc
          })

setMethod("refDosage",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, ...) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
            ## d <- seqBlockApply(gdsobj, "genotype",
            ##                    function(x) {colSums(x == 0)},
            ##                    margin="by.variant", as.is="list", ...)
            ## d <- do.call(cbind, d)
            d <- seqGetData(gdsobj, "$dosage", ...)
            if (use.names) {
                #dimnames(d) <- list(sample=NULL, variant=NULL)
                d <- .applyNames(gdsobj, d)
            } else {
                dimnames(d) <- NULL
            }
            d
          })

setMethod("altDosage",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, sparse=FALSE, parallel=FALSE, ...) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
            d <- seqBlockApply(gdsobj, "genotype",
                               function(x) {
                                   m <- colSums(x != 0)
                                   if (sparse) m <- Matrix(m, sparse=TRUE)
                                   m
                               }, margin="by.variant", as.is="list",
                               parallel=parallel, ...)
            #d <- do.call(cbind, d)
            d <- Reduce(cbind, d[-1], d[[1]])
            if (use.names) {
                dimnames(d) <- list(sample=NULL, variant=NULL)
                .applyNames(gdsobj, d)
            } else d
          })

## uses too much memory
## setMethod("altDosage2",
##           "SeqVarGDSClass",
##           function(gdsobj, use.names=TRUE, sparse=FALSE, ...) {
##             if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
##             d <- seqGetData(gdsobj, "$dosage_alt", ...)
##             if (sparse) d <- Matrix(d, sparse=TRUE)
##             if (use.names) {
##                 d <- .applyNames(gdsobj, d)
##             } else {
##                 dimnames(d) <- if (sparse) list(NULL,NULL) else NULL
##             }
##             d
##           })

setMethod("alleleDosage",
          c("SeqVarGDSClass", "numeric"),
          function(gdsobj, n=0, use.names=TRUE, parallel=FALSE) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
            if (length(n) == 1) n <- rep(n, .nVar(gdsobj))
            stopifnot(length(n) == .nVar(gdsobj))
            stopifnot(all(n <= nAlleles(gdsobj)))
            d <- seqApply(gdsobj, "genotype",
                          function(index, x) {colSums(x == n[index])},
                          margin="by.variant", as.is="list",
                          var.index="relative", parallel=parallel)
            d <- matrix(unlist(d, use.names=FALSE), ncol=length(d))
            if (use.names) {
                dimnames(d) <- list(sample=NULL, variant=NULL)
                .applyNames(gdsobj, d)
            } else d
          })

## setMethod("alleleDosage",
##           c("SeqVarGDSClass", "list"),
##           function(gdsobj, n, use.names=TRUE) {
##             stopifnot(length(n) == .nVar(gdsobj))
##             d <- seqApply(gdsobj, "genotype",
##                           function(index, x) {
##                               tmp <- matrix(x %in% n[[index]], ncol=ncol(x),
##                                              nrow=nrow(x))
##                               tmp[is.na(x)] <- NA
##                               colSums(tmp)
##                           },
##                           margin="by.variant", as.is="list",
##                           var.index="relative")
##             d <- matrix(unlist(d, use.names=FALSE), ncol=length(d),
##                          dimnames=list(sample=NULL, variant=NULL))
##             if (use.names) .applyNames(gdsobj, d) else d
##           })

setMethod("alleleDosage",
          c("SeqVarGDSClass", "list"),
          function(gdsobj, n, use.names=TRUE, parallel=FALSE) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))
            stopifnot(length(n) == .nVar(gdsobj))
            samp.names <- if (use.names) seqGetData(gdsobj, "sample.id") else NULL
            d <- seqApply(gdsobj, "genotype",
                          function(index, x) {
                              tmp <- lapply(n[[index]], function(allele) colSums(x == allele))
                              matrix(unlist(tmp, use.names=FALSE), ncol=length(tmp),
                                     dimnames=list(sample=samp.names, allele=n[[index]]))
                          },
                          margin="by.variant", as.is="list",
                          var.index="relative", parallel=parallel)
            if (use.names) names(d) <- seqGetData(gdsobj, "variant.id")
            d
          })

setMethod("expandedAltDosage",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, sparse=FALSE, parallel=FALSE) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))

            n <- nAlleles(gdsobj) - 1
            # if we have only biallelic variants, faster to use altDosage
            if (all(n == 1)) return(altDosage(gdsobj, use.names=use.names, sparse=sparse))
            
            samp.names <- if (use.names) seqGetData(gdsobj, "sample.id") else NULL
            variant.id <- if (use.names) seqGetData(gdsobj, "variant.id") else NULL
            d <- seqApply(gdsobj, "genotype",
                          function(index, x) {
                              tmp <- lapply(1:n[index], function(allele) colSums(x == allele))
                              var.names <- rep(variant.id[index], n[index])
                              m <- matrix(unlist(tmp, use.names=FALSE), ncol=length(tmp),
                                          dimnames=list(sample=samp.names, variant=var.names))
                              if (sparse) m <- Matrix(m, sparse=TRUE)
                              m
                          },
                          margin="by.variant", as.is="list",
                          var.index="relative", parallel=parallel)
            #https://stackoverflow.com/questions/37581417/getting-node-stack-overflow-when-cbind-multiple-sparse-matrices
            #d <- do.call(cbind, d)
            d <- Reduce(cbind, d[-1], d[[1]])
            if (use.names) names(dimnames(d)) <- c("sample", "variant")
            d
          })

setMethod("expandedVariantIndex",
          "SeqVarGDSClass",
          function(gdsobj) {
              nv <- .nVar(gdsobj)
              if (nv == 0) return(integer())
              nAlt <- nAlleles(gdsobj) - 1
              # keep variants with no alternate alleles
              nAlt[nAlt == 0] <- 1
              rep(1:nv, nAlt)
          })


### tidyverse version
## setMethod("variantInfo",
##           "SeqVarGDSClass",
##           function(gdsobj, alleles=TRUE, expanded=FALSE) {
##               x <- data.frame(variant.id=seqGetData(gdsobj, "variant.id"),
##                               chr=seqGetData(gdsobj, "chromosome"),
##                               pos=seqGetData(gdsobj, "position"),
##                               stringsAsFactors=FALSE)
##               if (nrow(x) == 0) return(x)
##               if (alleles) {
##                   allele <- seqGetData(gdsobj, "allele")
##                   x$ref <- .parseRefAllele(allele)
##                   x$alt <- .parseAltAllele(allele)
##               }
##               if (expanded) {
##                   if (alleles) {
##                       x <- separate_rows(x, .data$alt, sep=",")
##                   } else {
##                       x <- x[expandedVariantIndex(gdsobj),]
##                   }
##                   x <- group_by(x, .data$variant.id)
##                   x <- mutate(x, allele.index = 1:n())
##                   x <- as.data.frame(x)
##               }
##               x
##           })

setMethod("variantInfo",
          "SeqVarGDSClass",
          function(gdsobj, alleles=TRUE, expanded=FALSE) {
              x <- data.frame(variant.id=seqGetData(gdsobj, "variant.id"),
                              chr=seqGetData(gdsobj, "chromosome"),
                              pos=seqGetData(gdsobj, "position"),
                              stringsAsFactors=FALSE)
              if (nrow(x) == 0) return(x)
              if (alleles) {
                  allele <- seqGetData(gdsobj, "allele")
                  x <- as.data.table(x)
                  x[, `:=`(ref = .parseRefAllele(allele),
                           alt = .parseAltAllele(allele)
                           )]
              }
              if (expanded) {
                  x <- as.data.table(x)
                  if (alleles) {
                      alt <- NULL
                      x <- x[, strsplit(alt, ",", fixed=TRUE), by = c("variant.id", "chr", "pos", "ref", "alt")
                        ][, alt := NULL
                          ][, setnames(.SD, "V1", "alt")]
                  } else {
                      x <- x[expandedVariantIndex(gdsobj),]
                  }
                  x[, "allele.index" := seq_len(.N), by="variant.id"]
              }
              as.data.frame(x)
          })

setMethod("getVariableLengthData",
          c("SeqVarGDSClass", "character"),
          function(gdsobj, var.name, use.names=TRUE, parallel=FALSE) {
            var.list <- seqApply(gdsobj, var.name, function(x) {x},
                                 margin="by.variant", as.is="list",
                                 parallel=parallel)
            .ncol <- function(x) ifelse(is.null(x), 0, ncol(x))
            n.alleles <- sapply(var.list, .ncol)
            nsamp <- .nSamp(gdsobj)
            if (length(unique(n.alleles)) > 1) {
                ## fill in missing values
                maxn <- max(n.alleles)
                var.list <- lapply(var.list, function(x) {
                    cbind(x, matrix(NA, nrow=nsamp, ncol=(maxn - .ncol(x))))
                })
            }
            var <- array(unlist(var.list, use.names=FALSE),
                         dim=c(nrow(var.list[[1]]), ncol(var.list[[1]]),
                             length(var.list)))
            var <- aperm(var, c(2,1,3))
            
            ## if first array dimension is 1, simplify to a matrix
            if (dim(var)[1] == 1) {
              var <- var[1,,]
            }
            if (length(dim(var)) == 3) {
              dimnames(var) <- list(n=NULL, sample=NULL, variant=NULL)
            } else {
              dimnames(var) <- list(sample=NULL, variant=NULL)
            }
            
            if (use.names) .applyNames(gdsobj, var) else var
          })


setMethod("imputedDosage",
          "SeqVarGDSClass",
          function(gdsobj, dosage.field="DS", use.names=TRUE) {
            if (.emptyDim(gdsobj)) return(.emptyGenoMatrix(gdsobj, use.names=use.names))

            d <- seqGetData(gdsobj, paste0("annotation/format/", dosage.field))
            if (!is.null(names(d))) {
                if (!all(d$length == 1)) stop("multiple dosage values per variant")
                d <- d$data
            }
            if (use.names) {
                dimnames(d) <- list(sample=NULL, variant=NULL)
                .applyNames(gdsobj, d)
            } else d
          })
          


## metrics    
setMethod("titv",
          "SeqVarGDSClass",
          function(gdsobj, by.sample=FALSE, use.names=FALSE) {
            ref <- refChar(gdsobj)
            alt <- altChar(gdsobj)
            ti <- .isTransition(ref, alt)
            tv <- .isTransversion(ref, alt)
            if (by.sample) {
              isVar <- function(x) {colSums(x, na.rm=TRUE) > 0}
              tisum <- integer(.nSamp(gdsobj))
              tvsum <- integer(.nSamp(gdsobj))
              seqApply(gdsobj, "genotype",
                       function(index, x) {
                         tisum <<- tisum + (ti[index] & isVar(x));
                         tvsum <<- tvsum + (tv[index] & isVar(x))
                       },
                       margin="by.variant", as.is="none", var.index="relative")
              titv <- tisum / tvsum
              if (use.names)
                names(titv) <- seqGetData(gdsobj, "sample.id")
              titv
            } else {
              sum(ti) / sum(tv)
            }
          })

## n=0: REF allele freq
## n>0: freq of nth ALT allele
setMethod("alleleFrequency",
          "SeqVarGDSClass",
          function(gdsobj, n=0, use.names=FALSE, parallel=FALSE) {
            ## af <- seqApply(gdsobj, "genotype",
            ##                function(x) {mean(x == n, na.rm=TRUE)},
            ##                margin="by.variant", as.is="double")
            af <- seqAlleleFreq(gdsobj, ref.allele=n, parallel=parallel)
            ## frequency of alt=2, etc. should be 0 if there is no such allele
            af[is.na(af)] <- 0
            if (use.names)
              names(af) <- seqGetData(gdsobj, "variant.id")
            af
          })

## n=0: REF allele count
## n>0: count of nth ALT allele
setMethod("alleleCount",
          "SeqVarGDSClass",
          function(gdsobj, n=0, use.names=FALSE, parallel=FALSE) {
            ac <- seqAlleleCount(gdsobj, ref.allele=n, parallel=parallel)
            ## count of alt=2, etc. should be 0 if there is no such allele
            ac[is.na(ac)] <- 0L
            if (use.names)
              names(ac) <- seqGetData(gdsobj, "variant.id")
            ac
          })

setMethod("minorAlleleCount",
          "SeqVarGDSClass",
          function(gdsobj, use.names=FALSE, parallel=FALSE) {
              ref.cnt <- seqAlleleCount(gdsobj, parallel=parallel)
              n.obs <- .nSampObserved(gdsobj)
              ac <- pmin(ref.cnt, 2L*n.obs - ref.cnt)
              if (use.names)
                  names(ac) <- seqGetData(gdsobj, "variant.id")
              ac
          })

setMethod("missingGenotypeRate",
          "SeqVarGDSClass",
          function(gdsobj, margin=c("by.variant", "by.sample"), use.names=FALSE, parallel=FALSE) {
            margin <- match.arg(margin)
            if (margin == "by.variant") {
              miss <- seqMissing(gdsobj, per.variant=TRUE, parallel=parallel)
              if (use.names)
                names(miss) <- seqGetData(gdsobj, "variant.id")
            } else {
              miss <- seqMissing(gdsobj, per.variant=FALSE, parallel=parallel)
              if (use.names)
                names(miss) <- seqGetData(gdsobj, "sample.id")
            } 
            miss

##             if (margin == "by.variant") {
##               miss <- seqApply(gdsobj, "genotype",
##                                function(x) {sum(is.na(x[1,]))},
##                                margin=margin, as.is="integer")
##               if (use.names)
##                 names(miss) <- seqGetData(gdsobj, "variant.id")
##               miss / .nSamp(gdsobj)
##             } else {
##               miss <- integer(.nSamp(gdsobj))
##               ## use "<<-" operator to find "miss" in the parent environment
##               seqApply(gdsobj, "genotype",
##                        function(x) {miss <<- miss + is.na(x[1,])},
##                        margin="by.variant", as.is="none")
##               if (use.names)
##                 names(miss) <- seqGetData(gdsobj, "sample.id")
##               miss / .nVar(gdsobj)
##             } 

## REPLACE when by.sample works in seqApply
##              miss <- seqApply(gdsobj, "genotype",
##                               function(x) {sum(is.na(x[1,]))},
##                               margin=margin, as.is="integer")
##             if (margin == "by.variant") {
##               miss / .nSamp(gdsobj)
##             } else {
##               miss / .nVar(gdsobj)
##             } 
          })

setMethod("heterozygosity",
          "SeqVarGDSClass",
          function(gdsobj, margin=c("by.variant", "by.sample"), use.names=FALSE, parallel=FALSE) {
            margin <- match.arg(margin)
            if (margin == "by.variant") {
            het <- seqApply(gdsobj, "genotype",
                            function(x) {
                              sum(x[1,] != x[2,], na.rm=TRUE) /
                              sum(!is.na(x[1,]) & !is.na(x[2,]))
                            },
                            margin=margin, as.is="double", parallel=parallel)
            if (use.names)
              names(het) <- seqGetData(gdsobj, "variant.id")
            het
          } else {
            het <- integer(.nSamp(gdsobj))
            nonmiss <- integer(.nSamp(gdsobj))
            ## use "<<-" operator to find "het" in the parent environment
            seqApply(gdsobj, "genotype",
                     function(x) {
                       nm <- !is.na(x[1,]) & !is.na(x[2,])
                       het <<- het + (x[1,] != x[2,] & nm)
                       nonmiss <<- nonmiss + nm
                     },
                     margin="by.variant", as.is="none")
            if (use.names)
              names(het) <- seqGetData(gdsobj, "sample.id")
            het / nonmiss
          }
            
## REPLACE when by.sample works in seqApply
##             het <- seqApply(gdsobj, "genotype",
##                             function(x) {
##                               sum(x[1,] != x[2,], na.rm=TRUE) /
##                                 sum(!is.na(x[1,]) & !is.na(x[2,]))
##                             },
##                             margin=margin, as.is="integer")
          })

setMethod("homozygosity",
          "SeqVarGDSClass",
          function(gdsobj, allele=c("any", "ref", "alt"),
                   margin=c("by.variant", "by.sample"), use.names=FALSE, parallel=FALSE) {
            hom.func <- switch(match.arg(allele),
                               any=function(a,b) {a == b},
                               ref=function(a,b) {a == b & a == 0},
                               alt=function(a,b) {a == b & a > 0})
            margin <- match.arg(margin)
            if (margin == "by.variant") {
              hom <- seqApply(gdsobj, "genotype",
                              function(x) {
                                sum(hom.func(x[1,], x[2,]), na.rm=TRUE) /
                                  sum(!is.na(x[1,]) & !is.na(x[2,]))
                              },
                              margin=margin, as.is="double", parallel=parallel)
              if (use.names)
                names(hom) <- seqGetData(gdsobj, "variant.id")
              hom
            } else {
              hom <- integer(.nSamp(gdsobj))
              nonmiss <- integer(.nSamp(gdsobj))
              ## use "<<-" operator to find "hom" in the parent environment
              seqApply(gdsobj, "genotype",
                       function(x) {
                         nm <- !is.na(x[1,]) & !is.na(x[2,])
                         hom <<- hom + (hom.func(x[1,], x[2,]) & nm)
                         nonmiss <<- nonmiss + nm
                       },
                       margin="by.variant", as.is="none")
              if (use.names)
                names(hom) <- seqGetData(gdsobj, "sample.id")
              hom / nonmiss
            }
          })


setMethod("hethom",
          "SeqVarGDSClass",
          function(gdsobj, use.names=FALSE) {
            hom.func <- function(a,b) {a == b & a > 0}
            het <- integer(.nSamp(gdsobj))
            hom <- integer(.nSamp(gdsobj))
            seqApply(gdsobj, "genotype",
                     function(x) {
                       nm <- !is.na(x[1,]) & !is.na(x[2,])
                       het <<- het + (x[1,] != x[2,] & nm)
                       hom <<- hom + (hom.func(x[1,], x[2,]) & nm)
                     },
                     margin="by.variant", as.is="none")
             if (use.names)
               names(het) <- seqGetData(gdsobj, "sample.id")
             het / hom
           })
            

setMethod("meanBySample",
          "SeqVarGDSClass",
          function(gdsobj, var.name, use.names=FALSE) {
            ns <- .nSamp(gdsobj)
            tot <- double(ns)
            nm <- double(ns)
            seqApply(gdsobj, var.name,
                     function(x) {
                       val <- !is.na(x)
                       tot[val] <<- tot[val] + x[val]
                       nm[val] <<- nm[val] + 1
                     }, margin="by.variant", as.is="none")
            if (use.names)
              names(tot) <- seqGetData(gdsobj, "sample.id")
            tot / nm
          })


setMethod("countSingletons",
          "SeqVarGDSClass",
          function(gdsobj, use.names=FALSE) {
            st <- integer(.nSamp(gdsobj))
            seqApply(gdsobj, "genotype",
                     function(x) {
                       nonref <- which(colSums(x, na.rm=TRUE) > 0)
                       if (length(nonref) == 1) {
                           st[nonref] <<- st[nonref] + 1
                       }
                     },
                     margin="by.variant", as.is="none")
             if (use.names)
               names(st) <- seqGetData(gdsobj, "sample.id")
             st
           })

.triosFromPedigree <- function(pedigree) {
  fams <- unique(pedigree$family)
  trios <- list()
  for (f in fams) {
    fam <- pedigree[pedigree$family == f,]
    for (i in 1:nrow(fam)) {
      if (fam$mother[i] %in% fam$individ & fam$father[i] %in% fam$individ) {
        trio <- c(child=fam$sample.id[i],
                  mother=fam$sample.id[fam$individ == fam$mother[i]],
                  father=fam$sample.id[fam$individ == fam$father[i]])
        trios[[length(trios)+1]] <- trio
      } else if (fam$mother[i] %in% fam$individ) {
        trio <- c(child=fam$sample.id[i],
                  mother=fam$sample.id[fam$individ == fam$mother[i]])
        trios[[length(trios)+1]] <- trio
      } else if (fam$father[i] %in% fam$individ) {
        trio <- c(child=fam$sample.id[i],
                  father=fam$sample.id[fam$individ == fam$father[i]])
        trios[[length(trios)+1]] <- trio
      }
    }
  }
  trios
}

.trioType <- function(trio) {
  if ("mother" %in% trio & "father" %in% trio) {
    "both.parents"
  } else if ("mother" %in% trio) {
    "mother.only"
  } else if ("father" %in% trio) {
    "father.only"
  }
}

.alleleMatch <- function(geno, allele, parent) {
  match1 <- geno[allele,"child",] == geno[1,parent,]
  match2 <- geno[allele,"child",] == geno[2,parent,]
  match1 | match2
}

## classes of Mendelian errors
## autosomes, pseudoautosomal region, and X chrom female child
## father  mother  child
## AA      AA      AB
## BB      BB      AB
## BB      **      AA
## **      BB      AA
## AA      **      BB
## **      AA      BB
.autosomeErr <- function(geno) {
  type <- .trioType(dimnames(geno)$sample)
  hom <- geno[1,"child",] == geno[2,"child",]
  if (type == "both.parents") {
    mother1match <- .alleleMatch(geno, 1, "mother")
    mother2match <- .alleleMatch(geno, 2, "mother")
    father1match <- .alleleMatch(geno, 1, "father")
    father2match <- .alleleMatch(geno, 2, "father")
    err.het <- !(mother1match | father1match) | !(mother2match | father2match)
    err.mother <- hom & !(mother1match | mother2match)
    err.father <- hom & !(father1match | father2match)
    err <- err.het | err.mother | err.father
    ## deal with missing values
    if (sum(is.na(err) > 0)) {
      hom <- geno[1,"child",] == geno[2,"child",]
      mother.missing <- is.na(mother1match) | is.na(mother2match)
      err[mother.missing] <- err.father[mother.missing]
      father.missing <- is.na(father1match) | is.na(father2match)
      err[father.missing] <- err.mother[father.missing]
    }
  } else if (type == "mother.only") {
    ## can only test for homozygote errors
    mother1match <- .alleleMatch(geno, 1, "mother")
    mother2match <- .alleleMatch(geno, 2, "mother")
    err <- hom & !(mother1match | mother2match)
  } else if (type == "father.only") {
    ## can only test for homozygote errors
    father1match <- .alleleMatch(geno, 1, "father")
    father2match <- .alleleMatch(geno, 2, "father")
    err <- hom & !(father1match | father2match)
  }
  err[is.na(err)] <- FALSE
  err
}

## X chromosome male child
## father  mother  child
## **      AA      BB
## **      BB      AA
.xMaleErr <- function(geno) {
  type <- .trioType(dimnames(geno)$sample)
  if (type == "both.parents" | type == "mother.only") {
    mother1match <- .alleleMatch(geno, 1, "mother")
    mother2match <- .alleleMatch(geno, 2, "mother")
    err <- !(mother1match | mother2match)
  } else {
    err <- rep(FALSE, dim(geno)[3])
  }
  err[is.na(err)] <- FALSE
  err
}

## Y chromosome male child
## father  mother  child
## AA      **      BB
## BB      **      AA
.yMaleErr <- function(geno) {
  type <- .trioType(dimnames(geno)$sample)
  if (type == "both.parents" | type == "father.only") {
    father1match <- .alleleMatch(geno, 1, "father")
    father2match <- .alleleMatch(geno, 2, "father")
    err <- !(father1match | father2match)
  } else {
    err <- rep(FALSE, dim(geno)[3])
  }
  err[is.na(err)] <- FALSE
  err
}


setMethod("mendelErr",
          "SeqVarGDSClass",
          function(gdsobj, pedigree, use.names=FALSE,
                   autosomes=1:22, xchrom="X", ychrom="Y",
                   verbose=TRUE) {
            ## pedigree shoud have format (family, individ, father, mother, sex, sample.id)
            ## generate trios (and duos) from pedigree
            trios <- .triosFromPedigree(pedigree)
            
            ## deal with duplicate samples (to-do later)
            
            ## chromosome types
            chrom <- seqGetData(gdsobj, "chromosome")
            auto <- chrom %in% autosomes
            x <- chrom %in% xchrom
            y <- chrom %in% ychrom
            ## pseudoautosomal?
            ## mitochondrial?           

            ## get original sample filter
            filt.orig <- seqGetFilter(gdsobj)$sample.sel
            
            ## vector to store errors by variant
            var.err <- integer(.nVar(gdsobj))

            ## for each trio
            trio.err <- integer()
            for (t in trios) {
              ## get child + parent(s) genotype
              seqSetFilter(gdsobj, sample.id=t, verbose=verbose)
              geno <- seqGetData(gdsobj, "genotype")
              sample.id <- seqGetData(gdsobj, "sample.id")
              dimnames(geno)$sample <- names(t)[match(sample.id, t)]
              
              ## add to count of errors 
              child.sex <- pedigree$sex[pedigree$sample.id == t["child"]]
              aut.type <- switch(child.sex, F=(auto | x), M=auto)
              aut.err <- .autosomeErr(geno[,,aut.type])
              var.err[aut.type] <- var.err[aut.type] + aut.err

              if (child.sex == "M") {
                x.err <- .xMaleErr(geno[,,x])
                var.err[x] <- var.err[x] + x.err
                y.err <- .yMaleErr(geno[,,y])
                var.err[y] <- var.err[y] + y.err
              } else {
                x.err <- NULL
                y.err <- NULL
              }

              ## keep track of errors by trio
              trio.err[as.character(t["child"])] <- sum(c(aut.err, x.err, y.err))
            }
              
            ## reset original sample filter
            seqSetFilter(gdsobj, sample.sel=filt.orig, verbose=FALSE)

            if (use.names)
              names(var.err) <- seqGetData(gdsobj, "variant.id")
            list(by.variant=var.err, by.trio=trio.err)
          })

test_triosFromPedigree <- function() {
  ped <- data.frame(family=c(rep("a",4), rep("b",5), rep("c",2)),
                    individ=c("a1","a2","am","af", "b1","bm","bf","bfm","bff", "c1","cm"),
                    father=c("af","af",0,0, "bf",0,"bff",0,0, 0,0),
                    mother=c("am","am",0,0, "bm",0,"bfm",0,0, "cm",0),
                    sample.id=c(11:14, 21:25, 31:32),
                    stringsAsFactors=FALSE)
  trios <- list(c(child=11, mother=13, father=14),
                c(child=12, mother=13, father=14),
                c(child=21, mother=22, father=23),
                c(child=23, mother=24, father=25),
                c(child=31, mother=32))
  checkEquals(trios, SeqVarTools:::.triosFromPedigree(ped))
}


test_alleleMatch <- function() {
  geno <- array(dim=c(2,2,5),
                dimnames=list(NULL,c("child","parent"),NULL))
  geno[1,1,] <- c(0,0,0,0,0)
  geno[2,1,] <- c(1,1,1,1,1)
  geno[1,2,] <- c(0,0,0,1,1)
  geno[2,2,] <- c(0,0,0,0,1)
  match1 <- c(TRUE,TRUE,TRUE,TRUE,FALSE)
  match2 <- c(FALSE,FALSE,FALSE,TRUE,TRUE)
  checkIdentical(match1, SeqVarTools:::.alleleMatch(geno,1,"parent"))
  checkIdentical(match2, SeqVarTools:::.alleleMatch(geno,2,"parent"))
}


test_autosomeErr <- function() {
  geno <- array(dim=c(2,3,9),
                dimnames=list(NULL, sample=c("child","mother","father"), NULL))
  ## father  mother  child
  ## AB      AB      AB     1 - no err
  ## AB      AB      AA     2 - no err
  ## AB      AB      BB     3 - no err
  ## AA      BB      AB     4 - no err
  ## AA      BB      AA     5 - err
  ## AA      BB      BB     6 - err
  ## BB      AA      AB     7 - no err
  ## BB      AA      AA     8 - err
  ## BB      AA      BB     9 - err
  geno[1,"father",] <- c(0,0,0,0,0,0,1,1,1)
  geno[2,"father",] <- c(1,1,1,0,0,0,1,1,1)
  geno[1,"mother",] <- c(0,0,0,1,1,1,0,0,0)
  geno[2,"mother",] <- c(1,1,1,1,1,1,0,0,0)
  geno[1,"child",]  <- c(0,0,1,0,0,1,0,0,1)
  geno[2,"child",]  <- c(1,0,1,1,0,1,1,0,1) 
  err <- c(rep(FALSE,3),rep(c(FALSE,TRUE,TRUE),2))
  checkIdentical(err, SeqVarTools:::.autosomeErr(geno))
  err.mother <- c(rep(FALSE,3),FALSE,TRUE,FALSE,FALSE,FALSE,TRUE)
  checkIdentical(err.mother, SeqVarTools:::.autosomeErr(geno[,c("child","mother"),]))
  err.father <- c(rep(FALSE,3),FALSE,FALSE,TRUE,FALSE,TRUE,FALSE)
  checkIdentical(err.father, SeqVarTools:::.autosomeErr(geno[,c("child","father"),]))
                                          
  geno <- array(dim=c(2,3,6),
                dimnames=list(NULL, sample=c("child","mother","father"), NULL))
  ## father  mother  child
  ## AA      AB      AB     1 - no err
  ## AA      AB      AA     2 - no err
  ## AA      AB      BB     3 - err
  ## AB      AA      AB     4 - no err
  ## AB      AA      AA     5 - no err
  ## AB      AA      BB     6 - err
  geno[1,"father",] <- c(0,0,0,0,0,0)
  geno[2,"father",] <- c(0,0,0,1,1,1)
  geno[1,"mother",] <- c(0,0,0,0,0,0)
  geno[2,"mother",] <- c(1,1,1,0,0,0)
  geno[1,"child",]  <- c(0,0,1,0,0,1)
  geno[2,"child",]  <- c(1,0,1,1,0,1) 
  err <- rep(c(FALSE,FALSE,TRUE),2)
  checkIdentical(err, SeqVarTools:::.autosomeErr(geno))
  err.mother <- c(rep(FALSE,3),FALSE,FALSE,TRUE)
  checkIdentical(err.mother, SeqVarTools:::.autosomeErr(geno[,c("child","mother"),]))
  err.father <- c(FALSE,FALSE,TRUE,rep(FALSE,3))
  checkIdentical(err.father, SeqVarTools:::.autosomeErr(geno[,c("child","father"),]))
  
  geno <- array(dim=c(2,3,6),
                dimnames=list(NULL, sample=c("child","mother","father"), NULL))
  ## father  mother  child
  ## AA      **      AB     1 - no err
  ## AA      **      AA     2 - no err
  ## AA      **      BB     3 - err
  ## **      AA      AB     4 - no err
  ## **      AA      AA     5 - no err
  ## **      AA      BB     6 - err
  geno[1,"father",] <- c(0,0,0,NA,NA,NA)
  geno[2,"father",] <- c(0,0,0,NA,NA,NA)
  geno[1,"mother",] <- c(NA,NA,NA,0,0,0)
  geno[2,"mother",] <- c(NA,NA,NA,0,0,0)
  geno[1,"child",]  <- c(0,0,1,0,0,1)
  geno[2,"child",]  <- c(1,0,1,1,0,1) 
  err <- rep(c(FALSE,FALSE,TRUE),2)
  checkIdentical(err, SeqVarTools:::.autosomeErr(geno))
}

test_xMaleErr <- function() {
  geno <- array(dim=c(2,2,6),
                dimnames=list(NULL, sample=c("child","mother"), NULL))
  ## father  mother  child
  ## **      AA      AB    1 - no err
  ## **      AA      AA    2 - no err
  ## **      AA      BB    3 - err
  ## **      BB      AB    4 - no err
  ## **      BB      AA    5 - err
  ## **      BB      BB    6 - no err
  geno[1,"mother",] <- c(0,0,0,1,1,1)
  geno[2,"mother",] <- c(0,0,0,1,1,1)
  geno[1,"child",]  <- c(0,0,1,0,0,1)
  geno[2,"child",]  <- c(1,0,1,1,0,1)
  err <- c(FALSE,FALSE,TRUE,FALSE,TRUE,FALSE)
  checkIdentical(err, SeqVarTools:::.xMaleErr(geno))
}

test_yMaleErr <- function() {
  geno <- array(dim=c(2,2,6),
                dimnames=list(NULL, sample=c("child","father"), NULL))
  ## father  mother  child
  ## AA      **      AB    1 - no err
  ## AA      **      AA    2 - no err
  ## AA      **      BB    3 - err
  ## BB      **      AB    4 - no err
  ## BB      **      AA    5 - err
  ## BB      **      BB    6 - no err
  geno[1,"father",] <- c(0,0,0,1,1,1)
  geno[2,"father",] <- c(0,0,0,1,1,1)
  geno[1,"child",]  <- c(0,0,1,0,0,1)
  geno[2,"child",]  <- c(1,0,1,1,0,1)
  err <- c(FALSE,FALSE,TRUE,FALSE,TRUE,FALSE)
  checkIdentical(err, SeqVarTools:::.yMaleErr(geno))
}

test_mendelErr <- function() {
  gds <- SeqVarTools:::.testData()
  data(pedigree)
  checkEquals(list(c(child="NA12878", mother="NA12892", father="NA12891")),
              SeqVarTools:::.triosFromPedigree(pedigree))
  err <- mendelErr(gds, pedigree)
  checkEquals(rep(0, length(seqGetData(gds, "variant.id"))),
              err$by.variant)
  checkEquals(c("NA12878"=0), err$by.trio)
  seqClose(gds)
}

test_mendelErr_apply <- function() {
  gds <- SeqVarTools:::.testData()
  data(pedigree)
  var.id <- 101:110
  seqSetFilter(gds, variant.id=var.id)
  err <- mendelErr(gds, pedigree)
  seqSetFilter(gds)
  checkIdentical(err,
                 applyMethod(gds, mendelErr, variant=var.id, pedigree=pedigree))
  seqClose(gds)
}

## define pseudoautosomal region

setMethod("chromWithPAR",
          "SeqVarGDSClass",
          function(gdsobj, genome.build=c("hg19", "hg38")) {
              build <- match.arg(genome.build)
              chr <- seqGetData(gdsobj, "chromosome")
              if (build == "hg19") {
                  PAR <- GRanges(seqnames=c("X", "X", "Y", "Y"),
                                 ranges=IRanges(start=c(60001, 154931044, 10001, 59034050),
                                                end=c(2699520, 155260560, 2649520, 59363566)))
              } else if (build == "hg38") {
                  PAR <- GRanges(seqnames=c("X", "X", "Y", "Y"),
                                 ranges=IRanges(start=c(10001, 155701383, 10001, 56887903),
                                                end=c(2781479, 156030895, 2781479, 57217415)))
              }
              chr[.rangesToSel(gdsobj, PAR)] <- "PAR"
              chr
          })

test_titv_bysamp <- function() {
  gds <- SeqVarTools:::.testData()
  ref <- refChar(gds)
  alt <- altChar(gds)
  ti <- SeqVarTools:::.isTransition(ref, alt)
  tv <- SeqVarTools:::.isTransversion(ref, alt)
  var <- isVariant(gds)
  timat <- matrix(ti, nrow=nrow(var), ncol=ncol(var), byrow=TRUE)
  tvmat <- matrix(tv, nrow=nrow(var), ncol=ncol(var), byrow=TRUE)
  checkIdentical(rowSums(timat & var) / rowSums(tvmat & var),
                 titv(gds, by.sample=TRUE))
  seqClose(gds)
}

test_titv_bysamp_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  ti.tv <- titv(gds, by.sample=TRUE)
  seqSetFilter(gds)
  checkIdentical(ti.tv,
                 applyMethod(gds, titv, variant=var.id, sample=samp.id,
                             by.sample=TRUE))
  seqClose(gds)
}

test_titv_byvar_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  ti.tv <- titv(gds)
  seqSetFilter(gds)
  checkIdentical(ti.tv,
                 applyMethod(gds, titv, variant=var.id, sample=samp.id))
  seqClose(gds)
}

.applyMethod <- function(gdsobj, FUN, variant.id=NULL, sample.id=NULL, ...) {
  seqSetFilter(gdsobj, sample.id=sample.id, variant.id=variant.id, action="push+set")
  result <- FUN(gdsobj, ...)
  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
  result
}

.rangesToID <- function(gdsobj, ranges) {
  gds.ranges <- granges(gdsobj, id=seqGetData(gdsobj, "variant.id"))
  mcols(subsetByOverlaps(gds.ranges, ranges))$id
}

.rangesToSel <- function(gdsobj, ranges) {
  gds.ranges <- granges(gdsobj)
  # don't warn if no sequence levels in common
  queryHits(suppressWarnings(findOverlaps(gds.ranges, ranges)))
}

.ploidy <- function(gdsobj) {
  seqSummary(gdsobj, "genotype", check="none", verbose=FALSE)$dim[1L]
}

.nSamp <- function(gdsobj) {
  #sum(seqGetFilter(gdsobj)$sample.sel)
  seqSummary(gdsobj, "genotype", check="none", verbose=FALSE)$seldim[2L]
}

.nVar <- function(gdsobj) {
  #sum(seqGetFilter(gdsobj)$variant.sel)
  seqSummary(gdsobj, "genotype", check="none", verbose=FALSE)$seldim[3L]
}

.nSampUnfiltered <- function(gdsobj) {
  seqSummary(gdsobj, "sample.id", check="none", verbose=FALSE)
}

.nVarUnfiltered <- function(gdsobj) {
  seqSummary(gdsobj, "variant.id", check="none", verbose=FALSE)
}

.nSampObserved <- function(gdsobj) {
  ns <- .nSamp(gdsobj)
  if (ns > 0) {
      return(as.integer(round(ns * (1-seqMissing(gdsobj)))))
  } else {
      return(rep(0L, .nVar(gdsobj)))
  }
}

.emptySampFilter <- function(x, verbose=FALSE) {
    seqSetFilter(x, sample.sel=raw(.nSampUnfiltered(x)), action="push+set", verbose=verbose)
}

.emptyVarFilter <- function(x, verbose=FALSE) {
    seqSetFilter(x, variant.sel=raw(.nVarUnfiltered(x)), action="push+set", verbose=verbose)
}

.emptyDim <- function(x) {
    .nSamp(x) == 0 | .nVar(x) == 0
}

.emptyGenoMatrix <- function(x, use.names=FALSE) {
    m <- matrix(nrow=.nSamp(x), ncol=.nVar(x),
                dimnames=list(sample=NULL, variant=NULL))
    if (use.names) .applyNames(x, m) else m
}

.applyNames <- function(gdsobj, var) {
  if ("sample" %in% names(dimnames(var)))
    dimnames(var)$sample <- seqGetData(gdsobj, "sample.id")
  if ("variant" %in% names(dimnames(var)))
    dimnames(var)$variant <- seqGetData(gdsobj, "variant.id")
  var
}

.parseRefAllele <- function(x) {
  endRef <- regexpr(",", x, fixed=TRUE) - 1L
  noAlt <- endRef < 0L
  if (any(noAlt))
      endRef[noAlt] <- nchar(x[noAlt])
  substr(x, 1L, endRef)
}

.parseAltAllele <- function(x, n=0) {
  if (n == 0) {
    ## if n=0, return string with all ALT alleles
    begAlt <- regexpr(",", x, fixed=TRUE) + 1L
    noAlt <- begAlt == 0L
    if (any(noAlt))
        begAlt[noAlt] <- nchar(x[noAlt]) + 1L
    substr(x, begAlt, nchar(x))
  } else {
    ## if n>0, return nth ALT allele
    unlist(lapply(strsplit(x, ",", fixed=TRUE), function(x) x[n+1]),
           use.names=FALSE)
  }
}

## .parseNumAlleles <- function(x) {
##   #unlist(lapply(strsplit(x, ",", fixed=TRUE), length), use.names=FALSE)
##   str_count(x, ",") + 1L
## }

.maxAlleleLength <- function(x) {
  a <- gregexpr("[ACGT]+", x)
  unlist(lapply(a, function(y) max(attr(y, "match.length"))), use.names=FALSE)
}

.parseVariableLength <- function(x) {
  if (all(x$length == 1)) {
    x$data
  } else {
    var <- array(dim=c(max(x$length), nrow(x$data), length(x$length)))

    ## assign each element of length to an index of first array dimension
    n.ind <- rep(NA, ncol(x$data))
    j <- 1
    for (i in 1:length(x$length)) {
      len <- x$length[i]
      if (len > 0) {
        n.ind[j:(j + len - 1)] <- 1:len
        j <- j + len
      }
    }

    ## for each index of first array dimension, get values
    for (n in 1:dim(var)[1]) {
      var.ind <- which(x$length >= n)
      var[n,,var.ind] <- x$data[,n.ind == n]
    }

    ## if first array dimension is 1, simplify to a matrix
    if (dim(var)[1] == 1) {
      var <- var[1,,]
    }
    var
  }
}

.isTransition <- function(ref, alt) {
  (ref %in% c("C","T") & alt %in% c("C","T")) |
  (ref %in% c("A","G") & alt %in% c("A","G"))
}

.isTransversion <- function(ref, alt) {
  (ref %in% c("C","T") & alt %in% c("A","G")) |
  (ref %in% c("A","G") & alt %in% c("C","T"))
}

## number of homozygote reference genotypes
.nHomRef <- function(x) {
    sum(x[1,] == 0 & x[2,] == 0, na.rm=TRUE)
}

## number of heterozyotes with one reference allele
.nHetRef <- function(x) {
    sum((x[1,] == 0 & x[2,] != 0) |
        (x[1,] != 0 & x[2,] == 0), na.rm=TRUE)
}

## number of genotypes with no reference alleles
.nHomAlt <- function(x) {
    sum(x[1,] != 0 & x[2,] != 0, na.rm=TRUE)
}

.permuteGenotypes <- function(x) {
    ## get subset of matrix with no missing values
    ind <- colSums(is.na(x)) == 0
    nm <- x[,ind,drop=FALSE]
    ## permute alleles
    nm <- matrix(sample(nm), nrow=nrow(nm), ncol=ncol(nm))
    ## replace non-missing genotypes
    x[,ind] <- nm
    x
}

.countGenotypes <- function(gdsobj, permute=FALSE, parallel=FALSE) {
    n <- seqApply(gdsobj, "genotype", function(x) {
        if (permute) x <- .permuteGenotypes(x)
        c(nAA=.nHomRef(x), nAa=.nHetRef(x), naa=.nHomAlt(x))
    }, margin="by.variant", as.is="list", parallel=parallel)
    as.data.frame(do.call(rbind, n))
}

.subjectByQuery <- function(query, subject, hits.only=FALSE) {
    ol <- findOverlaps(query, subject)
    if (hits.only) {
        hits <- unique(queryHits(ol))
    } else {
        hits <- seq_along(query)
    }
    subj.hits <- lapply(hits, function(h) {
        subjectHits(ol)[queryHits(ol) == h]
    })
    list(queryHits=hits, subjectHits=subj.hits)
}

# behaves like subsetByOverlaps, except removes query ranges that have
# duplicate sets of overlapping subject ranges
.subsetByUniqueOverlaps <- function(query, subject) {
    # look for any subject ranges that have identical query ranges
    # we want to eliminate these
    sbq <- .subjectByQuery(query, subject, hits.only=TRUE)
    query[sbq$queryHits[!duplicated(sbq$subjectHits)]]
}

.testData <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- seqExampleFileName("gds")
    seqOpen(gdsfile)
}

.testSeqVarData <- function() {
    SeqVarData(.testData())
}

.testDosageData <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "gl_chr1.gds", package="SeqVarTools")
    seqOpen(gdsfile)
}

.alleleCount <- function(gdsobj, n=0) {
    freq <- seqAlleleFreq(gdsobj, ref.allele=n)
    nsamp <- .nSamp(gdsobj)*(1-seqMissing(gdsobj))
    2*freq*nsamp
}

.testSampleData <- function(gds) {
    require(Biobase)
    sample.id <- seqGetData(gds, "sample.id")
    set.seed(66); sex <- sample(c("M","F"), replace=TRUE, length(sample.id))
    df <- data.frame(sample.id, sex, stringsAsFactors=FALSE)
    AnnotatedDataFrame(df)
}

.testVariantData <- function(gds) {
    require(Biobase)
    variant.id <- seqGetData(gds, "variant.id")
    set.seed(77); weight <- runif(length(variant.id))
    df <- data.frame(variant.id, weight, stringsAsFactors=FALSE)
    AnnotatedDataFrame(df)
}

test_SeqVarData <- function() {
    ## test creation from existing object
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gds.fn <- seqExampleFileName("gds")
    gds <- seqOpen(gds.fn)
    sample.id <- seqGetData(gds, "sample.id")
    adf <- .testSampleData(gds)
    svd <- SeqVarData(gds, sampleData=adf)
    checkIdentical(sample.id, seqGetData(svd, "sample.id"))
    seqClose(svd)

    ## test creation from filename
    svd <- SeqVarData(gds.fn, sampleData=adf)
    checkIdentical(sample.id, seqGetData(svd, "sample.id"))
    seqClose(svd)

    ## test errors
    checkException(SeqVarData("abc"))
    gds <- seqOpen(gds.fn)
    checkException(SeqVarData(gds, pData(adf)))
    adf$sample.id <- paste0("a", adf$sample.id)
    checkException(SeqVarData(gds, adf))
    adf$sample.id <- NULL
    checkException(SeqVarData(gds, adf))
    seqClose(gds)
}

test_variantData <- function() {
    gds <- SeqVarTools:::.testData()
    adf <- .testVariantData(gds)
    svd <- SeqVarData(gds, variantData=adf)
    checkIdentical(adf$variant.id, seqGetData(svd, "variant.id"))

    ## test errors
    adf$variant.id <- adf$variant.id + 1
    checkException(SeqVarData(gds, variantData=adf))
    
    seqClose(svd)
}

test_filters <- function() {
    gds <- SeqVarTools:::.testData()
    sdf <- .testSampleData(gds)
    vdf <- .testVariantData(gds)
    svd <- SeqVarData(gds, sampleData=sdf, variantData=vdf)
    seqSetFilter(svd, sample.id=sdf$sample.id[1:10], variant.id=vdf$variant.id[1:10], verbose=FALSE)
    checkEquals(seqGetData(svd, "sample.id"), sampleData(svd)$sample.id)
    checkEquals(seqGetData(svd, "variant.id"), variantData(svd)$variant.id)
    seqClose(svd)
}

test_replacement <- function() {
    gds <- SeqVarTools:::.testData()
    sdf <- .testSampleData(gds)
    vdf <- .testVariantData(gds)
    svd <- SeqVarData(gds, sampleData=sdf, variantData=vdf)
    
    sdf$tmp <- "a"
    sampleData(svd) <- sdf
    checkIdentical(sdf$tmp, sampleData(svd)$tmp)
    checkException(sampleData(svd) <- sdf[1:10,])

    vdf$tmp <- "a"
    variantData(svd) <- vdf
    checkIdentical(vdf$tmp, variantData(svd)$tmp)
    checkException(variantData(svd) <- vdf[1:10,])
    
    seqClose(svd)
}

test_missing <- function() {
    ## test creation with missing arguments
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gds.fn <- seqExampleFileName("gds")
    svd <- SeqVarData(gds.fn)
    checkEquals(length(seqGetData(svd, "sample.id")), nrow(sampleData(svd)), checkNames=FALSE)
    checkEquals(length(seqGetData(svd, "variant.id")), nrow(variantData(svd)), checkNames=FALSE)
    seqClose(svd)
}

test_granges <- function() {
    gds <- SeqVarTools:::.testData()
    svd <- SeqVarData(gds)
    checkEquals(granges(gds), granges(svd))
    
    vdf <- .testVariantData(gds)
    variantData(svd) <- vdf
    x <- pData(vdf)[,-1,drop=FALSE]
    row.names(x) <- as.character(row.names(x))
    checkEquals(x, as.data.frame(S4Vectors::mcols(granges(svd))), checkNames=FALSE)
    
    seqClose(svd)
}

test_filtered <- function() {
    gds <- SeqVarTools:::.testData()
    sdf <- .testSampleData(gds)
    vdf <- .testVariantData(gds)
    seqSetFilter(gds, sample.sel=1:10, variant.sel=1:10, verbose=FALSE)
    svd <- SeqVarData(gds)
    checkEquals(10, nrow(sampleData(svd)), checkNames=FALSE)
    checkEquals(10, nrow(variantData(svd)), checkNames=FALSE)
    svd <- SeqVarData(gds, sampleData=sdf, variantData=vdf)
    checkEquals(seqGetData(svd, "sample.id"), sampleData(svd)$sample.id)
    checkEquals(seqGetData(svd, "variant.id"), variantData(svd)$variant.id)
    seqClose(svd)
}

test_validateSex <- function() {
    svd <- SeqVarTools:::.testSeqVarData()
    checkTrue(is.null(validateSex(svd)))
    sdf <- .testSampleData(svd)
    sampleData(svd) <- sdf
    checkTrue(!is.null(validateSex(svd)))
    set.seed(88); sampleData(svd)$sex <- sample(c(1,2,NA), nrow(sdf), replace=TRUE)
    checkTrue(setequal(c("M","F",NA), validateSex(svd)))
    seqClose(svd)
}

test_dosage <- function() {
    gds <- SeqVarTools:::.testDosageData()
    dos <- imputedDosage(gds)
    checkTrue(is(dos, "matrix"))
    checkEquals(rownames(dos), seqGetData(gds, "sample.id"))
    checkEquals(colnames(dos), as.character(seqGetData(gds, "variant.id")))
    checkTrue(all(dos >= 0 & dos <= 2))
    seqClose(gds)
}

test_alleleFrequency_sum <- function() {
  gds <- SeqVarTools:::.testData()
  maxn <- max(nAlleles(gds))
  af <- matrix(nrow=SeqVarTools:::.nVar(gds), ncol=maxn)
  for (n in 1:maxn) af[,n] <- alleleFrequency(gds, n=(n-1))
  checkTrue(all(rowSums(af) == 1))
  seqClose(gds)
}

test_alleleFrequency_info <- function() {
  gds <- SeqVarTools:::.testData()
  ac <- seqGetData(gds, "annotation/info/AC")
  an <- seqGetData(gds, "annotation/info/AN")
  checkEquals(ac/an, alleleFrequency(gds, n=1))
  seqClose(gds)
}

test_alleleFrequency_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  af <- alleleFrequency(gds)
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(af,
                 applyMethod(gds, alleleFrequency, variant=var.id, sample=samp.id))
  seqClose(gds)
}

.testGdsXY <- function() {
    # make up file with sex chroms
    gds.fn <- tempfile()
    invisible(file.copy(seqExampleFileName("gds"), gds.fn))
    gds <- openfn.gds(gds.fn, readonly=FALSE)
    node <- index.gdsn(gds, "chromosome")
    compression.gdsn(node, "")
    chr <- read.gdsn(node)
    chr[chr == 1] <- "X"
    chr[chr == 2] <- "Y"
    write.gdsn(node, chr)
    closefn.gds(gds)
    seqOptimize(gds.fn, target="chromosome", verbose=FALSE)
    gds <- seqOpen(gds.fn)
    sample.id <- seqGetData(gds, "sample.id")
    set.seed(44); sex <- sample(c("M","F"), replace=TRUE, length(sample.id))
    df <- data.frame(sample.id, sex, stringsAsFactors=FALSE)
    SeqVarData(gds, sampleData=Biobase::AnnotatedDataFrame(df))
}

.cleanupGds <- function(gds) {
    fn <- seqSummary(gds, check="none", verbose=FALSE)$filename
    seqClose(gds)
    unlink(fn)
}


test_alleleFrequency_sex <- function() {
    svd <- .testGdsXY()
    sex <- sampleData(svd)$sex

    af <- alleleFrequency(svd)

    geno <- refDosage(svd, use.names=FALSE)
    chr <- chromWithPAR(svd)
    auto <- chr %in% 1:22
    checkEquals(0.5*colMeans(geno[,auto], na.rm=TRUE), af[auto])
    
    X <- chr == "X"
    female <- sex == "F"
    male <- sex == "M"
    F.count <- colSums(geno[female, X], na.rm=TRUE)
    F.nsamp <- colSums(!is.na(geno[female, X]))
    M.count <- 0.5*colSums(geno[male, X], na.rm=TRUE)
    M.nsamp <- colSums(!is.na(geno[male, X]))
    checkEquals((F.count + M.count)/(2*F.nsamp + M.nsamp), af[X])

    Y <- chr == "Y"
    checkEquals(0.5*colMeans(geno[male,Y], na.rm=TRUE), af[Y])

    # PAR
    checkTrue(all(chr[1:3] == "PAR"))
    checkEquals(0.5*colMeans(geno[,1:3], na.rm=TRUE), af[1:3])

    # names
    af <- alleleFrequency(svd, use.names=TRUE)
    checkEquals(as.character(seqGetData(svd, "variant.id")), names(af))
    
    .cleanupGds(svd)
}

test_alleleFrequency_nosex <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    # make sure there is no warning when we don't want to check sex
    options(warn=2)
    tmp <- alleleFrequency(gds, sex.adjust=FALSE)
    tmp <- alleleFrequency(gds, sex.adjust=TRUE) # no sex chrom in this data
    options(warn=1)
    seqClose(gds)
}


test_alleleCount_sex <- function() {
    svd <- .testGdsXY()
    sex <- sampleData(svd)$sex

    ac <- alleleCount(svd)

    geno <- refDosage(svd, use.names=FALSE)
    chr <- chromWithPAR(svd)
    auto <- chr %in% 1:22
    checkEquals(colSums(geno[,auto], na.rm=TRUE), ac[auto])
    
    X <- chr == "X"
    female <- sex == "F"
    male <- sex == "M"
    F.count <- colSums(geno[female, X], na.rm=TRUE)
    M.count <- 0.5*colSums(geno[male, X], na.rm=TRUE)
    checkEquals((F.count + M.count), ac[X])

    Y <- chr == "Y"
    checkEquals(0.5*colSums(geno[male,Y], na.rm=TRUE), ac[Y])

    # MAC
    mac <- minorAlleleCount(svd)
    ac.alt <- alleleCount(svd, n=1) + alleleCount(svd, n=2)
    minor <- ac < ac.alt
    checkEquals(ac[minor], mac[minor])
    checkEquals(ac.alt[!minor], mac[!minor])
    
    # PAR
    checkTrue(all(chr[1:3] == "PAR"))
    checkEquals(colSums(geno[,1:3], na.rm=TRUE), ac[1:3])

    # names
    ac <- alleleCount(svd, use.names=TRUE)
    checkEquals(as.character(seqGetData(svd, "variant.id")), names(ac))
    
    .cleanupGds(svd)
}


## this file is badly formatted for current SeqArray
## test_1KG_Y <- function() {
##     gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
##     gds.fn <- system.file("extdata", "1KG_chrY.gds", package="SeqVarTools")
    
##     gds <- seqOpen(gds.fn)
##     sample.id <- seqGetData(gds, "sample.id")
##     df <- data.frame(sample.id, sex="M", stringsAsFactors=FALSE)
##     svd <- SeqVarData(gds, sampleData=Biobase::AnnotatedDataFrame(df))

##     af <- alleleFrequency(svd)

##     geno <- refDosage(svd, use.names=FALSE)
##     chr <- chromWithPAR(gds)
##     checkTrue(all(chr == "Y"))
    
##     checkEquals(colMeans(geno, na.rm=TRUE), af)

##     # AC
##     ac <- alleleCount(svd)
##     checkEquals(colSums(geno, na.rm=TRUE), ac)

##     # MAC
##     mac <- minorAlleleCount(svd)
##     ac.alt <- alleleCount(svd, n=1) + alleleCount(svd, n=2) + alleleCount(svd, n=3) + alleleCount(svd, n=4) 
##     minor <- ac < ac.alt
##     checkEquals(ac[minor], mac[minor])
##     checkEquals(ac.alt[!minor], mac[!minor])
    
##     seqClose(gds)
## }


test_alleleCount_allF <- function() {
    svd <- .testGdsXY()
    sex <- sampleData(svd)$sex
    seqSetFilter(svd, sample.sel=(sex == "F"), verbose=FALSE)

    ac <- alleleCount(svd)

    geno <- refDosage(svd, use.names=FALSE)
    chr <- chromWithPAR(svd)
    
    X <- chr == "X"
    F.count <- colSums(geno[, X], na.rm=TRUE)
    checkEquals((F.count), ac[X])

    Y <- chr == "Y"
    checkTrue(all(ac[Y] == 0))

    # MAC
    mac <- minorAlleleCount(svd)
    ac.alt <- alleleCount(svd, n=1) + alleleCount(svd, n=2)
    minor <- ac < ac.alt
    checkEquals(ac[minor], mac[minor])
    checkEquals(ac.alt[!minor], mac[!minor])
    
    .cleanupGds(svd)
}

test_meanBySample <- function() {
  gds <- SeqVarTools:::.testData()
  dp <- getVariableLengthData(gds, "annotation/format/DP", use.names=TRUE)
  checkEquals(rowMeans(dp, na.rm=TRUE),
              meanBySample(gds, "annotation/format/DP", use.names=TRUE))
  seqClose(gds)
}

test_meanBySample_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  mn <- meanBySample(gds, "annotation/format/DP")
  seqSetFilter(gds)
  checkIdentical(mn,
                 applyMethod(gds, meanBySample, variant=var.id, sample=samp.id,
                             var.name="annotation/format/DP"))
  seqClose(gds)
}

library(GenomicRanges)

test_iterator_block <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    it <- SeqVarBlockIterator(gds, variantBlock=100, verbose=FALSE)
    checkEquals(1:100, seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(101:200, seqGetData(it, "variant.id"))
    checkEquals(granges(it), currentRanges(it))
    seqClose(it)
}

test_iterator_block_large <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    it <- SeqVarBlockIterator(gds, variantBlock=10000, verbose=FALSE)
    checkEquals(1:SeqVarTools:::.nVar(it), seqGetData(it, "variant.id"))
    checkTrue(!iterateFilter(it, verbose=FALSE))
    checkEquals(GRanges(), currentRanges(it), checkNames=FALSE)
    seqClose(it)
}

test_iterator_block_prev <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    set.seed(11); var.sel <- sort(sample(1:SeqVarTools:::.nVar(gds), 250))
    seqSetFilter(gds, variant.sel=var.sel, verbose=FALSE)
    it <- SeqVarBlockIterator(gds, variantBlock=100, verbose=FALSE)
    checkEquals(var.sel[1:100], seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(var.sel[101:200], seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(var.sel[201:250], seqGetData(it, "variant.id"))
    checkTrue(!iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    seqClose(it)
}

test_iterator_block_samples <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    set.seed(22); samp.sel <- sort(sample(1:SeqVarTools:::.nSamp(gds), 50))
    seqSetFilter(gds, sample.sel=samp.sel, verbose=FALSE)
    it <- SeqVarBlockIterator(gds, variantBlock=100, verbose=FALSE)
    checkEquals(samp.sel, which(seqGetFilter(it)$sample.sel))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(samp.sel, which(seqGetFilter(it)$sample.sel))
    seqClose(it)
}

test_iterator_range <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    gr <- GRanges(seqnames=rep(1,3), ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
    it <- SeqVarRangeIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(1:3, seqGetData(it, "variant.id"))
    checkEquals(gr[1], currentRanges(it))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(4:7, seqGetData(it, "variant.id"))
    checkTrue(!iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    seqClose(it)
}

test_iterator_range_prev <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    seqSetFilter(gds, variant.sel=c(1,3,5,7), verbose=FALSE)
    gr <- GRanges(seqnames=rep(1,3), ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
    it <- SeqVarRangeIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(c(1,3), seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(c(5,7), seqGetData(it, "variant.id"))
    checkTrue(!iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    seqClose(it)
}

test_iterator_range_samples <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    set.seed(33); samp.sel <- sort(sample(1:SeqVarTools:::.nSamp(gds), 50))
    seqSetFilter(gds, sample.sel=samp.sel, verbose=FALSE)
    gr <- GRanges(seqnames=rep(1,3), ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
    it <- SeqVarRangeIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(samp.sel, which(seqGetFilter(it)$sample.sel))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(samp.sel, which(seqGetFilter(it)$sample.sel))
    seqClose(it)
}

test_unique_overlaps <- function() {
    subject <- GRanges(seqnames="1", ranges=IRanges(seq(10,50,10), (seq(10,50,10))))
    query <- GRanges(seqnames="1", ranges=IRanges(c(1,5,9,25,25,29), c(5,15,18,45,55,41)))
    chk <- list(queryHits=2:6, subjectHits=list(1, 1, 3:4, 3:5, 3:4))
    checkEquals(chk, SeqVarTools:::.subjectByQuery(query, subject, hits.only=TRUE))
    chk <- list(queryHits=1:6, subjectHits=list(integer(0), 1, 1, 3:4, 3:5, 3:4))
    checkEquals(chk, SeqVarTools:::.subjectByQuery(query, subject, hits.only=FALSE))
    keep <- c(2,4,5)
    checkEquals(query[keep], SeqVarTools:::.subsetByUniqueOverlaps(query, subject))
}

test_iterator_window <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    seqSetFilterChrom(gds, include="22", verbose=FALSE)
    it <- SeqVarWindowIterator(gds, verbose=FALSE)
    checkTrue(all(width(it@variantRanges) == 10000))
    var <- list(seqGetData(it, "variant.id"))
    i <- 2
    while (iterateFilter(it, verbose=FALSE)) {
        var[[i]] <- seqGetData(it, "variant.id")
        i <- i + 1
    }
    checkEquals(length(var), length(it@variantRanges))
    seqSetFilterChrom(gds, include="22", verbose=FALSE)
    checkEquals(sort(unique(unlist(var))), seqGetData(it, "variant.id"))
    seqClose(it)
}

test_iterator_list <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    gr <- GRangesList(
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(1e6, 3e6), width=1e6)),
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(3e6, 34e6), width=1e6)))
    it <- SeqVarListIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(1:7, seqGetData(it, "variant.id"))
    checkEquals(gr[[1]], currentRanges(it))
    checkTrue(iterateFilter(it, verbose=FALSE))
    checkEquals(4:11, seqGetData(it, "variant.id"))
    checkTrue(!iterateFilter(it, verbose=FALSE))
    checkEquals(integer(), seqGetData(it, "variant.id"))
    seqClose(it)
}

test_iterator_list_duplicates <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    gr <- granges(gds)
    grl <- GRangesList(gr[c(1,1,2)])
    it <- SeqVarListIterator(gds, variantRanges=grl, verbose=FALSE)
    checkEquals(1:2, seqGetData(it, "variant.id"))
    seqClose(it)
}


.testVarData <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    vdf <- Biobase::AnnotatedDataFrame(data.frame(
        variant.id=seqGetData(gds, "variant.id"),
        weight=1, stringsAsFactors=FALSE))
    variantData(gds) <- vdf
    gds
}

test_currentRanges_block <- function() {
    gds <- .testVarData()
    it <- SeqVarBlockIterator(gds, variantBlock=10, verbose=FALSE)
    checkEquals(currentRanges(it)$weight, rep(1,10))
    seqClose(it)
}

test_currentRanges_list <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    gr <- GRangesList(
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(1e6, 3e6), width=1e6), weight=rep(1,2)),
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(3e6, 34e6), width=1e6), weight=rep(2,2)))
    it <- SeqVarListIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(currentRanges(it)$weight, c(1,1))
    iterateFilter(it, verbose=FALSE)
    checkEquals(currentRanges(it)$weight, c(2,2))
    seqClose(it)
}

test_currentRanges_range <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    gr <- GRanges(seqnames=rep(1,3), ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6), weight=rep(1,3))
    it <- SeqVarRangeIterator(gds, variantRanges=gr, verbose=FALSE)
    checkEquals(currentRanges(it)$weight, 1)
    iterateFilter(it, verbose=FALSE)
    checkEquals(currentRanges(it)$weight, 1)
    seqClose(it)
}

test_currentVariants_window <- function() {
    gds <- .testVarData()
    it <- SeqVarWindowIterator(gds, verbose=FALSE)
    checkTrue(all(currentVariants(it)$weight == 1))
    seqClose(it)
}

test_currentVariants_range <- function() {
    gds <- .testVarData()
    gr <- GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(1e6, 3e6), width=1e6), weight2=rep(2,2))
    it <- SeqVarRangeIterator(gds, variantRanges=gr, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(rownames(cv), as.character(1:3))
    checkEquals(cv$weight, rep(1,3))
    checkEquals(cv$weight2, rep(2,3))
    iterateFilter(it, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(rownames(cv), as.character(4:7))
    checkEquals(cv$weight, rep(1,4))
    checkEquals(cv$weight2, rep(2,4))
    seqClose(it)
}

test_currentVariants_list <- function() {
    gds <- .testVarData()
    gr <- GRangesList(
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(1e6, 3e6), width=1e6), weight2=rep(2,2)),
        GRanges(seqnames=rep(1,2), ranges=IRanges(start=c(3e6, 34e6), width=1e6), weight2=rep(3,2)))
    it <- SeqVarListIterator(gds, variantRanges=gr, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(rownames(cv), as.character(1:7))
    checkEquals(cv$weight, rep(1,7))
    checkEquals(cv$weight2, rep(2,7))
    iterateFilter(it, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(rownames(cv), as.character(4:11))
    checkEquals(cv$weight, rep(1,8))
    checkEquals(cv$weight2, rep(3,8))
    seqClose(it)
}

test_currentVariants_block <- function() {
    gds <- .testVarData()
    it <- SeqVarBlockIterator(gds, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(names(cv), c("variant", "weight"))
    checkEquals(rownames(cv), as.character(seqGetData(it, "variant.id")))
    checkTrue(all(cv$weight == 1))
    seqClose(it)
}

test_currentVariants_no_variantData <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    it <- SeqVarWindowIterator(gds, verbose=FALSE)
    cv <- currentVariants(it)
    checkEquals(names(cv), c("variant", "range"))
    seqClose(it)
}

test_resetIterator <- function() {
    gds <- SeqVarTools:::.testSeqVarData()
    it <- SeqVarBlockIterator(gds, variantBlock=1000, verbose=FALSE)
    while(iterateFilter(it, verbose=FALSE)) {}
    checkEquals(SeqVarTools:::.nVar(it), 0)
    resetIterator(it, verbose=FALSE)
    checkEquals(SeqVarTools:::.nVar(it), 1000)
    seqClose(it)
}

test_countSingletons <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- altDosage(gds)
  hasalt <- geno > 0 & !is.na(geno)
  singleton <- colSums(hasalt) == 1
  singmat <- matrix(singleton, nrow=nrow(geno), ncol=ncol(geno), byrow=TRUE)
  ct <- rowSums(hasalt & singmat)
  checkEquals(ct, countSingletons(gds), checkNames=FALSE)
  seqClose(gds)
}

test_hethom_sum <- function() {
  gds <- SeqVarTools:::.testData()
  nsamp <- SeqVarTools:::.nSamp(gds)
  nvar <- SeqVarTools:::.nVar(gds)
  
  miss <- missingGenotypeRate(gds)
  n.miss <- miss*nsamp
  n.nonmiss <- (1-miss)*nsamp
  checkEquals(rep(nsamp, nvar),
              heterozygosity(gds)*n.nonmiss + homozygosity(gds)*n.nonmiss + n.miss)
  
  miss <- missingGenotypeRate(gds, margin="by.sample")
  n.miss <- miss*nvar
  n.nonmiss <- (1-miss)*nvar
  checkEquals(rep(nvar, nsamp),
              heterozygosity(gds, margin="by.sample")*n.nonmiss +
              homozygosity(gds, margin="by.sample")*n.nonmiss +
              n.miss)
  seqClose(gds)
}

test_hom_sum <- function() {
  gds <- SeqVarTools:::.testData()
  checkEquals(homozygosity(gds, allele="any"),
              homozygosity(gds, allele="ref") + homozygosity(gds, allele="alt"))
  checkEquals(homozygosity(gds, allele="any", margin="by.sample"),
              homozygosity(gds, allele="ref", margin="by.sample") +
              homozygosity(gds, allele="alt", margin="by.sample"))
  seqClose(gds)
}

test_heterozygosity <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  checkIdentical(colSums(geno[1,,] != geno[2,,], na.rm=TRUE) /
                 colSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 heterozygosity(gds, margin="by.variant"))
  checkIdentical(rowSums(geno[1,,] != geno[2,,], na.rm=TRUE) /
                 rowSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 heterozygosity(gds, margin="by.sample"))
  seqClose(gds)
}

test_homozygosity_any <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  hom <- geno[1,,] == geno[2,,]
  checkIdentical(colSums(hom, na.rm=TRUE) /
                 colSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="any", margin="by.variant"))
  checkIdentical(rowSums(hom, na.rm=TRUE) /
                 rowSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="any", margin="by.sample"))
  seqClose(gds)
}

test_homozygosity_ref <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  hom <- geno[1,,] == geno[2,,] & geno[1,,] == 0
  checkIdentical(colSums(hom, na.rm=TRUE) /
                 colSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="ref", margin="by.variant"))
  checkIdentical(rowSums(hom, na.rm=TRUE) /
                 rowSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="ref", margin="by.sample"))
  seqClose(gds)
}

test_homozygosity_alt <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  hom <- geno[1,,] == geno[2,,] & geno[1,,] > 0
  checkIdentical(colSums(hom, na.rm=TRUE) /
                 colSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="alt", margin="by.variant"))
  checkIdentical(rowSums(hom, na.rm=TRUE) /
                 rowSums(!is.na(geno[1,,]) & !is.na(geno[2,,])),
                 homozygosity(gds, allele="alt", margin="by.sample"))
  seqClose(gds)
}

test_hethom <- function() {
  gds <- SeqVarTools:::.testData()
  hh <- hethom(gds)
  geno <- seqGetData(gds, "genotype")
  hom <- geno[1,,] == geno[2,,] & geno[1,,] > 0
  checkIdentical(rowSums(geno[1,,] != geno[2,,], na.rm=TRUE) /
                 rowSums(hom, na.rm=TRUE),
                 hh)
  checkEquals(heterozygosity(gds, margin="by.sample") /
              homozygosity(gds, allele="alt", margin="by.sample"),
              hh)
  seqClose(gds)
}

test_heterozygosity_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  het.var <- heterozygosity(gds, margin="by.variant")
  het.samp <- heterozygosity(gds, margin="by.sample")
  seqSetFilter(gds)
  checkIdentical(het.var,
                 applyMethod(gds, heterozygosity, variant=var.id, sample=samp.id,
                             margin="by.variant"))
  checkIdentical(het.samp,
                 applyMethod(gds, heterozygosity, variant=var.id, sample=samp.id,
                             margin="by.sample"))
  seqClose(gds)
}

test_homozygosity_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  hom.var <- homozygosity(gds, margin="by.variant")
  hom.samp <- homozygosity(gds, margin="by.sample")
  seqSetFilter(gds)
  checkIdentical(hom.var,
                 applyMethod(gds, homozygosity, variant=var.id, sample=samp.id,
                             margin="by.variant"))
  checkIdentical(hom.samp,
                 applyMethod(gds, homozygosity, variant=var.id, sample=samp.id,
                             margin="by.sample"))
  seqClose(gds)
}

test_parallel_bysample  <- function() {
    gds <- SeqVarTools:::.testData()
    
    ic1 <- inbreedCoeff(gds, margin="by.sample", parallel=FALSE)
    ic2 <- inbreedCoeff(gds, margin="by.sample", parallel=TRUE)
    checkEquals(ic1, ic2)

    miss1 <- missingGenotypeRate(gds, margin="by.sample", parallel=FALSE)
    miss2 <- missingGenotypeRate(gds, margin="by.sample", parallel=TRUE)
    checkEquals(miss1, miss2)
    
    het1 <- heterozygosity(gds, margin="by.sample", parallel=FALSE)
    het2 <- heterozygosity(gds, margin="by.sample", parallel=TRUE)
    checkEquals(het1, het2)

    seqClose(gds)
}

test_parallel_byvariant  <- function() {
    gds <- SeqVarTools:::.testData()
    
    ic1 <- inbreedCoeff(gds, margin="by.variant", parallel=FALSE)
    ic2 <- inbreedCoeff(gds, margin="by.variant", parallel=TRUE)
    checkEquals(ic1, ic2)

    miss1 <- missingGenotypeRate(gds, margin="by.variant", parallel=FALSE)
    miss2 <- missingGenotypeRate(gds, margin="by.variant", parallel=TRUE)
    checkEquals(miss1, miss2)

    hwe1 <- hwe(gds, parallel=FALSE)
    hwe2 <- hwe(gds, parallel=TRUE)
    checkEquals(hwe1, hwe2)

    het1 <- heterozygosity(gds, margin="by.variant", parallel=FALSE)
    het2 <- heterozygosity(gds, margin="by.variant", parallel=TRUE)
    checkEquals(het1, het2)
    
    seqClose(gds)
}

test_pca <- function() {
  gds <- SeqVarTools:::.testData()
  pc <- pca(gds, eigen.cnt=8)
  checkEquals(8, length(pc$eigenval))
  checkEquals(c(length(seqGetData(gds, "sample.id")), 8),
              dim(pc$eigenvect))
  seqClose(gds)
}

test_pca_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  pc <- pca(gds)
  seqSetFilter(gds)
  checkIdentical(pc,
                 applyMethod(gds, pca, variant=var.id, sample=samp.id))
  seqClose(gds)
}

test_applyNames <- function() {
  gds <- SeqVarTools:::.testData()
  samp.id <- as.character(seqGetData(gds, "sample.id"))
  var.id <- as.character(seqGetData(gds, "variant.id"))
  
  var <- matrix(nrow=length(samp.id), ncol=length(var.id),
                dimnames=list(sample=NULL, variant=NULL))
  checkIdentical(list(sample=samp.id, variant=var.id),
                 dimnames(SeqVarTools:::.applyNames(gds, var)))
  
  var <- matrix(nrow=length(var.id), ncol=length(samp.id),
                dimnames=list(variant=NULL, sample=NULL))
  checkIdentical(list(variant=var.id, sample=samp.id),
                 dimnames(SeqVarTools:::.applyNames(gds, var)))

  var <- array(dim=c(2, length(samp.id), length(var.id)),
               dimnames=list(NULL, sample=NULL, variant=NULL))
  checkIdentical(list(NULL, sample=samp.id, variant=var.id),
                 dimnames(SeqVarTools:::.applyNames(gds, var)))
  
  var <- matrix(nrow=10, ncol=10)
  checkIdentical(NULL, dimnames(SeqVarTools:::.applyNames(gds, var)))
  seqClose(gds)
}


test_isVariant <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  var <- (!is.na(geno[1,,]) & geno[1,,] > 0) | (!is.na(geno[2,,]) & geno[2,,] > 0)
  checkIdentical(var, isVariant(gds))
  seqClose(gds)
}

test_isVariant_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  var <- isVariant(gds)
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(var,
                 applyMethod(gds, isVariant, variant=var.id, sample=samp.id))
  seqClose(gds)
}


test_getGenotype <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  gc <- matrix(paste(geno[1,,], geno[2,,], sep="/"),
               nrow=dim(geno)[2], ncol=dim(geno)[3])
  gc[gc == "NA/NA"] <- NA
  checkIdentical(gc, getGenotype(gds, use.names=FALSE))
  dimnames(gc) <- list(sample=seqGetData(gds, "sample.id"),
                       variant=seqGetData(gds, "variant.id"))
  checkIdentical(gc, getGenotype(gds, use.names=TRUE))
  seqClose(gds)
}

test_getGenotype_phased <- function() {
  gdsfmt::showfile.gds(closeall=TRUE)
  gds <- seqOpen(system.file("extdata", "gl_chr1.gds", package="SeqVarTools"))
  geno <- seqGetData(gds, "genotype")
  gc <- matrix(paste(geno[1,,], geno[2,,], sep="|"),
               nrow=dim(geno)[2], ncol=dim(geno)[3])
  gc[gc == "NA/NA"] <- NA
  checkIdentical(gc, getGenotype(gds, use.names=FALSE))
  seqClose(gds)
}

test_getGenotype_haploid <- function() {
  gdsfmt::showfile.gds(closeall=TRUE)
  gds <- seqOpen(system.file("extdata", "1KG_chrY.gds", package="SeqVarTools"))
  geno <- seqGetData(gds, "genotype")
  gc <- geno[1,,]
  checkIdentical(gc, getGenotype(gds, use.names=FALSE))
  dimnames(gc) <- list(sample=seqGetData(gds, "sample.id"),
                       variant=seqGetData(gds, "variant.id"))
  checkIdentical(gc, getGenotype(gds, use.names=TRUE))
  seqClose(gds)
}

test_getGenotype_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  geno <- getGenotype(gds)
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(geno,
                 applyMethod(gds, getGenotype, variant=var.id, sample=samp.id))
  seqClose(gds)
}


test_getGenotypeAlleles <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  alleles <- list(refChar(gds), altChar(gds, n=1), altChar(gds, n=2))
  for (i in 1:2) {
    for (j in 1:dim(geno)[2]) {
      for (a in 0:2) {
        x <- geno[i,j,] %in% a
        geno[i,j,x] <- alleles[[a+1]][x]
      }
    }
  }
  gc <- matrix(paste(geno[1,,], geno[2,,], sep="/"),
               nrow=dim(geno)[2], ncol=dim(geno)[3])
  gc[gc == "NA/NA"] <- NA
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=FALSE))
  dimnames(gc) <- list(sample=seqGetData(gds, "sample.id"),
                       variant=seqGetData(gds, "variant.id"))
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=TRUE))
  seqClose(gds)
}

test_getGenotypeAlleles_phased <- function() {
  gdsfmt::showfile.gds(closeall=TRUE)
  gds <- seqOpen(system.file("extdata", "gl_chr1.gds", package="SeqVarTools"))
  geno <- seqGetData(gds, "genotype")
  alleles <- list(refChar(gds), altChar(gds, n=1), altChar(gds, n=2))
  for (i in 1:2) {
    for (j in 1:dim(geno)[2]) {
      for (a in 0:2) {
        x <- geno[i,j,] %in% a
        geno[i,j,x] <- alleles[[a+1]][x]
      }
    }
  }
  gc <- matrix(paste(geno[1,,], geno[2,,], sep="|"),
               nrow=dim(geno)[2], ncol=dim(geno)[3])
  gc[gc == "NA/NA"] <- NA
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=FALSE))

  ## sort
  gc <- matrix(paste(pmin(geno[1,,], geno[2,,]),
                     pmax(geno[1,,], geno[2,,]), sep="/"),
               nrow=dim(geno)[2], ncol=dim(geno)[3])
  gc[gc == "NA/NA"] <- NA
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=FALSE, sort=TRUE))
  seqClose(gds)
}

test_getGenotypeAlleles_haploid <- function() {
  gdsfmt::showfile.gds(closeall=TRUE)
  gds <- seqOpen(system.file("extdata", "1KG_chrY.gds", package="SeqVarTools"))
  gc <- seqGetData(gds, "genotype")[1,,]
  alleles <- list(refChar(gds), altChar(gds, n=1), altChar(gds, n=2), altChar(gds, n=3), altChar(gds, n=4))
  for (j in 1:nrow(gc)) {
      for (a in 0:4) {
          x <- gc[j,] %in% a
          gc[j,x] <- alleles[[a+1]][x]
      }
  }
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=FALSE))
  dimnames(gc) <- list(sample=seqGetData(gds, "sample.id"),
                       variant=seqGetData(gds, "variant.id"))
  checkIdentical(gc, getGenotypeAlleles(gds, use.names=TRUE))
  seqClose(gds)
}

test_getGenotypeAlleles_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  geno <- getGenotypeAlleles(gds)
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(geno,
                 applyMethod(gds, getGenotypeAlleles, variant=var.id, sample=samp.id))
  seqClose(gds)
}

test_refDosage <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- getGenotype(gds)
  rd <- refDosage(gds)
  checkIdentical(geno %in% "0/0", rd %in% 2)
  checkIdentical(geno %in% c("0/1", "1/0"), rd %in% 1)
  checkIdentical(geno %in% "1/1", rd %in% 0)
  checkIdentical(is.na(geno), is.na(rd))
  seqClose(gds)
}

test_refDosage_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  geno <- refDosage(gds)
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(geno,
                 applyMethod(gds, refDosage, variant=var.id, sample=samp.id))
  seqClose(gds)
}

test_altDosage <- function() {
  gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
  gds <- seqOpen(gdsfile)
  geno <- getGenotype(gds)
  d <- altDosage(gds)
  hasref <- geno %in% paste0("0/", 0:(max(nAlleles(gds))-1))
  hasref[is.na(geno)] <- NA
  hasalt <- !(geno %in% "0/0")
  hasalt[is.na(geno)] <- NA
  checkEquals(!hasref, as.vector(d == 2))
  checkEquals(hasalt & hasref, as.vector(d == 1))
  checkEquals(geno == "0/0", d == 0)
  checkIdentical(is.na(geno), is.na(d))
  checkEquals(refDosage(gds), 2-d)
  seqClose(gds)
}

## former alleleDosage method for list, summing over all requested alleles
.alleleDosageSum <- function(gdsobj, n, use.names=TRUE) {
    stopifnot(length(n) == SeqVarTools:::.nVar(gdsobj))
    d <- seqApply(gdsobj, "genotype",
                  function(index, x) {
                      tmp <- matrix(x %in% n[[index]], ncol=ncol(x),
                                    nrow=nrow(x))
                      tmp[is.na(x)] <- NA
                      colSums(tmp)
                  },
                  margin="by.variant", as.is="list",
                  var.index="relative")
    d <- matrix(unlist(d, use.names=FALSE), ncol=length(d))
    if (use.names) {
        dimnames(d) <- list(sample=NULL, variant=NULL)
        SeqVarTools:::.applyNames(gdsobj, d)
    } else d
}

test_alleleDosage <- function() {
  gds <- SeqVarTools:::.testData()
  checkEquals(refDosage(gds), alleleDosage(gds, n=0))
  checkEquals(altDosage(gds), alleleDosage(gds, n=1))
  seqClose(gds)
  
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
  gds <- seqOpen(gdsfile)
  checkEquals(refDosage(gds), alleleDosage(gds, n=0))
  checkEquals(altDosage(gds), 2-alleleDosage(gds, n=0))

  ## different alleles for each variant
  nalleles <- nAlleles(gds)
  n <- sapply(nalleles, function(x) sample(0:(x-1), 1))
  geno <- getGenotype(gds)
  cnt <- matrix(nrow=nrow(geno), ncol=ncol(geno), dimnames=dimnames(geno))
  for (i in 1:nrow(geno)) {
      cnt[i,] <- stringr::str_count(geno[i,], as.character(n))
  }
  checkEquals(cnt, alleleDosage(gds, n))

  ## n is a list
  n <- lapply(nalleles, function(x) sample(0:(x-1), sample(1:x, 1)))
  cnt <- list()
  for (i in 1:ncol(geno)) {
  ##     tmp <- lapply(n[[i]], function(x) stringr::str_count(geno[,i], as.character(x)))
  ##     cnt[,i] <- colSums(do.call(rbind, tmp))
      cnt[[i]] <- do.call(cbind, lapply(n[[i]], function(x) stringr::str_count(geno[,i], as.character(x))))
      dimnames(cnt[[i]]) <- list(sample=rownames(geno), allele=n[[i]])
  }
  ad <- alleleDosage(gds, n)
  checkEquals(cnt, ad, checkNames=FALSE)

  adsum <- do.call(cbind, lapply(ad, function(x) rowSums(x)))
  names(dimnames(adsum)) <- c("sample", "variant")
  checkEquals(.alleleDosageSum(gds, n), adsum)
  
  ## invalid allele
  checkException(alleleDosage(gds, n=10))
  checkException(alleleDosage(gds, n=c(1,0)))
  
  seqClose(gds)
}

test_expandedAltDosage <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    ead <- expandedAltDosage(gds)
    checkEquals(sum(nAlleles(gds) - 1), ncol(ead)) 

    seqSetFilter(gds, variant.sel=(nAlleles(gds) == 3), verbose=FALSE)
    ad <- lapply(1:2, function(n) alleleDosage(gds, n=n))
    ead <- expandedAltDosage(gds)
    checkEquals(ad[[1]], ead[,c(TRUE,FALSE)])
    checkEquals(ad[[2]], ead[,c(FALSE,TRUE)])

    seqResetFilter(gds, verbose=FALSE)
    seqSetFilter(gds, variant.sel=(nAlleles(gds) == 4), verbose=FALSE)
    ad <- lapply(1:3, function(n) alleleDosage(gds, n=n))
    ead <- expandedAltDosage(gds)
    checkEquals(ad[[1]], ead[,c(TRUE,FALSE,FALSE)])
    checkEquals(ad[[2]], ead[,c(FALSE,TRUE,FALSE)])
    checkEquals(ad[[3]], ead[,c(FALSE,FALSE,TRUE)])

    seqClose(gds)
}

test_expandedVariantIndex <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    ind <- expandedVariantIndex(gds)
    checkEquals(sum(nAlleles(gds) - 1), length(ind)) 
    ead <- expandedAltDosage(gds)
    checkEquals(ncol(ead), length(ind))
    checkEquals(colnames(ead), as.character(seqGetData(gds, "variant.id")[ind]))
    checkEquals(as.vector(table(nAlleles(gds))), as.vector(table(table(ind))))
    seqClose(gds)
}

test_expandedVariantIndex_noAlt <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "gl_chr1.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    ind <- expandedVariantIndex(gds)
    checkEquals(1:9, ind) 
    seqClose(gds)
}

test_variantInfo <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    
    v <- variantInfo(gds, alleles=FALSE, expanded=FALSE)
    checkEquals(SeqVarTools:::.nVar(gds), nrow(v))
    
    v <- variantInfo(gds, alleles=TRUE, expanded=FALSE)
    checkTrue(all(c("ref", "alt") %in% names(v)))
    
    v <- variantInfo(gds, alleles=FALSE, expanded=TRUE)
    na <- table(nAlleles(gds))
    ai <- sapply(1:length(na), function(i) sum(na[i:length(na)]))
    checkEquals(ai, as.vector(table(v$allele.index)))

    v <- variantInfo(gds, alleles=TRUE, expanded=TRUE)
    checkEquals(ai, as.vector(table(v$allele.index)))
    for (i in 1:6) {
        ac <- altChar(gds, n=i)
        checkEquals(ac[!is.na(ac)], v$alt[v$allele.index == i])
    }
    
    seqClose(gds)
}

test_empty <- function() {
    gds <- SeqVarTools:::.testData()
    SeqVarTools:::.emptyVarFilter(gds)
    checkEquals(0, ncol(getGenotype(gds)))
    checkEquals(0, ncol(getGenotypeAlleles(gds)))
    checkEquals(0, ncol(refDosage(gds)))
    checkEquals(0, ncol(altDosage(gds)))
    checkEquals(0, ncol(expandedAltDosage(gds)))
    checkEquals(0, ncol(alleleDosage(gds, n=0)))
    checkEquals(0, length(expandedVariantIndex(gds)))
    checkEquals(0, nrow(variantInfo(gds)))
    checkEquals(0, nrow(variantInfo(gds, expanded=TRUE)))
    checkEquals(0, nrow(variantInfo(gds, alleles=FALSE, expanded=TRUE)))
    seqClose(gds)
}

.parseVariableLength <- function(x) {
  if (all(x$length == 1)) {
    x$data
  } else {
    var <- array(dim=c(max(x$length), nrow(x$data), length(x$length)))

    ## assign each element of length to an index of first array dimension
    n.ind <- rep(NA, ncol(x$data))
    j <- 1
    for (i in 1:length(x$length)) {
      len <- x$length[i]
      if (len > 0) {
        n.ind[j:(j + len - 1)] <- 1:len
        j <- j + len
      }
    }

    ## for each index of first array dimension, get values
    for (n in 1:dim(var)[1]) {
      var.ind <- which(x$length >= n)
      var[n,,var.ind] <- x$data[,n.ind == n]
    }

    ## if first array dimension is 1, simplify to a matrix
    if (dim(var)[1] == 1) {
      var <- var[1,,]
    }
    var
  }
}

test_parseVariableLength <- function() {
  x <- list(length=c(3,3,2,1,3),
            data=matrix(c(1,2,3,1,2,3,1,2,1,1,2,3), nrow=4, ncol=12, byrow=TRUE))
  y <- .parseVariableLength(x)
  for (n in 1:3) checkTrue(all(y[n,,] %in% c(n,NA)))
  checkTrue(all(is.na(y[2,,4])))
  checkTrue(all(is.na(y[3,,3:4])))
}

test_parseVariableLength_length1 <- function() {
  x <- list(length=c(1,1,1,1,1),
            data=matrix(1:5, nrow=4, ncol=5, byrow=TRUE))
  checkIdentical(x$data, .parseVariableLength(x))
}

test_parseVariableLength_length0 <- function() {
  x <- list(length=c(2,2,2,0,2),
            data=matrix(rep(1:2,4), nrow=3, ncol=8, byrow=TRUE))
  y <- array(rep(1:2,15), dim=c(2,3,5))
  y[,,4] <- NA
  checkIdentical(y, .parseVariableLength(x))
}

test_parseVariableLength_length01 <- function() {
  x <- list(length=c(1,1,1,0,1),
            data=matrix(1:4, nrow=3, ncol=4, byrow=TRUE))
  y <- matrix(c(1,2,3,NA,4), nrow=3, ncol=5, byrow=TRUE)
  checkEquals(y, .parseVariableLength(x))
}

test_getVariableLength_GL <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  gds <- seqOpen(system.file("extdata", "gl_chr1.gds", package="SeqVarTools"))
  gl <- getVariableLengthData(gds, "annotation/format/GL", use.names=FALSE)
  dimnames(gl) <- NULL
  checkIdentical(.parseVariableLength(seqGetData(gds, "annotation/format/GL")), gl)
  seqClose(gds)
}

test_getVariableLength_apply <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  gds <- seqOpen(system.file("extdata", "gl_chr1.gds", package="SeqVarTools"))
  var.id <- 2:6
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id, verbose=FALSE)
  vl <- getVariableLengthData(gds, "annotation/format/GL")
  seqSetFilter(gds, verbose=FALSE)
  checkIdentical(vl,
                 applyMethod(gds, getVariableLengthData, variant=var.id,
                             sample=samp.id, var.name="annotation/format/GL"))
  seqClose(gds)
}

## something is badly formatted in the AD field and SeqArray can no longer handle it
## test_getVariableLength_AD <- function() {
##     gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
##     gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
##     gds <- seqOpen(gdsfile)
##     ad <- getVariableLengthData(gds, "annotation/format/AD", use.names=FALSE)
##     dimnames(ad) <- NULL
##     checkIdentical(.parseVariableLength(seqGetData(gds, "annotation/format/AD")), ad)
##     seqClose(gds)
## }

test_getVariableLength_AB <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    ab <- getVariableLengthData(gds, "annotation/format/AB", use.names=FALSE)
    dimnames(ab) <- NULL
    checkIdentical(.parseVariableLength(seqGetData(gds, "annotation/format/AB")), ab)
    seqClose(gds)
}

test_rangesTo <- function() {
    gds <- SeqVarTools:::.testData()
    ranges <- GRanges("1", IRanges(start=c(1e6,3e6), end=c(2e6,4e6)))
    checkEquals(1:7, SeqVarTools:::.rangesToSel(gds, ranges))
    checkEquals(1:7, SeqVarTools:::.rangesToID(gds, ranges))

    # check empty ranges
    ranges <- GRanges("X", IRanges(start=c(1e6,3e6), end=c(2e6,4e6)))
    checkEquals(0, length(SeqVarTools:::.rangesToSel(gds, ranges)))
    seqClose(gds)
}

test_nSamp <- function() {
  gds <- SeqVarTools:::.testData()
  n <- sum(seqGetFilter(gds)$sample.sel)
  checkEquals(n, SeqVarTools:::.nSamp(gds))

  n <- 10
  seqSetFilter(gds, sample.sel=1:n)
  checkEquals(n, SeqVarTools:::.nSamp(gds))
  seqClose(gds)
}

test_nVar <- function() {
  gds <- SeqVarTools:::.testData()
  n <- sum(seqGetFilter(gds)$variant.sel)
  checkEquals(n, SeqVarTools:::.nVar(gds))

  n <- 10
  seqSetFilter(gds, variant.sel=1:n)
  checkEquals(n, SeqVarTools:::.nVar(gds))
  seqClose(gds)
}

test_nSampUnfiltered <- function() {
  gds <- SeqVarTools:::.testData()
  n <- SeqVarTools:::.nSamp(gds)
  seqSetFilter(gds, sample.sel=1:10)
  checkEquals(n, SeqVarTools:::.nSampUnfiltered(gds))
  seqClose(gds)
}

test_nVarUnfiltered <- function() {
  gds <- SeqVarTools:::.testData()
  n <- SeqVarTools:::.nVar(gds)
  seqSetFilter(gds, sample.sel=1:10)
  checkEquals(n, SeqVarTools:::.nVarUnfiltered(gds))
  seqClose(gds)
}

test_emptyGenoMatrix <- function() {
    gds <- SeqVarTools:::.testData()
    SeqVarTools:::.emptyVarFilter(gds)
    checkTrue(SeqVarTools:::.emptyDim(gds))
    m <- SeqVarTools:::.emptyGenoMatrix(gds, use.names=TRUE)
    checkEquals(SeqVarTools:::.nSamp(gds), nrow(m))
    checkEquals(0, ncol(m))
    checkEquals(seqGetData(gds, "sample.id"), rownames(m))

    seqResetFilter(gds, verbose=FALSE)
    SeqVarTools:::.emptySampFilter(gds)
    checkTrue(SeqVarTools:::.emptyDim(gds))
    m <- SeqVarTools:::.emptyGenoMatrix(gds, use.names=TRUE)
    checkEquals(SeqVarTools:::.nVar(gds), ncol(m))
    checkEquals(0, nrow(m))
    checkEquals(as.character(seqGetData(gds, "variant.id")), colnames(m))

    seqClose(gds)
}

test_nSampObserved <- function() {
    gds <- SeqVarTools:::.testData()
    nso <- SeqVarTools:::.nSampObserved(gds)
    checkEquals(SeqVarTools:::.nVar(gds), length(nso))
    checkEquals(SeqVarTools:::.nSamp(gds), max(nso))
    seqClose(gds)
}

test_nSamp_empty <- function() {
    gds <- SeqVarTools:::.testData()
    seqSetFilter(gds, sample.sel=rep(SeqVarTools:::.nSamp(gds), FALSE), verbose=FALSE)
    checkEquals(0, SeqVarTools:::.nSamp(gds))
    checkEquals(rep(0, SeqVarTools:::.nVar(gds)), SeqVarTools:::.nSampObserved(gds))
    seqClose(gds)
}

test_permute <- function() {
    x <- matrix(c(0,0,1,1,NA,NA), nrow=2)
    p <- SeqVarTools:::.permuteGenotypes(x)
    checkEquals(c(0,0,1,1), sort(as.vector(p[,1:2])))
    checkTrue(all(is.na(p[,3])))
}

test_count <- function() {
    gds <- SeqVarTools:::.testData()
    filt <- nAlleles(gds) == 2
    seqSetFilter(gds, variant.sel=filt)
    geno <- getGenotype(gds)
    counts <- data.frame(nAA=colSums(geno == "0/0", na.rm=TRUE),
                         nAa=colSums(geno == "0/1" | geno == "1/0", na.rm=TRUE),
                         naa=colSums(geno == "1/1", na.rm=TRUE),
                         row.names=1:ncol(geno))
    checkEquals(counts, SeqVarTools:::.countGenotypes(gds))
    perm <- SeqVarTools:::.countGenotypes(gds, permute=TRUE)
    checkEquals(rowSums(counts), rowSums(perm), checkNames=FALSE)
    checkEquals(2*counts$nAA + counts$nAa, 2*perm$nAA + perm$nAa, checkNames=FALSE)
    seqClose(gds)
}

test_hwe <- function() {
    gds <- SeqVarTools:::.testData()
    af <- alleleFrequency(gds)
    biallelic <- nAlleles(gds) == 2
    mono <- af %in% c(0,1)
    hw <- hwe(gds, permute=FALSE)
    checkEquals(mono | !biallelic, is.na(hw$p), checkNames=FALSE)
    
    filt <- biallelic & !mono
    seqSetFilter(gds, variant.sel=filt)
    hw <- hwe(gds, permute=FALSE)
    checkEquals(af[filt], hw$afreq)
    checkEquals(inbreedCoeff(gds), hw$f)

    hw <- hwe(gds, permute=TRUE)
    checkEquals(af[filt], hw$afreq)
    seqClose(gds)
}

setVariantID <- function(gdsfile, variant.id) {
  stopifnot(file.exists(gdsfile))
  if (length(unique(variant.id)) != length(variant.id)) {
    stop("variant.id must be unique")
  }

  gdsobj <- openfn.gds(gdsfile, readonly=FALSE)
  node <- index.gdsn(gdsobj, "variant.id")
  curr.id <- read.gdsn(node)
  if (length(curr.id) != length(variant.id)) {
    closefn.gds(gdsobj)
    stop(paste("variant.id must have length", length(curr.id)))
  }

  ## delete existing node and create a new one,
  ## as data type may be different
  compress <- objdesp.gdsn(node)$compress
  add.gdsn(gdsobj, "variant.id", variant.id,
           compress=compress, closezip=TRUE, replace=TRUE)
  closefn.gds(gdsobj)
}

## constructor
SeqVarData <- function(gds, sampleData, variantData) {
    closeOnFail <- FALSE
    if (is.character(gds) & length(gds) == 1) {
        gds <- seqOpen(gds)
        closeOnFail <- TRUE
    } else if (!is(gds, "SeqVarGDSClass")) {
        stop("gds must be a SeqVarGDSClass object or a valid filename")
    }
    class(gds) <- "SeqVarData"

    if (missing(sampleData)) {
        sampleData <- AnnotatedDataFrame(data.frame(matrix(nrow=.nSampUnfiltered(gds), ncol=0)))
    }
    gds@sampleData <- sampleData
    
    if (missing(variantData)) {
        variantData <- AnnotatedDataFrame(data.frame(matrix(nrow=.nVarUnfiltered(gds), ncol=0)))
    }
    gds@variantData <- variantData
    
    #check <- validObject(gds, test=TRUE)
    check <- .validObject(gds)
    if (is.character(check)) {
        if (closeOnFail) seqClose(gds)
        stop(check)
    }
    gds
}

#setValidity("SeqVarData",
.validObject <- function(object) {
                if (ncol(sampleData(object)) > 0) {
                    if (!("sample.id" %in% varLabels(sampleData(object)))) {
                        return("sampleData does not have column sample.id")
                    }
                    if (!identical(seqGetData(object, "sample.id"),
                                   sampleData(object)$sample.id)) {
                        return("sample.id in sampleData does not match GDS file")
                    }
                }
                if (ncol(variantData(object)) > 0) {
                    if (!("variant.id" %in% varLabels(variantData(object)))) {
                        return("variantData does not have column variant.id")
                    }
                    if (!identical(seqGetData(object, "variant.id"),
                                   variantData(object)$variant.id)) {
                        return("variant.id in variantData does not match GDS file")
                    }
                }
                TRUE
            }#)

setMethod("show",
          "SeqVarData",
          function(object) {
              cat(class(object), "object\n")
              cat(" | GDS:\n")
              print(object)
              cat(" | sampleData:\n")
              show(sampleData(object))
              cat(" | variantData:\n")
              show(variantData(object))
          })


setMethod("sampleData",
          "SeqVarData",
          function(x) {
              sd <- x@sampleData
              samp.sel <- seqGetFilter(x)$sample.sel
              if (sum(samp.sel) == length(samp.sel)) {
                  return(sd)
              } else {
                  return(sd[samp.sel,])
              }
          })

setReplaceMethod("sampleData",
                 c("SeqVarData", "AnnotatedDataFrame"),
                 function(x, value) {
                     seqSetFilter(x, sample.sel=1:.nSampUnfiltered(x), action="push+set", verbose=FALSE)
                     if (!identical(seqGetData(x, "sample.id"), value$sample.id)) {
                         stop("sample.id does not match GDS file")
                     }
                     seqSetFilter(x, action="pop", verbose=FALSE)
                     slot(x, "sampleData") <- value
                     x
                 })

            
setMethod("variantData",
          "SeqVarData",
          function(x) {
              vd <- x@variantData
              var.sel <- seqGetFilter(x)$variant.sel
              if (sum(var.sel) == length(var.sel)) {
                  return(vd)
              } else {
                  return(vd[var.sel,])
              }
          })

setReplaceMethod("variantData",
                 c("SeqVarData", "AnnotatedDataFrame"),
                 function(x, value) {
                     seqSetFilter(x, variant.sel=1:.nVarUnfiltered(x), action="push+set", verbose=FALSE)
                     if (!identical(seqGetData(x, "variant.id"), value$variant.id)) {
                         stop("variant.id does not match GDS file")
                     }
                     seqSetFilter(x, action="pop", verbose=FALSE)
                     slot(x, "variantData") <- value
                     x
                 })


setMethod("granges",
          "SeqVarData",
          function(x, ...) {
              gr <- callNextMethod(x, ...)
              df <- pData(variantData(x))
              df$variant.id <- NULL # redundant with names of gr
              mcols(gr) <- cbind(mcols(gr), as(df, "DataFrame"))
              gr
          })


setMethod("validateSex",
          "SeqVarData",
          function(x) {
              sex <- sampleData(x)$sex
              if (!is.null(sex)) {          
                  if (all(sex %in% c(1,2,NA))) {
                      sex <- c("M", "F")[as.integer(sex)]
                  }
                  if (!all(sex %in% c("M", "F", NA))) {
                      sex <- NULL
                  }
              }
              sex
          })


setMethod("alleleFrequency",
          "SeqVarData",
          function(gdsobj, n=0, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE, genome.build=c("hg19", "hg38"), parallel=FALSE) {
              if (!sex.adjust) {
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }
              
              # check chromosome
              chr <- chromWithPAR(gdsobj, genome.build)
              if (!any(chr %in% c("X", "Y"))) {
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }

              # check sex
              sex <- validateSex(gdsobj)
              if (is.null(sex)) {
                  warning("No valid sex coding provided in sampleData. Frequencies will not be calculated correctly for X and Y chromosomes.")
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }
              female <- sex %in% "F"
              male <- sex %in% "M"

              # check ploidy
              if (.ploidy(gdsobj) == 1) male.diploid <- FALSE
              
              # empty vector to fill in
              freq <- rep(NA, length(chr))
              if (use.names) names(freq) <- seqGetData(gdsobj, "variant.id")

              # autosomal
              auto <- !(chr %in% c("X", "Y"))
              if (any(auto)) {
                  seqSetFilter(gdsobj, variant.sel=auto, action="push+intersect", verbose=FALSE)
                  freq[auto] <- callNextMethod(gdsobj, n=n, use.names=FALSE, parallel=parallel)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }
              
              # X chrom
              X <- chr %in% "X"
              if (any(X)) {
                  seqSetFilter(gdsobj, sample.sel=female, variant.sel=X, action="push+intersect", verbose=FALSE)
                  #freq.X.F <- callNextMethod(gdsobj, n=n, use.names=FALSE)
                  n.F <- .nSampObserved(gdsobj)
                  #count.X.F <- freq.X.F * 2*n.F
                  count.X.F <- seqAlleleCount(gdsobj, ref.allele=n, parallel=parallel)
                  count.X.F[is.na(count.X.F)] <- 0L
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=X, action="push+intersect", verbose=FALSE)
                  #freq.X.M <- callNextMethod(gdsobj, n=n, use.names=FALSE)
                  n.M <- .nSampObserved(gdsobj)
                  #count.X.M <- freq.X.M * n.M
                  count.X.M <- seqAlleleCount(gdsobj, ref.allele=n, parallel=parallel)
                  count.X.M[is.na(count.X.M)] <- 0L
                  if (male.diploid) {
                      count.X.M <- count.X.M / 2
                  }
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
                  
                  freq[X] <- (count.X.F + count.X.M)/(2*n.F + n.M)
              }

              # Y chrom
              Y <- chr %in% "Y"
              if (any(Y)) {
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=Y, action="push+intersect", verbose=FALSE)
                  freq[Y] <- callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }
              freq
          })


setMethod("alleleCount",
          "SeqVarData",
          function(gdsobj, n=0, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE, genome.build=c("hg19", "hg38"), parallel=FALSE) {
              if (!sex.adjust) {
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }
              
              # check chromosome
              chr <- chromWithPAR(gdsobj, genome.build)
              if (!any(chr %in% c("X", "Y"))) {
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }

              # check sex
              sex <- validateSex(gdsobj)
              if (is.null(sex)) {
                  warning("No valid sex coding provided in sampleData. Counts will not be calculated correctly for X and Y chromosomes.")
                  return(callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel))
              }
              female <- sex %in% "F"
              male <- sex %in% "M"

              # check ploidy
              if (.ploidy(gdsobj) == 1) male.diploid <- FALSE
              
              # empty vector to fill in
              count <- rep(NA, length(chr))
              if (use.names) names(count) <- seqGetData(gdsobj, "variant.id")

              # autosomal
              auto <- !(chr %in% c("X", "Y"))
              if (any(auto)) {
                  seqSetFilter(gdsobj, variant.sel=auto, action="push+intersect", verbose=FALSE)
                  count[auto] <- callNextMethod(gdsobj, n=n, use.names=FALSE, parallel=parallel)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }
              
              # X chrom
              X <- chr %in% "X"
              if (any(X)) {
                  seqSetFilter(gdsobj, sample.sel=female, variant.sel=X, action="push+intersect", verbose=FALSE)
                  count.X.F <- callNextMethod(gdsobj, n=n, use.names=FALSE, parallel=parallel)
                  count.X.F[is.na(count.X.F)] <- 0L
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=X, action="push+intersect", verbose=FALSE)
                  count.X.M <- callNextMethod(gdsobj, n=n, use.names=FALSE, parallel=parallel)
                  count.X.M[is.na(count.X.M)] <- 0L
                  if (male.diploid) {
                      ## correct for X chromosome coded as diploid in males
                      count.X.M <- count.X.M / 2 
                  }
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
                  
                  count[X] <- count.X.F + count.X.M
              }

              # Y chrom
              Y <- chr %in% "Y"
              if (any(Y)) {
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=Y, action="push+intersect", verbose=FALSE)
                  count.Y <- callNextMethod(gdsobj, n=n, use.names=use.names, parallel=parallel)
                  count.Y[is.na(count.Y)] <- 0L
                  count[Y] <- count.Y
                  if (male.diploid) {
                      ## correct for Y chromosome coded as diploid in males
                      count[Y] <- count[Y] / 2
                  }
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }

              count
          })


setMethod("minorAlleleCount",
          "SeqVarData",
          function(gdsobj, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE, genome.build=c("hg19", "hg38"), parallel=FALSE) {
              if (!sex.adjust) {
                  return(callNextMethod(gdsobj, use.names=use.names, parallel=parallel))
              }
              
              # check chromosome
              chr <- chromWithPAR(gdsobj, genome.build)
              if (!any(chr %in% c("X", "Y"))) {
                  return(callNextMethod(gdsobj, use.names=use.names, parallel=parallel))
              }

              # check sex
              sex <- validateSex(gdsobj)
              if (is.null(sex)) {
                  warning("No valid sex coding provided in sampleData. Counts will not be calculated correctly for X and Y chromosomes.")
                  return(callNextMethod(gdsobj, use.names=use.names, parallel=parallel))
              }
              female <- sex %in% "F"
              male <- sex %in% "M"

              # check ploidy
              if (.ploidy(gdsobj) == 1) male.diploid <- FALSE
              
              # empty vector to fill in
              count <- rep(NA, length(chr))
              possible <- rep(NA, length(chr))
              if (use.names) names(count) <- seqGetData(gdsobj, "variant.id")

              # autosomal
              auto <- !(chr %in% c("X", "Y"))
              if (any(auto)) {
                  seqSetFilter(gdsobj, variant.sel=auto, action="push+intersect", verbose=FALSE)
                  count[auto] <- seqAlleleCount(gdsobj, parallel=parallel)
                  possible[auto] <- 2L * .nSampObserved(gdsobj)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }
              
              # X chrom
              X <- chr %in% "X"
              if (any(X)) {
                  seqSetFilter(gdsobj, sample.sel=female, variant.sel=X, action="push+intersect", verbose=FALSE)
                  count.X.F <- seqAlleleCount(gdsobj, parallel=parallel)
                  count.X.F[is.na(count.X.F)] <- 0L
                  possible.X.F <- 2L * .nSampObserved(gdsobj)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=X, action="push+intersect", verbose=FALSE)
                  count.X.M <- seqAlleleCount(gdsobj, parallel=parallel)
                  count.X.M[is.na(count.X.M)] <- 0L
                  if (male.diploid) {
                      ## correct for X chromosome coded as diploid in males
                      count.X.M <- count.X.M / 2 
                  }
                  possible.X.M <- .nSampObserved(gdsobj)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
                  
                  count[X] <- count.X.F + count.X.M
                  possible[X] <- possible.X.F + possible.X.M
              }

              # Y chrom
              Y <- chr %in% "Y"
              if (any(Y)) {
                  seqSetFilter(gdsobj, sample.sel=male, variant.sel=Y, action="push+intersect", verbose=FALSE)
                  count.Y <- seqAlleleCount(gdsobj, parallel=parallel)
                  count.Y[is.na(count.Y)] <- 0L
                  count[Y] <- count.Y
                  if (male.diploid) {
                      ## correct for Y chromosome coded as diploid in males
                      count[Y] <- count[Y] / 2 
                  }
                  possible[Y] <- .nSampObserved(gdsobj)
                  seqSetFilter(gdsobj, action="pop", verbose=FALSE)
              }

              pmin(count, possible - count)
          })

setClass("SeqVarData", contains="SeqVarGDSClass",
         slots=c(sampleData = "AnnotatedDataFrame",
                 variantData = "AnnotatedDataFrame"))

setClass("SeqVarIterator", contains="SeqVarData",
         slots=c(variantFilter="list",
                 lastFilter="environment")) # allow pass-by-reference for this slot

setClass("SeqVarBlockIterator", contains="SeqVarIterator",
         slots=c(variantBlock="integer")) # number of variants

setClass("SeqVarRangeIterator", contains="SeqVarIterator",
         slots=c(variantRanges="GRanges"))

setClass("SeqVarWindowIterator", contains="SeqVarRangeIterator",
         slots=c(windowSize="integer", # base pairs
                 windowShift="integer"))

setClass("SeqVarListIterator", contains="SeqVarIterator",
         slots=c(variantRanges="GRangesList"))

.waldTest <- function(Est, cov) {
    if (length(Est) == 1) {
        W <- (Est^2)/cov
    } else {
        W <- as.numeric(t(Est) %*% solve(cov) %*% Est)
    }
    pval <- pchisq(W, df=length(Est), lower.tail=FALSE)
    c(W, pval)
}

.runRegression <- function(model.string, model.data, model.type) {
    model.formula <- as.formula(model.string)
    tryCatch({
        if (model.type == "linear") {
            mod <- lm(model.formula, data=model.data)
        } else if (model.type == "logistic") {
            mod <- glm(model.formula, data=model.data, family=binomial())
        }
    
        Est <- unname(coef(mod)["genotype"])
        cov <- vcov(mod)["genotype","genotype"]
        
        c(Est, sqrt(cov), .waldTest(Est, cov))
    }, warning=function(w) rep(NA, 4), error=function(e) rep(NA, 4))
}

.runFirth <- function(model.string, model.data, geno.index=NULL) {
    model.formula <- as.formula(model.string)
    tryCatch({
        mod <- tryCatch({
            logistf(model.formula, data=model.data, plconf=geno.index, dataout=FALSE)
        }, error=function(e) {
            ## test will fail if geno.index is too large
            geno.index <- which(colnames(model.matrix(model.formula, model.data)) == "genotype")
            logistf(model.formula, data=model.data, plconf=geno.index, dataout=FALSE)
        })
        ## test will be wrong if geno.index is too small
        ind <- which(mod$terms == "genotype")
        if (ind != geno.index) {
            mod <- logistf(model.formula, data=model.data, plconf=ind, dataout=FALSE)
        }
    
        Est <- unname(coef(mod)[ind])
        cov <- vcov(mod)[ind,ind]
        pval <- unname(mod$prob[ind])
        Stat <- qchisq(pval, df=1, lower.tail=FALSE)
        
        c(Est, sqrt(cov), Stat, pval)
    }, warning=function(w) rep(NA, 4), error=function(e) rep(NA, 4))
}

.outputNames <- function(model.type) {
    switch(model.type,
           linear=c("Est", "SE", "Wald.Stat", "Wald.Pval"),
           logistic=c("Est", "SE", "Wald.Stat", "Wald.Pval"),
           firth=c("Est", "SE", "PPL.Stat", "PPL.Pval"))
}
    
.freqByOutcome <- function(model.data, model.type, outcome) {
    if (model.type %in% c("logistic", "firth")) {
        c0 <- model.data[[outcome]] == 0
        c1 <- model.data[[outcome]] == 1
        c(n0=sum(!is.na(model.data[["genotype"]][c0])),
          n1=sum(!is.na(model.data[["genotype"]][c1])),
          freq0=0.5*mean(model.data[["genotype"]][c0], na.rm=TRUE),
          freq1=0.5*mean(model.data[["genotype"]][c1], na.rm=TRUE))
    } else {
        c(n=sum(!is.na(model.data[["genotype"]])),
          freq=0.5*mean(model.data[["genotype"]], na.rm=TRUE))
    }
}

.freqFromBinary <- function(freq) {
    freq[is.na(freq)] <- 0
    unname((freq["n0"]*freq["freq0"] + freq["n1"]*freq["freq1"]) /
           (freq["n0"] + freq["n1"]))
}

setMethod("regression",
          "SeqVarData",
          function(gdsobj, outcome, covar=NULL,
                   model.type=c("linear", "logistic", "firth"), parallel=FALSE) {
              model.type <- match.arg(model.type)
              
              ## get covariates
              dat <- pData(sampleData(gdsobj))[,c(outcome, covar),drop=FALSE]
              
              ## create model formula
              model.string <- paste(outcome, "~", paste(c(covar, "genotype"), collapse=" + "))
              
              ## for firth test - determine index of genotype in model matrix
              if (model.type == "firth") {
                  tmp <- cbind(dat, "genotype"=0)
                  geno.index <- which(colnames(model.matrix(as.formula(model.string), tmp)) == "genotype")
                  rm(tmp)
              }

              ## apply function over variants
              res <- seqApply(gdsobj, "genotype", function(x) {
              #res <- seqApply(gdsobj, c(geno="genotype", id="variant.id"), function(x) {
                  #print(x$id)
                  ## assume we want effect of reference allele
                  model.data <- cbind(dat, genotype=colSums(x == 0))
                  model.data <- droplevels(model.data[complete.cases(model.data),])
                  
                  ## don't bother with monomorphic variants
                  freq <- .freqByOutcome(model.data, model.type, outcome)
                  freq.all <- if (model.type == "linear") freq["freq"] else .freqFromBinary(freq)
                  if (freq.all %in% c(0,1)) {
                      reg <- rep(NA, 4)
                  } else {
                      if (model.type %in% c("linear", "logistic")) {
                          reg <- .runRegression(model.string, model.data, model.type)
                      } else if (model.type == "firth") {
                          reg <- .runFirth(model.string, model.data, geno.index)
                      }
                  }
                  c(freq, setNames(reg, .outputNames(model.type)))
              }, margin="by.variant", as.is="list", parallel=parallel)
              res <- do.call(rbind, res)
              data.frame(variant.id=seqGetData(gdsobj, "variant.id"), res,
                         stringsAsFactors=FALSE)
          })

library(logistf)

test_lm <- function() {
    n <- 100
    set.seed(900); outcome <- rnorm(n)
    set.seed(901); covarA <- rnorm(n)
    set.seed(902); covarB <- sample(letters[1:3], n, replace=TRUE)
    set.seed(903); genotype <- sample(0:2, n, replace=TRUE)
    dat <- data.frame(outcome, covarA, covarB, genotype)
    model.string <- "outcome ~ covarA + covarB + genotype"

    res <- SeqVarTools:::.runRegression(model.string, dat, "linear")
    exp <- lm(model.string, dat)
    checkEquals(summary(exp)$coef["genotype",1:2], res[1:2], check.names=FALSE)
}

test_glm <- function() {
    n <- 100
    set.seed(904); outcome <- rbinom(n,1,0.3)
    set.seed(905); covarA <- rnorm(n)
    set.seed(906); covarB <- sample(letters[1:3], n, replace=TRUE)
    set.seed(907); genotype <- sample(0:2, n, replace=TRUE)
    dat <- data.frame(outcome, covarA, covarB, genotype)
    model.string <- "outcome ~ covarA + covarB + genotype"

    res <- SeqVarTools:::.runRegression(model.string, dat, "logistic")
    exp <- glm(model.string, dat, family="binomial")
    checkEquals(summary(exp)$coef["genotype",1:2], res[1:2], check.names=FALSE)
}

test_firth <- function() {
    n <- 100
    set.seed(908); outcome <- rbinom(n,1,0.3)
    set.seed(909); covarA <- rnorm(n)
    set.seed(910); covarB <- sample(letters[1:3], n, replace=TRUE)
    set.seed(911); genotype <- sample(0:2, n, replace=TRUE)
    dat <- data.frame(outcome, covarA, covarB, genotype)
    model.string <- "outcome ~ covarA + covarB + genotype"

    res <- SeqVarTools:::.runFirth(model.string, dat, geno.index=5)
    exp <- logistf(as.formula(model.string), dat)
    checkEquals(coef(exp)["genotype"], res[1], check.names=FALSE)
}

.testData <- function(binary=FALSE, nv=100) {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gds <- seqOpen(seqExampleFileName("gds"))
    seqSetFilter(gds, variant.id=1:nv, verbose=FALSE)
    
    require(Biobase)
    sample.id <- seqGetData(gds, "sample.id")
    n <- length(sample.id)
    set.seed(912); outcome <- rbinom(n,1,0.3)
    set.seed(913); covarA <- rnorm(n)
    set.seed(914); covarB <- sample(letters[1:3], n, replace=TRUE)
    df <- data.frame(sample.id, outcome, covarA, covarB,
                     stringsAsFactors=FALSE)
    adf <- AnnotatedDataFrame(df)

    SeqVarData(gds, adf)
}

test_regression <- function() {
    gds <- .testData()
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"))

    dat <- cbind(pData(sampleData(gds)), genotype=refDosage(gds)[,1])
    exp <- lm("outcome ~ covarA + covarB + genotype", dat)
    seqClose(gds)
    checkEquals(summary(exp)$coef["genotype",1], res[1,"Est"], check.names=FALSE)
    checkEquals(summary(exp)$coef["genotype",2], res[1,"SE"], check.names=FALSE)
}

test_regression_firth <- function() {
    gds <- .testData(binary=TRUE)
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"),
                      model.type="firth")

    dat <- cbind(pData(sampleData(gds)), genotype=refDosage(gds)[,1])
    seqClose(gds)
    exp <- logistf(as.formula("outcome ~ covarA + covarB + genotype"), dat)
    checkEquals(coef(exp)["genotype"], res[1,"Est"], check.names=FALSE)
}

test_freq <- function() {
    gds <- .testData()
    geno <- refDosage(gds)
    freq <- alleleFrequency(gds)
    for (i in 1:ncol(geno)) {
        model.data <- cbind(pData(sampleData(gds)), genotype=geno[,i])
        fo <- SeqVarTools:::.freqByOutcome(model.data, "linear", "outcome")
        checkEquals(freq[i], fo["freq"], check.names=FALSE)
        checkEquals(sum(!is.na(geno[,i])), fo["n"], check.names=FALSE)
    }
    seqClose(gds)
}

test_freq_binary <- function() {
    gds <- .testData(binary=TRUE)
    status <- sampleData(gds)$outcome
    geno <- refDosage(gds)
    n <- lapply(c(0,1), function(c) colSums(!is.na(geno[(status == c),])))
    samp.id <- sampleData(gds)$sample.id
    freq <- lapply(c(0,1), function(c)
                   applyMethod(gds, alleleFrequency, sample=samp.id[status == c]))
    for (i in 1:ncol(geno)) {
        model.data <- cbind(pData(sampleData(gds)), genotype=geno[,i])
        fo <- SeqVarTools:::.freqByOutcome(model.data, "logistic", "outcome")
        for (c in c(0,1)) {
            checkEquals(freq[[c+1]][i], fo[paste0("freq", c)], check.names=FALSE)
            checkEquals(n[[c+1]][i], fo[paste0("n", c)], check.names=FALSE)
        }
    }
    seqClose(gds)
}

mono <- function(x) {
    (x["freq0"] == 0 & x["freq1"] == 0) | (x["freq0"] == 1 & x["freq1"] == 1)
}

test_freqFromBinary <- function() {
    gds <- .testData(binary=TRUE)
    geno <- refDosage(gds)
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"),
                      model.type="logistic")
    for (i in 1:ncol(geno)) {
        model.data <- cbind(pData(sampleData(gds)), genotype=geno[,i])
        fo <- SeqVarTools:::.freqByOutcome(model.data, "logistic", "outcome")
        fc <- SeqVarTools:::.freqByOutcome(model.data, "linear", "outcome")
        checkEquals(fc["freq"], SeqVarTools:::.freqFromBinary(fo), checkNames=FALSE)
        checkEquals(mono(fo), fc["freq"] %in% c(0,1), checkNames=FALSE)
    }
    seqClose(gds)
}

test_freq_mono <- function() {
    gds <- .testData(binary=FALSE)
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"),
                      model.type="linear")
    for (i in 1:nrow(res)) {
        if (res[i,"freq"] %in% c(0,1)) checkTrue(is.na(res[i,"Est"]))
    }
    seqClose(gds)
    
    gds <- .testData(binary=TRUE)
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"),
                      model.type="logistic")
    for (i in 1:nrow(res)) {
        if (mono(res[i,])) checkTrue(is.na(res[i,"Est"]))
    }
    seqClose(gds)
}

test_firth_badindex <- function() {
    n <- 100
    set.seed(916); outcome <- rbinom(n,1,0.3)
    set.seed(917); covarA <- rnorm(n)
    set.seed(918); covarB <- sample(letters[1:3], n, replace=TRUE)
    set.seed(919); genotype <- sample(0:2, n, replace=TRUE)
    dat <- data.frame(outcome, covarA, covarB, genotype)
    model.string <- "outcome ~ covarA + covarB + genotype"

    exp <- logistf(as.formula(model.string), dat)
    res <- SeqVarTools:::.runFirth(model.string, dat, geno.index=4) # too small
    checkEquals(coef(exp)["genotype"], res[1], check.names=FALSE)
    res <- SeqVarTools:::.runFirth(model.string, dat, geno.index=6) # too big
    checkEquals(coef(exp)["genotype"], res[1], check.names=FALSE)
}

test_droplevels <- function() {
    gds <- .testData(binary=TRUE)
    seqSetFilter(gds, sample.sel=sampleData(gds)$covarB != "c")
    res <- regression(gds, outcome="outcome", covar=c("covarA", "covarB"),
                      model.type="firth")

    dat <- droplevels(cbind(pData(sampleData(gds)), genotype=refDosage(gds)[,1]))
    seqClose(gds)
    exp <- logistf(as.formula("outcome ~ covarA + covarB + genotype"), dat)
    checkEquals(coef(exp)["genotype"], res[1,"Est"], check.names=FALSE)
}

test_nocovar <- function() {
    gds <- .testData(binary=FALSE)
    res <- regression(gds, outcome="outcome", covar=NULL, model.type="linear")
    
    dat <- cbind(pData(sampleData(gds)), genotype=refDosage(gds)[,1])
    exp <- lm("outcome ~ genotype", dat)
    seqClose(gds)
    checkEquals(summary(exp)$coef["genotype",1], res[1,"Est"], check.names=FALSE)
    checkEquals(summary(exp)$coef["genotype",2], res[1,"SE"], check.names=FALSE)
}

# this function uses a lot of subfunctions defined in duplicateDiscordance.R

# some additional subfunctions need to be defined here.
# geno1 = sequence
# geno2 = array
.truePos <- function(geno1, geno2) {
  2*(geno1 == "alt" & geno2 == "alt") +
    (geno1 == "alt" & geno2 == "het") +
    (geno1 == "het" & geno2 == "alt") +
    (geno1 == "het" & geno2 == "het")
}

.trueNeg <- function(geno1, geno2) {
  (geno1 == "het" & geno2 == "het") +
    (geno1 == "het" & geno2 == "ref") +
    (geno1 == "ref" & geno2 == "het") +
    2*(geno1 == "ref" & geno2 == "ref")
}

.falsePos <- function(geno1, geno2) {
  (geno1 == "alt" & geno2 == "het") +
    2*(geno1 == "alt" & geno2 == "ref") +
    (geno1 == "het" & geno2 == "ref")
}

.falseNeg <- function(geno1, geno2) {
  
  (geno1 == "het" & geno2 == "alt") +
    2*(geno1 == "ref" & geno2 == "alt") +
    (geno1 == "ref" & geno2 == "het")
}



# assumes filters are already set, and will reset filters
setMethod("alternateAlleleDetection",
          c("SeqVarData", "SeqVarData"),
          function(gdsobj, gdsobj2, match.samples.on=c("subject.id", "subject.id"), verbose=TRUE){
            
            # save original filters
            originalVariants1 <- seqGetData(gdsobj, "variant.id")
            originalSamples1 <- seqGetData(gdsobj, "sample.id")
            
            originalVariants2 <- seqGetData(gdsobj2, "variant.id")
            originalSamples2 <- seqGetData(gdsobj2, "sample.id")
            
            
            # match samples -- no subject matching for now
            if (verbose) message("matching samples... ", appendLF=FALSE)
            samp1 <- pData(sampleData(gdsobj))[, c("sample.id", match.samples.on[1])]
            names(samp1)[2] <- "subject.id"
            
            samp2 <- pData(sampleData(gdsobj2))[, c("sample.id", match.samples.on[2])]
            names(samp2)[2] <- "subject.id"
            
            samples <- .matchSamples(samp1, samp2)
            if (verbose) message(paste(nrow(samples), "pairs identified!"))
            
            # match variants
            if (verbose) message("matching variants... ", appendLF=FALSE)
            gr1 <- .getGRanges(gdsobj)
            gr2 <- .getGRanges(gdsobj2)
            
            overlappingVariants <- .matchVariants(gr1, gr2, allowOverlaps=FALSE)
            
            if (verbose) message(paste(nrow(overlappingVariants), "non-overlapping variant matches identified!"))
            
            # set filters for the variants -- will still need to order them properly
            seqSetFilter(gdsobj, variant.id=overlappingVariants$variant.id.1, verbose=FALSE)
            seqSetFilter(gdsobj2, variant.id=overlappingVariants$variant.id.2, verbose=FALSE)
            
            overlappingVariants$n.samples <- 0
            overlappingVariants$true.pos <- 0
            overlappingVariants$true.neg <- 0
            overlappingVariants$false.pos <- 0
            overlappingVariants$false.neg <- 0
            
            for (i in seq_along(samples$subject.id)){
              
              if (verbose & (i %% 10) == 0){
                message(paste("sample pair", i, "out of", nrow(samples)))
              }
              
              # set sample filter for this pair
              seqSetFilter(gdsobj, sample.id=samples$sample.id.1[i], verbose=FALSE)
              seqSetFilter(gdsobj2, sample.id=samples$sample.id.2[i], verbose=FALSE)
              
              # prepare genotype data frame
              dos1 <- refDosage(gdsobj)
              dos2 <- refDosage(gdsobj2)
              
              # order genotypes appropriately
              dos1 <- dos1[, as.character(overlappingVariants$variant.id.1)]
              dos2 <- dos2[, as.character(overlappingVariants$variant.id.2)]
              
              # recoded variants where ref in one = alt in other
              dos2[overlappingVariants$recode] <- 2 - dos2[overlappingVariants$recode]
              
              ## remove missing genotypes
              #sel <- !is.na(dos1) & !is.na(dos2)
              #dos1 <- dos1[sel]
              #dos2 <- dos2[sel]
              
              class1 <- .getGenotypeClass(dos1)
              class2 <- .getGenotypeClass(dos2)
              
              nonmiss <- (!is.na(dos1) & !is.na(dos2))
              overlappingVariants$n.samples <- overlappingVariants$n.samples + nonmiss
              overlappingVariants$true.pos <- overlappingVariants$true.pos + .truePos(class1, class2)
              overlappingVariants$true.neg <- overlappingVariants$true.neg + .trueNeg(class1, class2)
              overlappingVariants$false.pos <- overlappingVariants$false.pos + .falsePos(class1, class2)
              overlappingVariants$false.neg <- overlappingVariants$false.neg + .falseNeg(class1, class2)
              
            }
            
            # reset original filters
            seqSetFilter(gdsobj, sample.id=originalSamples1, variant.id=originalVariants1, verbose=FALSE)
            seqSetFilter(gdsobj2, sample.id=originalSamples2, variant.id=originalVariants2, verbose=FALSE)
            
            # return results data frame
            overlappingVariants
            
          })

library(Biobase)

test_posNeg <- function() {
  TP <- matrix(c(2,1,0,0,
                 1,1,0,0,
                 0,0,0,0,
                 0,0,0,0), nrow=4, byrow=TRUE)
  TN <- matrix(c(0,0,0,0,
                 0,1,1,0,
                 0,1,2,0,
                 0,0,0,0), nrow=4, byrow=TRUE)
  FP <- matrix(c(0,0,0,0,
                 1,0,0,0,
                 2,1,0,0,
                 0,0,0,0), nrow=4, byrow=TRUE)
  FN <- matrix(c(0,1,2,0,
                 0,0,1,0,
                 0,0,0,0,
                 0,0,0,0), nrow=4, byrow=TRUE)
  
  geno1 <- matrix(c("alt", "het", "ref", "miss"),
                  nrow=4, ncol=4, byrow=TRUE)
  geno2 <-  matrix(c("alt", "het", "ref", "miss"),
                   nrow=4, ncol=4)
  
  checkIdentical(TP, SeqVarTools:::.truePos(geno1, geno2))
  checkIdentical(TN, SeqVarTools:::.trueNeg(geno1, geno2))
  checkIdentical(FP, SeqVarTools:::.falsePos(geno1, geno2))
  checkIdentical(FN, SeqVarTools:::.falseNeg(geno1, geno2))
  
}



test_alternateAlleleDetection <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  
  filename <- seqExampleFileName("gds")
  tmpfile <- tempfile()
  file.copy(filename, tmpfile)
  
  gds1 <- seqOpen(filename)
  gds2 <- seqOpen(tmpfile)
  
  # add sample data
  samples1 <- data.frame(sample.id=seqGetData(gds1, "sample.id"), stringsAsFactors=F)
  samples1$subject.id <- samples1$sample.id
  seqData1 <- SeqVarData(gds1, sampleData=AnnotatedDataFrame(samples1))
  
  samples2 <- data.frame(sample.id=seqGetData(gds2, "sample.id"), stringsAsFactors=F)
  samples2$subject.id <- samples2$sample.id
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))
  
  seqSetFilter(seqData1)
  seqSetFilter(seqData2)
  
  seqSetFilter(seqData1, sample.id=seqGetData(seqData1, "sample.id")[1:3])
  seqSetFilter(seqData2, sample.id=seqGetData(seqData2, "sample.id")[2:4])
  
  # check filters
  filt.1 <- seqGetFilter(seqData1)
  filt.2 <- seqGetFilter(seqData2)
  
  # makes sure it runs
  res <- alternateAlleleDetection(seqData1, seqData2) 
  
  # check filters
  checkEquals(seqGetFilter(seqData1), filt.1)
  checkEquals(seqGetFilter(seqData2), filt.2)
  
  # check data
  checkTrue(all(res$false.pos == 0))
  checkTrue(all(res$false.neg == 0))
  
  samps <- intersect(seqGetData(seqData1, "sample.id"), seqGetData(seqData2, "sample.id"))
  
  seqSetFilter(seqData1, sample.id=samps, variant.id=res$variant.id.1)
  refdos <- refDosage(seqData1)
  # number of samples with nonmissing data
  checkEquals(res$n.samples, colSums(!is.na(refdos)), checkNames=FALSE)
  checkEquals(res$true.neg, colSums(refdos, na.rm=T), checkNames=FALSE)
  checkEquals(res$true.pos, colSums(2-refdos, na.rm=T), checkNames=FALSE)

  # change the mapping
  samples2$subject.id[2:1] <- samples2$subject.id[1:2]
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))
  
  
  seqSetFilter(seqData1)
  seqSetFilter(seqData2)
  
  seqSetFilter(seqData1, sample.id=samples1$sample.id[1:2], variant.id=seqGetData(seqData1, "variant.id")[1:50])
  seqSetFilter(seqData2, sample.id=samples2$sample.id[1:2], variant.id=seqGetData(seqData2, "variant.id")[25:75])
  
  res <- alternateAlleleDetection(seqData1, seqData2) 
  
  checkEquals(res$variant.id.1, intersect(seqGetData(seqData1, "variant.id"), seqGetData(seqData2, "variant.id")))
  
  # set filter to only read overlaps
  seqSetFilter(seqData1, variant.id=res$variant.id.1)
  dos <- refDosage(seqData1)
  
  class1 <- SeqVarTools:::.getGenotypeClass(dos[1, ])
  class2 <- SeqVarTools:::.getGenotypeClass(dos[2, ])
  
  checkEquals(res$true.pos, 2*SeqVarTools:::.truePos(class1, class2))
  checkEquals(res$true.neg, 2*SeqVarTools:::.trueNeg(class1, class2))
  checkEquals(res$false.pos, SeqVarTools:::.falsePos(class1, class2) + SeqVarTools:::.falsePos(class2, class1))
  checkEquals(res$false.neg, SeqVarTools:::.falseNeg(class1, class2) + SeqVarTools:::.falseNeg(class2, class1))

  seqClose(seqData1)
  seqClose(seqData2)
  
  unlink(tmpfile)
  
}

test_missingGenotypeRate <- function() {
  gds <- SeqVarTools:::.testData()
  geno <- seqGetData(gds, "genotype")
  checkIdentical(colSums(is.na(geno[1,,])) / dim(geno)[2],
                 missingGenotypeRate(gds, "by.variant"))
  checkIdentical(rowSums(is.na(geno[1,,])) / dim(geno)[3],
                 missingGenotypeRate(gds, "by.sample"))
  checkIdentical(as.character(seqGetData(gds, "variant.id")),
                 names(missingGenotypeRate(gds, "by.variant", use.names=TRUE)))
  checkIdentical(seqGetData(gds, "sample.id"),
                 names(missingGenotypeRate(gds, "by.sample", use.names=TRUE)))
  seqClose(gds)
}

test_missingGenotypeRate_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  miss.var <- missingGenotypeRate(gds, margin="by.variant")
  miss.samp <- missingGenotypeRate(gds, margin="by.sample")
  seqSetFilter(gds)
  checkIdentical(miss.var,
                 applyMethod(gds, missingGenotypeRate, variant=var.id, sample=samp.id,
                             margin="by.variant"))
  checkIdentical(miss.samp,
                 applyMethod(gds, missingGenotypeRate, variant=var.id, sample=samp.id,
                             margin="by.sample"))
  seqClose(gds)
}

test_inbreedCoeff_apply <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- 101:110
  samp.id <- seqGetData(gds, "sample.id")[6:10]
  seqSetFilter(gds, variant.id=var.id, sample.id=samp.id)
  ic.var <- inbreedCoeff(gds, margin="by.variant")
  ic.samp <- inbreedCoeff(gds, margin="by.sample")
  seqSetFilter(gds)
  checkIdentical(ic.var,
                 applyMethod(gds, inbreedCoeff, variant=var.id, sample=samp.id,
                             margin="by.variant"))
  checkIdentical(ic.samp,
                 applyMethod(gds, inbreedCoeff, variant=var.id, sample=samp.id,
                             margin="by.sample"))
  seqClose(gds)
}

setMethod("refFrac",
          "SeqVarGDSClass",
          function(gdsobj, use.names=TRUE, parallel=FALSE) {
            vars <- seqSummary(gdsobj, "annotation/format", check="none", verbose=FALSE)$ID
            if (!("AD" %in% vars)) {
              stop("annotation/format/AD must be present to compute allelic balance")
            }
            
            rf <- seqApply(gdsobj, "annotation/format/AD",
                           function(x) {x[,1] / rowSums(x)},
                           margin="by.variant", as.is="list", parallel=parallel)
            rf <- matrix(unlist(rf), ncol=length(rf),
                         dimnames=list(sample=NULL, variant=NULL))
            if (use.names) .applyNames(gdsobj, rf) else rf
          })


setMethod("refFracOverHets",
          "SeqVarGDSClass",
          function(gdsobj, FUN=mean, use.names=TRUE, parallel=FALSE) {
            vars <- seqSummary(gdsobj, "annotation/format", check="none", verbose=FALSE)$ID
            if (!("AD" %in% vars)) {
              stop("annotation/format/AD must be present to compute allelic balance")
            }

            rf <- seqApply(gdsobj,
                           c(ad="annotation/format/AD", geno="genotype"),
                           function(x) {
                             het <- x$geno[1,] != x$geno[2,]
                             FUN(x$ad[het,1,drop=FALSE] /
                                 rowSums(x$ad[het,,drop=FALSE]), na.rm=TRUE)
                           },
                           margin="by.variant", as.is="double", parallel=parallel)   
            if (use.names)
              names(rf) <- seqGetData(gdsobj, "variant.id")
            rf
          })


## code by genotype
.callType <- function(x) {
  pcol <- character(length(x))
  a <- substr(x, 1, 1)
  b <- substr(x, 3, 3)
  pcol[a == "0" | b == "0"] <- "refhet"
  pcol[a == "0" & b == "0"] <- "refhom"
  pcol[a != "0" & b != "0" & a == b] <- "althom"
  pcol[a != "0" & b != "0" & a != b] <- "althet"
  pcol[is.na(x)] <- "na"
  pcol
}


## title with variant.id, chr:position, SNV status, genotype counts
.varTitle <- function(gdsobj) {
  var.id <- seqGetData(gdsobj, "variant.id")
  loc <- paste(seqGetData(gdsobj, "chromosome"),
               seqGetData(gdsobj, "position"), sep=":")
  snv <- isSNV(gdsobj, biallelic=FALSE)
  nalleles <- nAlleles(gdsobj)
  counts <- .countGenotypes(gdsobj)
  paste0("Variant ", var.id, ", chr", loc, ", ",
        ifelse(snv, "SNV", "INDEL"), ", ",
        "nAlt=", nalleles-1, "\n",
         apply(counts, 1, function(x) paste(rev(paste(names(counts), x, sep="=")), collapse=", ")))
}


setMethod("refFracPlot",
          "SeqVarGDSClass",
          function(gdsobj, variant.id, highlight=NULL,...) {
            filt.orig <- seqGetFilter(gdsobj)$variant.sel
            seqSetFilter(gdsobj, variant.id=variant.id, verbose=FALSE)

            ## number of reference reads
            num.ref <- seqApply(gdsobj, "annotation/format/AD",
                           function(x) {x[,1]},
                           margin="by.variant", as.is="list")
            num.ref <- matrix(unlist(num.ref), ncol=length(num.ref))

            ## total read depth
            tot.reads <- seqApply(gdsobj, "annotation/format/AD",
                           rowSums,
                           margin="by.variant", as.is="list")
            tot.reads <- matrix(unlist(tot.reads), ncol=length(tot.reads))

            ## reference allele fraction
            ref.frac <- num.ref / tot.reads

            ## color-code by genotype call
            geno <- getGenotype(gdsobj)
            dimnames(ref.frac) <- dimnames(geno)
            dimnames(tot.reads) <- dimnames(geno)

            ## title with variant.id, chr:position, SNV status
            title <- .varTitle(gdsobj)

            ## plot in same order as variant.id
            for (i in match(variant.id, seqGetData(gdsobj, "variant.id"))) {            
              ## median fraction for hets
              calls <- .callType(geno[,i])
              refhet <- calls == "refhet"
              med.frac <- median(ref.frac[refhet, i], na.rm=TRUE)
              ## hets significantly different from 0.5
              pch <- rep(16, length(refhet))
              alpha <- 0.05 / sum(refhet)  # Bonferroni correction
              for (j in which(refhet)) {
                if (tot.reads[j,i] > 0) {
                  sig <- binom.test(num.ref[j,i], tot.reads[j,i], 0.5,
                                    alternative="two.sided")$p.value
                  if (sig < alpha) pch[j] <- 17
                }
              }
              ## missing genotype
              pch[calls == "na"] <- 4

              colmap <- c("refhom"="#1b9e77",
                          "refhet"="#d95f02",
                          "althom"="#7570b3",
                          "althet"="#e7298a",
                          "na"="black")
              
              plot(ref.frac[,i], tot.reads[,i],
                   col=adjustcolor(colmap[calls], alpha.f=0.5), pch=pch,
                   xlab="Fraction of reference allele reads",
                   ylab="Total reads (unfiltered)",
                   main=title[i], xlim=c(0,1), ...)
              abline(v=0.5)
              abline(v=med.frac, col=colmap["refhet"])
              
              ## highlight samples?
              if (!is.null(highlight)) {
                  text(ref.frac[highlight[[i]],i], tot.reads[highlight[[i]],i],
                       col="black")
              }
            }
            
            seqSetFilter(gdsobj, variant.sel=filt.orig, verbose=FALSE)
          })

test_applyMethod_util <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- seqGetData(gds, "variant.id") 
  checkIdentical(var.id,
                 SeqVarTools:::.applyMethod(gds, seqGetData, var.name="variant.id"))
  checkIdentical(var.id[1:5],
                 SeqVarTools:::.applyMethod(gds, seqGetData, variant=var.id[1:5], var.name="variant.id")) 
  samp.id <- seqGetData(gds, "sample.id")
  checkIdentical(samp.id,
                 SeqVarTools:::.applyMethod(gds, seqGetData, var.name="sample.id"))
  checkIdentical(samp.id[1:5],
                 SeqVarTools:::.applyMethod(gds, seqGetData, sample=samp.id[1:5], var.name="sample.id"))
  seqClose(gds)
}

test_rangesToID <- function() {
  gds <- SeqVarTools:::.testData()
  chrom <- seqGetData(gds, "chromosome")
  var.id <- seqGetData(gds, "variant.id")[chrom == 1]
  pos <- seqGetData(gds, "position")[chrom == 1]
  ranges <- GRanges(seqnames="1", IRanges(start=min(pos), end=max(pos)))
  checkIdentical(var.id, SeqVarTools:::.rangesToID(gds, ranges))
  seqClose(gds)
}
  
test_applyMethod_vector <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- seqGetData(gds, "variant.id") 
  checkIdentical(var.id[1:5],
                 applyMethod(gds, seqGetData, variant=var.id[1:5], var.name="variant.id"))
  seqClose(gds)
}

test_applyMethod_ranges <- function() {
  gds <- SeqVarTools:::.testData()
  chrom <- seqGetData(gds, "chromosome")
  var.id <- seqGetData(gds, "variant.id")[chrom == 1]
  pos <- seqGetData(gds, "position")[chrom == 1]
  ranges <- GRanges(seqnames="1", IRanges(start=min(pos), end=max(pos)))
  checkIdentical(var.id,
                 applyMethod(gds, seqGetData, variant=ranges, var.name="variant.id"))
  seqClose(gds)
}

test_applyMethod_missing <- function() {
  gds <- SeqVarTools:::.testData()
  var.id <- seqGetData(gds, "variant.id") 
  checkIdentical(var.id,
                 applyMethod(gds, seqGetData, var.name="variant.id"))
  seqClose(gds)
}

test_applyMethod_pushpop <- function() {
  gds <- SeqVarTools:::.testData()
  sample.id <- seqGetData(gds, "sample.id")
  seqSetFilter(gds, sample.id=sample.id[1:5], variant.id=1:10)
  samp.orig <- seqGetData(gds, "sample.id")
  var.orig <- seqGetData(gds, "variant.id")
  applyMethod(gds, print, sample=sample.id[6:10], variant=11:20)
  checkIdentical(samp.orig, seqGetData(gds, "sample.id"))
  checkIdentical(var.orig, seqGetData(gds, "variant.id"))
  seqClose(gds)
}

test_setVariantID <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  oldfile <- system.file("extdata", "gl_chr1.gds", package="SeqVarTools")
  newfile <- tempfile()
  file.copy(oldfile, newfile)

  # variant length
  gds <- seqOpen(newfile)
  var.len <- length(seqGetData(gds, "variant.id"))
  seqClose(gds)

  # integer id
  int.id <- seq(101, 100+var.len)
  setVariantID(newfile, int.id)
  gds <- seqOpen(newfile)
  checkIdentical(int.id, seqGetData(gds, "variant.id"))
  seqClose(gds)
 
  # character id
  char.id <- letters[1:var.len]
  setVariantID(newfile, char.id)
  gds <- seqOpen(newfile)
  checkIdentical(char.id, seqGetData(gds, "variant.id"))
  seqClose(gds)

  # non-unique
  rep.id <- rep("a", var.len)
  checkException(setVariantID(newfile, rep.id))

  # wrong length
  long.id <- 1:(var.len+1)
  checkException(setVariantID(newfile, long.id))
 
  unlink(newfile)
}

setGeneric("applyMethod",
           signature=c("gdsobj", "FUN", "variant"),
           function(gdsobj, FUN, variant=NULL, sample=NULL, ...)
             standardGeneric("applyMethod"))


## allele methods
setGeneric("refChar",
           function(gdsobj, ...)
             standardGeneric("refChar"))

setGeneric("altChar",
           function(gdsobj, ...)
             standardGeneric("altChar"))

setGeneric("nAlleles",
           function(gdsobj, ...)
             standardGeneric("nAlleles"))


## descriptive methods
setGeneric("isVariant",
           function(gdsobj, ...)
             standardGeneric("isVariant"))

setGeneric("isSNV",
           function(gdsobj, ...)
             standardGeneric("isSNV"))

setGeneric("chromWithPAR",
           function(gdsobj, ...)
             standardGeneric("chromWithPAR"))


## data retrieval and formatting
setGeneric("getGenotype",
           function(gdsobj, ...)
             standardGeneric("getGenotype"))

setGeneric("getGenotypeAlleles",
           function(gdsobj, ...)
             standardGeneric("getGenotypeAlleles"))

setGeneric("refDosage",
           function(gdsobj, ...)
             standardGeneric("refDosage"))

setGeneric("altDosage",
           function(gdsobj, ...)
             standardGeneric("altDosage"))

setGeneric("alleleDosage",
           function(gdsobj, n, ...)
             standardGeneric("alleleDosage"))

setGeneric("expandedAltDosage",
           function(gdsobj, ...)
             standardGeneric("expandedAltDosage"))

setGeneric("expandedVariantIndex",
           function(gdsobj, ...)
             standardGeneric("expandedVariantIndex"))

setGeneric("variantInfo",
           function(gdsobj, ...)
             standardGeneric("variantInfo"))

setGeneric("getVariableLengthData",
           function(gdsobj, var.name, ...)
             standardGeneric("getVariableLengthData"))

setGeneric("imputedDosage",
           function(gdsobj, ...)
             standardGeneric("imputedDosage"))


## metrics
setGeneric("titv",
           function(gdsobj, ...)
             standardGeneric("titv"))

setGeneric("alleleFrequency",
           function(gdsobj, ...)
             standardGeneric("alleleFrequency"))

setGeneric("alleleCount",
           function(gdsobj, ...)
             standardGeneric("alleleCount"))

setGeneric("minorAlleleCount",
           function(gdsobj, ...)
             standardGeneric("minorAlleleCount"))

setGeneric("missingGenotypeRate",
           function(gdsobj, ...)
             standardGeneric("missingGenotypeRate"))

setGeneric("heterozygosity",
           function(gdsobj, ...)
             standardGeneric("heterozygosity"))

setGeneric("homozygosity",
           function(gdsobj, ...)
             standardGeneric("homozygosity"))

setGeneric("hethom",
           function(gdsobj, ...)
             standardGeneric("hethom"))

setGeneric("meanBySample",
           function(gdsobj, ...)
             standardGeneric("meanBySample"))

setGeneric("countSingletons",
           function(gdsobj, ...)
             standardGeneric("countSingletons"))

setGeneric("refFrac",
           function(gdsobj, ...)
             standardGeneric("refFrac"))

setGeneric("refFracOverHets",
           function(gdsobj, ...)
             standardGeneric("refFracOverHets"))

setGeneric("refFracPlot",
           function(gdsobj, ...)
             standardGeneric("refFracPlot"))


## other
setGeneric("pca",
           function(gdsobj, ...)
             standardGeneric("pca"))

setGeneric("hwe",
           function(gdsobj, ...)
             standardGeneric("hwe"))

setGeneric("inbreedCoeff",
           function(gdsobj, ...)
             standardGeneric("inbreedCoeff"))

setGeneric("mendelErr",
           function(gdsobj, ...)
             standardGeneric("mendelErr"))

setGeneric("duplicateDiscordance",
           function(gdsobj, obj2, ...)
             standardGeneric("duplicateDiscordance"))

setGeneric("alternateAlleleDetection",
           function(gdsobj, gdsobj2, ...)
             standardGeneric("alternateAlleleDetection"))

setGeneric("regression",
           function(gdsobj, ...)
             standardGeneric("regression"))


## SeqVarData
setGeneric("sampleData",
           function(x)
             standardGeneric("sampleData"))

setGeneric("sampleData<-",
           function(x, value)
             standardGeneric("sampleData<-"))

setGeneric("variantData",
           function(x)
             standardGeneric("variantData"))

setGeneric("variantData<-",
           function(x, value)
             standardGeneric("variantData<-"))

setGeneric("validateSex",
           function(x)
             standardGeneric("validateSex"))


## Iterators
setGeneric("variantFilter",
           function(x)
               standardGeneric("variantFilter"))

setGeneric("lastFilter",
           function(x)
               standardGeneric("lastFilter"))

setGeneric("lastFilter<-",
           function(x, value)
               standardGeneric("lastFilter<-"))

setGeneric("variantBlock",
           function(x)
               standardGeneric("variantBlock"))

setGeneric("variantRanges",
           function(x)
               standardGeneric("variantRanges"))

setGeneric("iterateFilter",
           function(x, ...)
               standardGeneric("iterateFilter"))

setGeneric("resetIterator",
           function(x, ...)
               standardGeneric("resetIterator"))

setGeneric("currentRanges",
           function(x)
               standardGeneric("currentRanges"))

setGeneric("currentVariants",
           function(x)
               standardGeneric("currentVariants"))

.samplePairs1 <- function(samples) {
  stopifnot(all(c("sample.id", "subject.id") %in% names(samples)))

  ## keep only subjects with duplicate samples
  subjects <- unique(samples$subject.id[duplicated(samples$subject.id)])
  if (length(subjects) == 0) {
    stop("No duplicate samples - all values of subject.id are unique")
  }
  samples <- samples[samples$subject.id %in% subjects,]

  sample1 <- character(length(subjects))
  sample2 <- character(length(subjects))
  for (s in 1:length(subjects)) {
    samp.s <- samples$sample.id[samples$subject.id %in% subjects[s]]
    if (length(samp.s) > 2) {
      warning("More than two samples for subject ", s,
              "\nSelecting first two samples: ", samp.s[1], ", ", samp.s[2])
      samp.s <- samp.s[1:2]
    }
    sample1[s] <- samp.s[1]
    sample2[s] <- samp.s[2]
  }
  data.frame(sample1, sample2, row.names=subjects, stringsAsFactors=FALSE)
}

.samplePairs <- function(samples, all.pairs=TRUE) {
  stopifnot(all(c("sample.id", "subject.id") %in% names(samples)))

  ## keep only subjects with duplicate samples
  subjects <- unique(na.omit(samples$subject.id[duplicated(samples$subject.id)]))
  if (length(subjects) == 0) {
    stop("No duplicate samples - all values of subject.id are unique")
  }
  samples <- samples[samples$subject.id %in% subjects,]
  samples$order <- 1:nrow(samples)

  x <- merge(samples, samples, by="subject.id", suffixes = c(".1",".2"))
  x <- x[x$order.1 < x$order.2, c("subject.id", "sample.id.1", "sample.id.2")] 

  if (!all.pairs) {
      x <- x[!duplicated(x$subject.id),]
  }
  row.names(x) <- 1:nrow(x)
  return(x)
}

.genoMatch <- function(geno, check.phase) {
  allele1Match <- geno[1,1,] == geno[1,2,]
  allele2Match <- geno[2,1,] == geno[2,2,]
  if (check.phase) {
    allele1Match & allele2Match
  } else {
    allele12Match <- geno[1,1,] == geno[2,2,]
    allele21Match <- geno[2,1,] == geno[1,2,]
    (allele1Match & allele2Match) |
    (allele12Match & allele21Match)
  }
}

## setMethod("duplicateDiscordance",
##           c("SeqVarData", "missing"),
duplicateDiscordance1 <-
        function(gdsobj, match.samples.on="subject.id", check.phase=FALSE, verbose=TRUE) {
            ## samples should have columns of sample.id, subject.id
            ## find matching sample pairs for subjects (one sample per subject)
            samples <- pData(sampleData(gdsobj))
            if (!(match.samples.on %in% names(samples))) stop(sprintf("%s is not a column in sampleData", match.samples.on))
            samples <- samples[, c("sample.id", match.samples.on)]
            names(samples)[2] <- "subject.id"
            
            samp.pairs <- .samplePairs1(samples)

            ## get original sample filter
            filt.orig <- seqGetFilter(gdsobj)$sample.sel

            var.id <- seqGetData(gdsobj, "variant.id")
            nvar <- length(var.id)
            var.discord <- integer(nvar)
            var.npair <- integer(nvar)
            
            ## for each subject
            nsubj <- nrow(samp.pairs)
            subj.discord <- integer(nsubj)
            subj.nvar <- integer(nsubj)
            for (i in 1:nsubj) {
              if (verbose) message("subject ", i, " of ", nsubj)
              
              ## get genotypes for all samples
              seqSetFilter(gdsobj, sample.id=unlist(samp.pairs[i,], use.names=FALSE),
                           verbose=FALSE)
              geno <- seqGetData(gdsobj, "genotype")
              match <- .genoMatch(geno, check.phase)
              
              ## store concordance by sample in sample list
              subj.discord[i] <- sum(!match, na.rm=TRUE)
              subj.nvar[i] <- sum(!is.na(match))
              
              ## add concordance by variant to variant list
              var.npair <- var.npair + !is.na(match)
              disc <- as.integer(!match)
              disc[is.na(match)] <- 0
              var.discord <- var.discord + disc
            }

            var.df <- data.frame(num.discord=var.discord, num.pair=var.npair,
                                 discord.rate=var.discord/var.npair,
                                 row.names=var.id)

            samp.pairs$num.discord <- subj.discord
            samp.pairs$num.var <- subj.nvar
            samp.pairs$discord.rate <- subj.discord/subj.nvar
            
            ## reset original sample filter
            seqSetFilter(gdsobj, sample.sel=filt.orig, verbose=FALSE)

            list(by.variant=var.df, by.subject=samp.pairs)
            ## })
            }



#####################################################################
# definition for the signature with two SeqVarGDSClass objects
#
#

# returns number of concordant genotypes
.getMatchesConc <- function(geno1, geno2){
  sel <- (geno1 == geno2) & (geno1 != "miss")
  sel
}

# returns index selecting genotype pairs that involve the alt allele
.getAlt <- function(geno1, geno2){
  (geno1 == "het" | geno1 == "alt" | geno2 == "het" | geno2 == "alt") & (geno1 != "miss") & (geno2 != "miss")
}


# returns the number of concordant pairs involving the alt allele
.getMatchesAltConc <- function(geno1, geno2){
  sel <- .getAlt(geno1, geno2) & (geno1 == geno2) & (geno1 != "miss")
  sel
}

.getMatchesHetAlt <- function(geno1, geno2) {
  sel <- (geno1 == "alt" & geno2 == "het") |
    (geno1 == "het" & geno2 == "alt")
  sel
}

.getMatchesHetRef <- function(geno1, geno2) {
  sel <- (geno1 == "ref" & geno2 == "het") |
    (geno1 == "het" & geno2 == "ref")
  sel
}

.getMatchesRefAlt <- function(geno1, geno2){
  sel <- (geno1 == "ref" & geno2 == "alt") | (geno1 == "alt" & geno2 == "ref")
  sel
}


.getMatchesHetHom <- function(geno1, geno2){
  sel <- (geno1 %in% c("ref", "alt") & geno2 %in% c("ref", "alt")) |
    (geno1 == "het" & geno2 == "het")
  sel
}

.getNonMissing <- function(geno1, geno2){
  sel <- (geno1 != "miss") & (geno2 != "miss")
  sel
}


.matchVariants <- function(gr1, gr2, match.on=c("alleles", "position"), allowOverlaps=TRUE){
  
  match.on = match.arg(match.on)
  
  required <- c("variant.id", "ref", "alt", "snv")
  names1 <- names(elementMetadata(gr1))
  if (length(setdiff(required, names1)) > 0) stop(paste("gr1 must have metadata", paste(required, collapse=", ")))
  
  names2 <- names(elementMetadata(gr2))
  if (length(setdiff(required, names2)) > 0) stop(paste("gr2 must have metadata", paste(required, collapse=", ")))
  
  # subset to only biallelic snvs
  #sel.snv <- gr1$snv & gr2$snv
  gr1 <- gr1[gr1$snv]
  gr2 <- gr2[gr2$snv]
  
  # find overlaps
  overlaps <- findOverlaps(gr1, gr2)
  

  
  overlapping.1 <- gr1[queryHits(overlaps)]
  overlapping.2 <- gr2[subjectHits(overlaps)]
  
  # data frame to track overlaps
  overlapping <- data.frame(variant.id.1=overlapping.1$variant.id,
                            variant.id.2=overlapping.2$variant.id,
                            stringsAsFactors=F)

  # check alleles if requested
  sel.same <- (overlapping.1$ref == overlapping.2$ref) & (overlapping.1$alt == overlapping.2$alt)
  sel.flip <- (overlapping.1$ref == overlapping.2$alt) & (overlapping.1$alt == overlapping.2$ref)
  if (match.on == "alleles"){
    overlapping$recode <- sel.flip
    overlapping <- overlapping[sel.same | sel.flip, ]
  } else {
    overlapping$recode <- FALSE
  }
  
  if (!allowOverlaps){
    # choose the first overlapping variant
    overlapping <- overlapping[!duplicated(overlapping$variant.id.1), ]
    overlapping <- overlapping[!duplicated(overlapping$variant.id.2), ]
  }
  
  overlapping
  
}

.matchSamples <- function(samp1, samp2) {
  
  if (!("sample.id" %in% names(samp1)) | !("sample.id" %in% names(samp2))) stop("sample data frames must have sample.id")
  if (!("subject.id" %in% names(samp1)) | !("subject.id" %in% names(samp2))) stop("sample data frames must have subject.id")
  
  # match on subject.ids
  subjects <- intersect(samp1$subject.id, samp2$subject.id)
  
  samp1 <- samp1[samp1$subject.id %in% subjects, ]
  samp2 <- samp2[samp2$subject.id %in% subjects, ]
  
  names(samp1)[names(samp1) == "sample.id"] <- "sample.id.1"
  names(samp2)[names(samp2) == "sample.id"] <- "sample.id.2"
  
  samp <- merge(samp1, samp2)
  samp
  
}


.getGRanges <- function(gds){
  gr <- granges(gds)
  gr$variant.id <- seqGetData(gds, "variant.id")
  gr$ref <- refChar(gds)
  gr$alt <- altChar(gds)
  gr$snv <- isSNV(gds)
  
  gr
}


.getGenotypeClass <- function(x){
  
  # map for genotype classes
  class.map <- c("alt", "het", "ref")
  
  # 0 = alt/alt, 1 = het, 2 = ref/ref, so we can just subset class map by the dosage plus 1
  tmp <- class.map[x + 1]
  tmp[is.na(tmp)] <- "miss"
  
  if (is.matrix(x)) tmp <- matrix(tmp, nrow=nrow(x))
  
  tmp
  
}


# assumes filters are already set, and will reset filters
setMethod("duplicateDiscordance",
          c("SeqVarData", "SeqVarData"),
          function(gdsobj, obj2, match.samples.on=c("subject.id", "subject.id"), match.variants.on=c("alleles", "position"), discordance.type=c("genotype", "hethom"), by.variant=FALSE, verbose=TRUE){
            
            # deal with arguments
            match.variants.on = match.arg(match.variants.on)
            discordance.type = match.arg(discordance.type)
            
            # save original filters
            originalVariants1 <- seqGetData(gdsobj, "variant.id")
            originalSamples1 <- seqGetData(gdsobj, "sample.id")
            
            originalVariants2 <- seqGetData(obj2, "variant.id")
            originalSamples2 <- seqGetData(obj2, "sample.id")
            
            
            # match samples
            # construct sample data frames for matching
            if (verbose) message("matching samples... ", appendLF=FALSE)
            samp1 <- pData(sampleData(gdsobj))[, c("sample.id", match.samples.on[1])]
            names(samp1)[2] <- "subject.id"

            samp2 <- pData(sampleData(obj2))[, c("sample.id", match.samples.on[2])]
            names(samp2)[2] <- "subject.id"

            samples <- .matchSamples(samp1, samp2)
            if (verbose) message(paste(nrow(samples), "pairs identified!"))
            
            # match variants
            if (verbose) message("matching variants... ", appendLF=FALSE)
            gr1 <- .getGRanges(gdsobj)
            gr2 <- .getGRanges(obj2)
            overlappingVariants <- .matchVariants(gr1, gr2, match.on=match.variants.on)
            if (verbose) message(paste(nrow(overlappingVariants), "variant matches identified!"))
            
            # set up results data frame -- can just add columns to the samples data frame
            if (!by.variant){

              nsamp <- nrow(samples)
              n.variants <- integer(nsamp)
              n.concordant <- integer(nsamp)
              
              if (discordance.type == "genotype") {
                n.alt <- integer(nsamp)
                n.alt.conc <- integer(nsamp)
                n.het.ref <- integer(nsamp)
                n.het.alt <- integer(nsamp)
                n.ref.alt <- integer(nsamp)
              }
            } else {

              nvar <- nrow(overlappingVariants)
              n.samples <- integer(nvar)
              n.concordant <- integer(nvar)
              
              if (discordance.type == "genotype"){
                n.alt <- integer(nvar)
                n.alt.conc <- integer(nvar)
                n.het.ref <- integer(nvar)
                n.het.alt <- integer(nvar)
                n.ref.alt <- integer(nvar)
              }
            }
            
            # set filters for the variants -- will still need to order them properly
            seqSetFilter(gdsobj, variant.id=overlappingVariants$variant.id.1, verbose=FALSE)
            seqSetFilter(obj2, variant.id=overlappingVariants$variant.id.2, verbose=FALSE)
            
            for (i in seq_along(samples$subject.id)){
              
              if (verbose & (i %% 10) == 0){
                message(paste("sample pair", i, "of", nsamp))
              }
              
              # set sample filter for this pair
              seqSetFilter(gdsobj, sample.id=samples$sample.id.1[i], verbose=FALSE)
              seqSetFilter(obj2, sample.id=samples$sample.id.2[i], verbose=FALSE)
              
              # prepare genotype data frame
              dos1 <- refDosage(gdsobj)
              dos2 <- refDosage(obj2)
              
              # order genotypes appropriately
              dos1 <- dos1[, as.character(overlappingVariants$variant.id.1)]
              dos2 <- dos2[, as.character(overlappingVariants$variant.id.2)]
              
              # recode if the alt/ref alleles are switched
              dos2[overlappingVariants$recode] <- 2 - dos2[overlappingVariants$recode]
              
              # remove missing genotypes
              sel <- !is.na(dos1) & !is.na(dos2)
              #dos1 <- dos1[sel]
              #dos2 <- dos2[sel]
              
              class1 <- .getGenotypeClass(dos1)
              class2 <- .getGenotypeClass(dos2)
              
              if (!by.variant){
                
                # store information about discordant variants for this pair
                n.variants[i] <- sum(sel)
                
                if (discordance.type == "hethom"){
                  n.concordant[i] <- sum(.getMatchesHetHom(class1, class2)[sel])
                } else {
                  n.concordant[i] <- sum(.getMatchesConc(class1, class2)[sel])
                  n.alt[i] <- sum(.getAlt(class1, class2)[sel])
                  n.alt.conc[i] <- sum(.getMatchesAltConc(class1, class2)[sel])
                  n.het.ref[i] <- sum(.getMatchesHetRef(class1, class2)[sel])
                  n.het.alt[i] <- sum(.getMatchesHetAlt(class1, class2)[sel])
                  n.ref.alt[i] <- sum(.getMatchesRefAlt(class1, class2)[sel])
                }
              } else {
                
                n.samples <- n.samples + sel
                
                if (discordance.type == "hethom"){
                  n.concordant <- n.concordant + .getMatchesHetHom(class1, class2)  
                } else if (discordance.type == "genotype") {
                  n.concordant <- n.concordant + .getMatchesConc(class1, class2)
                  n.alt <- n.alt + .getAlt(class1, class2)
                  n.alt.conc <- n.alt.conc + .getMatchesAltConc(class1, class2)
                  n.het.ref <- n.het.ref + .getMatchesHetRef(class1, class2)
                  n.het.alt <- n.het.alt + .getMatchesHetAlt(class1, class2)
                  n.ref.alt <- n.ref.alt + .getMatchesRefAlt(class1, class2)
                }
              }
              
            }
            
            # reset original filters
            seqSetFilter(gdsobj, sample.id=originalSamples1, variant.id=originalVariants1, verbose=FALSE)
            seqSetFilter(obj2, sample.id=originalSamples2, variant.id=originalVariants2, verbose=FALSE)
            
            # return results data frame
            if (!by.variant){
              samples <- cbind(samples, n.variants, n.concordant)
              if (discordance.type == "genotype") {
                samples <- cbind(samples,
                                 n.alt, n.alt.conc,
                                 n.het.ref, n.het.alt, n.ref.alt)
              }
              return(samples)
            } else {
              overlappingVariants <- cbind(overlappingVariants,
                                           n.samples, n.concordant)
              if (discordance.type == "genotype") {
                overlappingVariants <- cbind(overlappingVariants,
                                             n.alt, n.alt.conc,
                                             n.het.ref, n.het.alt, n.ref.alt)
              }
              return(overlappingVariants)
            }
            
          })


setMethod("duplicateDiscordance",
          c("SeqVarData", "missing"),
          function(gdsobj,  match.samples.on="subject.id", by.variant=FALSE, all.pairs=TRUE, verbose=TRUE){
              
              ## samples should have columns of sample.id, subject.id
              ## find matching sample pairs for subjects (one sample per subject)
              samples <- pData(sampleData(gdsobj))
              if (!(match.samples.on %in% names(samples))) stop(sprintf("%s is not a column in sampleData", match.samples.on))
              samples <- samples[, c("sample.id", match.samples.on)]
              names(samples)[2] <- "subject.id"
              
              samp.pairs <- .samplePairs(samples, all.pairs=all.pairs)

              ## get original sample filter
              filt.orig <- seqGetFilter(gdsobj)$sample.sel

              nsamp <- nrow(samp.pairs)
              
              if (!by.variant) {
                  n.variants <- integer(nsamp)
                  n.concordant <- integer(nsamp)
                  n.alt <- integer(nsamp)
                  n.alt.conc <- integer(nsamp)
                  n.het.ref <- integer(nsamp)
                  n.het.alt <- integer(nsamp)
                  n.ref.alt <- integer(nsamp)
              } else {
                  variant.id <- seqGetData(gdsobj, "variant.id")
                  nvar <- length(variant.id)
                  n.samples <- integer(nvar)
                  n.concordant <- integer(nvar)
                  n.alt <- integer(nvar)
                  n.alt.conc <- integer(nvar)
                  n.het.ref <- integer(nvar)
                  n.het.alt <- integer(nvar)
                  n.ref.alt <- integer(nvar)
              }
              
              for (i in 1:nsamp) {
                  if (verbose) message("sample pair ", i, " of ", nsamp)
                  
                  ## get genotypes for all samples
                  seqSetFilter(gdsobj, sample.id=unlist(samp.pairs[i,-1], use.names=FALSE),
                               verbose=FALSE)
                  geno <- refDosage(gdsobj, use.names=FALSE)
                  dos1 <- geno[1,]
                  dos2 <- geno[2,]
                  rm(geno)
                  
                  sel <- !is.na(dos1) & !is.na(dos2)
                  
                  class1 <- .getGenotypeClass(dos1)
                  class2 <- .getGenotypeClass(dos2)
                  
                  if (!by.variant){
                      # store information about discordant variants for this pair
                      n.variants[i] <- sum(sel)
                      n.concordant[i] <- sum(.getMatchesConc(class1, class2)[sel])
                      n.alt[i] <- sum(.getAlt(class1, class2)[sel])
                      n.alt.conc[i] <- sum(.getMatchesAltConc(class1, class2)[sel])
                      n.het.ref[i] <- sum(.getMatchesHetRef(class1, class2)[sel])
                      n.het.alt[i] <- sum(.getMatchesHetAlt(class1, class2)[sel])
                      n.ref.alt[i] <- sum(.getMatchesRefAlt(class1, class2)[sel])
                  } else {
                      n.samples <- n.samples + sel
                      n.concordant <- n.concordant + .getMatchesConc(class1, class2)
                      n.alt <- n.alt + .getAlt(class1, class2)
                      n.alt.conc <- n.alt.conc + .getMatchesAltConc(class1, class2)
                      n.het.ref <- n.het.ref + .getMatchesHetRef(class1, class2)
                      n.het.alt <- n.het.alt + .getMatchesHetAlt(class1, class2)
                      n.ref.alt <- n.ref.alt + .getMatchesRefAlt(class1, class2)
                  }
              }
              
              ## reset original sample filter
              seqSetFilter(gdsobj, sample.sel=filt.orig, verbose=FALSE)

              # return results data frame
              if (!by.variant){
                  subj.df <- cbind(samp.pairs,
                                   n.variants,
                                   n.concordant,
                                   n.alt,
                                   n.alt.conc,
                                   n.het.ref,
                                   n.het.alt,
                                   n.ref.alt)
                  return(subj.df)
                  
              } else {
                  var.df <- data.frame(variant.id,
                                       n.samples,
                                       n.concordant,
                                       n.alt,
                                       n.alt.conc,
                                       n.het.ref,
                                       n.het.alt,
                                       n.ref.alt)
                  return(var.df)
              }
              
          })



setMethod("duplicateDiscordance",
          c("SeqVarIterator", "missing"),
          function(gdsobj,  match.samples.on="subject.id", by.variant=FALSE, all.pairs=TRUE, verbose=TRUE){
              
              ## samples should have columns of sample.id, subject.id
              ## find matching sample pairs for subjects (one sample per subject)
              samples <- pData(sampleData(gdsobj))
              if (!(match.samples.on %in% names(samples))) stop(sprintf("%s is not a column in sampleData", match.samples.on))
              samples <- samples[, c("sample.id", match.samples.on)]
              names(samples)[2] <- "subject.id"
              
              samp.pairs <- .samplePairs(samples, all.pairs=all.pairs)
              sample.id <- unique(c(samp.pairs$sample.id.1, samp.pairs$sample.id.2))

              ## get original sample filter
              filt.orig <- seqGetFilter(gdsobj)$sample.sel
              seqSetFilter(gdsobj, sample.id=sample.id, verbose=FALSE)
              sample.id <- seqGetData(gdsobj, "sample.id") # in case order is different

              nsamp <- nrow(samp.pairs)
              
              # results
              res <- list()
              n.iter <- length(variantFilter(gdsobj))
              b <- 1
              iterate <- TRUE
              while (iterate) {
                  ## get genotypes for all samples
                  geno <- refDosage(gdsobj, use.names=FALSE)
                  class <- .getGenotypeClass(geno)
                  rownames(class) <- sample.id
                  rm(geno)
                  
                  class1 <- class[samp.pairs$sample.id.1,,drop=FALSE]
                  class2 <- class[samp.pairs$sample.id.2,,drop=FALSE]
                  
                  if (!by.variant){
                      res[[b]] <- cbind(
                          n.variants = rowSums(.getNonMissing(class1, class2)),
                          n.concordant = rowSums(.getMatchesConc(class1, class2)),
                          n.alt = rowSums(.getAlt(class1, class2)),
                          n.alt.conc = rowSums(.getMatchesAltConc(class1, class2)),
                          n.het.ref = rowSums(.getMatchesHetRef(class1, class2)),
                          n.het.alt = rowSums(.getMatchesHetAlt(class1, class2)),
                          n.ref.alt = rowSums(.getMatchesRefAlt(class1, class2))
                      )
                  } else {
                      res[[b]] <- data.frame(
                          variant.id = seqGetData(gdsobj, "variant.id"),
                          n.samples = colSums(.getNonMissing(class1, class2)),
                          n.concordant = colSums(.getMatchesConc(class1, class2)),
                          n.alt = colSums(.getAlt(class1, class2)),
                          n.alt.conc = colSums(.getMatchesAltConc(class1, class2)),
                          n.het.ref = colSums(.getMatchesHetRef(class1, class2)),
                          n.het.alt = colSums(.getMatchesHetAlt(class1, class2)),
                          n.ref.alt = colSums(.getMatchesRefAlt(class1, class2))
                      )
                  }
                  
                  if (verbose & b %% 100 == 0) {
                      message(paste("Iteration", b , "of", n.iter, "completed"))
                  }
                  b <- b + 1
                  iterate <- iterateFilter(gdsobj, verbose=FALSE)
              }
              
              ## reset original sample filter
              seqSetFilter(gdsobj, sample.sel=filt.orig, verbose=FALSE)

              # return results data frame
              if (!by.variant){
                  ## sum over variant blocks
                  res <- Reduce(`+`, res)
                  subj.df <- cbind(samp.pairs, res)
                  return(subj.df)
                  
              } else {
                  var.df <- do.call(rbind, res)
                  return(var.df)
              }
          })

library(Biobase)
library(GenomicRanges)

test_samplePairs1 <- function() {
  samples <- data.frame(sample.id=paste("samp", 1:10, sep=""),
                        subject.id=paste("subj", c(1,1,2,2,3,3,4,4,4,5), sep=""),
                        stringsAsFactors=FALSE)
  subj.df <- data.frame(sample1=paste("samp", c(1,3,5,7), sep=""),
                        sample2=paste("samp", c(2,4,6,8), sep=""),
                        row.names=paste("subj", 1:4, sep=""),
                        stringsAsFactors=FALSE)
  checkIdentical(subj.df, SeqVarTools:::.samplePairs1(samples))
}

test_samplePairs <- function() {
  samples <- data.frame(sample.id=paste("samp", 1:10, sep=""),
                        subject.id=paste("subj", c(1,1,2,2,3,3,4,4,4,5), sep=""),
                        stringsAsFactors=FALSE)
  subj.df <- data.frame(subject.id=paste("subj", 1:4, sep=""),
                        sample.id.1=paste("samp", c(1,3,5,7), sep=""),
                        sample.id.2=paste("samp", c(2,4,6,8), sep=""),
                        stringsAsFactors=FALSE)
  sp <- SeqVarTools:::.samplePairs(samples, all.pairs=FALSE)
  checkIdentical(subj.df, sp)
  sp1 <- suppressWarnings(SeqVarTools:::.samplePairs1(samples))
  checkIdentical(row.names(sp1), sp$subject.id)
  checkIdentical(sp1$sample1, sp$sample.id.1)
  checkIdentical(sp1$sample2, sp$sample.id.2)

  subj.df <- data.frame(subject.id=paste("subj", c(1:4,4,4), sep=""),
                        sample.id.1=paste("samp", c(1,3,5,7,7,8), sep=""),
                        sample.id.2=paste("samp", c(2,4,6,8,9,9), sep=""),
                        stringsAsFactors=FALSE)
  sp <- SeqVarTools:::.samplePairs(samples, all.pairs=TRUE)
  checkIdentical(subj.df, sp)
}

test_genoMatch <- function() {
  geno <- array(dim=c(2,2,5))
  geno[1,1,] <- c(0,0,0,0,0)
  geno[2,1,] <- c(1,1,1,1,1)
  geno[1,2,] <- c(0,1,0,1,NA)
  geno[2,2,] <- c(1,0,0,1,NA)
  match.phase <- c(TRUE,FALSE,FALSE,FALSE,NA)
  match.unphase <- c(TRUE,TRUE,FALSE,FALSE,NA)
  checkIdentical(match.phase, SeqVarTools:::.genoMatch(geno, check.phase=TRUE))
  checkIdentical(match.unphase, SeqVarTools:::.genoMatch(geno, check.phase=FALSE))
}

## test_duplicateDiscordance_apply <- function() {
##   gds <- SeqVarTools:::.testData()
##   sample.id <- seqGetData(gds, "sample.id")
##   samples <- data.frame(subject.id=c(rep(c("subj1", "subj2"), each=2), sample.id[5:length(sample.id)]),
##                         sample.id=sample.id,
##                         stringsAsFactors=FALSE)
##   var.id <- 101:110
  
##   seqData <- SeqVarData(gds, sampleData=AnnotatedDataFrame(samples))
  
##   seqSetFilter(gds, variant.id=var.id, sample.id=sample.id[1:4])
  
##   disc <- duplicateDiscordance(seqData)
##   seqSetFilter(gds)
##   checkIdentical(disc,
##                  applyMethod(seqData, duplicateDiscordance,
##                              variant=var.id))
##   seqClose(gds)
## }



test_duplicateDiscordance <- function() {
  gds <- SeqVarTools:::.testData()
  sample.id <- seqGetData(gds, "sample.id")
  samples <- data.frame(subject.id=c(rep(c("subj1", "subj2"), each=2), sample.id[5:length(sample.id)]),
                        sample.id=sample.id,
                        stringsAsFactors=FALSE)
  var.id <- 101:110
  
  seqData <- SeqVarData(gds, sampleData=AnnotatedDataFrame(samples))
  
  seqSetFilter(gds, variant.id=var.id, sample.id=sample.id[1:4], verbose=FALSE)
  disc1 <- SeqVarTools:::duplicateDiscordance1(seqData, verbose=FALSE)
  disc.subj <- duplicateDiscordance(seqData, by.variant=FALSE, all.pairs=FALSE, verbose=FALSE)
  disc.var <- duplicateDiscordance(seqData, by.variant=TRUE, all.pairs=FALSE, verbose=FALSE)

  checkEquals(disc.subj$subject.id, row.names(disc1$by.subject))
  checkEquals(disc.subj$sample.id.1, disc1$by.subject$sample1)
  checkEquals(disc.subj$sample.id.2, disc1$by.subject$sample2)
  checkEquals(disc.subj$n.variants, disc1$by.subject$num.var)
  checkEquals(disc.subj$n.concordant, disc1$by.subject$num.var - disc1$by.subject$num.discord)

  checkEquals(disc.var$variant.id, as.integer(rownames(disc1$by.variant)))
  checkEquals(disc.var$n.samples, disc1$by.variant$num.pair)
  checkEquals(disc.var$n.concordant, disc1$by.variant$num.pair - disc1$by.variant$num.discord)
  
  seqClose(gds)
}


test_duplicateDiscordance_iterator <- function() {
  gds <- SeqVarTools:::.testData()
  sample.id <- seqGetData(gds, "sample.id")
  samples <- data.frame(subject.id=c(rep(c("subj1", "subj2"), each=3), sample.id[7:length(sample.id)]),
                        sample.id=sample.id,
                        stringsAsFactors=FALSE)
  
  seqData <- SeqVarData(gds, sampleData=AnnotatedDataFrame(samples))
  disc.subj <- duplicateDiscordance(seqData, by.variant=FALSE, all.pairs=TRUE, verbose=FALSE)
  disc.var <- duplicateDiscordance(seqData, by.variant=TRUE, all.pairs=TRUE, verbose=FALSE)

  it <- SeqVarBlockIterator(seqData, variantBlock=500, verbose=FALSE)
  disc.subj.it <- duplicateDiscordance(it, by.variant=FALSE, all.pairs=TRUE, verbose=FALSE)
  resetIterator(it, verbose=FALSE)
  disc.var.it <- duplicateDiscordance(it, by.variant=TRUE, all.pairs=TRUE, verbose=FALSE)

  checkEquals(disc.subj, disc.subj.it)
  checkEquals(disc.var, disc.var.it)

  seqClose(gds)
}



## tests for duplicate discordance on two gds files

.getTestGeno <- function() {
  vals <- c("ref", "het", "alt", "miss")
  df <- expand.grid(geno1=vals, geno2=vals, stringsAsFactors=F)
  
  #n.random <- 20
  #df2 <- data.frame(geno1=sample(vals, n.random, replace=T), geno2=sample(vals, n.random, replace=T), stringsAsFactors=F)
  #df <- rbind(df, df2)
  df  
}

test_getMatchesConc <- function(){
  df <- .getTestGeno()
  truth <- c(T, F, F, F, F, T, F, F, F, F, T, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesConc(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesConc(df$geno1, df$geno2), SeqVarTools:::.getMatchesConc(df$geno2, df$geno1))
}


test_getAlt <- function(){
  df <- .getTestGeno()
  truth <- c(F, T, T, F, T, T, T, F, T, T, T, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getAlt(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getAlt(df$geno1, df$geno2), SeqVarTools:::.getAlt(df$geno2, df$geno1))  
}


test_getMatchesAltConc <- function(){
  df <- .getTestGeno()
  truth <- c(F, F, F, F, F, T, F, F, F, F, T, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesAltConc(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesAltConc(df$geno1, df$geno2), SeqVarTools:::.getMatchesAltConc(df$geno2, df$geno1))
}


test_getMatchesHetRef <- function(){
  df <- .getTestGeno()
  truth <- c(F, T, F, F, T, F, F, F, F, F, F, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesHetRef(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesHetRef(df$geno1, df$geno2), SeqVarTools:::.getMatchesHetRef(df$geno2, df$geno1))
}

test_getMatchesHetAlt <- function(){
  df <- .getTestGeno()
  truth <- c(F, F, F, F, F, F, T, F, F, T, F, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesHetAlt(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesHetAlt(df$geno1, df$geno2), SeqVarTools:::.getMatchesHetAlt(df$geno2, df$geno1))
}


test_getMatchesRefAlt <- function() {
  df <- .getTestGeno()
  truth <- c(F, F, T, F, F, F, F, F, T, F, F, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesRefAlt(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesRefAlt(df$geno1, df$geno2), SeqVarTools:::.getMatchesRefAlt(df$geno2, df$geno1))
}

test_getMatchesHetHom <- function (){
  df <- .getTestGeno()
  truth <- c(T, F, T, F, F, T, F, F, T, F, T, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getMatchesHetHom(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getMatchesHetHom(df$geno1, df$geno2), SeqVarTools:::.getMatchesHetHom(df$geno2, df$geno1))
}

test_getNonMissing <- function (){
  df <- .getTestGeno()
  truth <- c(T, T, T, F, T, T, T, F, T, T, T, F, F, F, F, F)
  checkEquals(SeqVarTools:::.getNonMissing(df$geno1, df$geno2), truth)
  checkIdentical(SeqVarTools:::.getNonMissing(df$geno1, df$geno2), SeqVarTools:::.getNonMissing(df$geno2, df$geno1))
}

test_getGentoypeClass <- function() {
  vals <- c(NA, 0, 1, 2)
  class <- SeqVarTools:::.getGenotypeClass(vals)
  checkEquals(class, c("miss", "alt", "het", "ref"))

}

test_matchVariants <- function() {
  
  gr1 <- GRanges(seqnames=c("1", "1", "1", "1", "2", "2", "4"),
                 ranges=IRanges(start=c(1, 1, 2, 3, 1, 3, 1),
                                end  =c(1, 2, 2, 3, 1, 3, 1))
  )
  gr1$ref <- rep("A", length(gr1))
  gr1$alt <- rep("B", length(gr1))
  gr1$alt[2] <- "BC"
  gr1$snv <- rep(TRUE, length(gr1))
  gr1$snv[2] <- FALSE
  gr1$variant.id <- 1:length(gr1) + 2
  
  gr2 <- GRanges(seqnames=c("1", "1", "1", "1", "1", "2", "2", "3", "4"),
                 ranges=IRanges(start=c(1, 1, 1, 2, 2, 1, 3, 1, 1),
                                end = c(1, 1, 2, 2, 2, 1, 3, 1, 1)))
  gr2$ref <- rep("A", length(gr2))
  gr2$alt <- rep("B", length(gr2))
  gr2$alt[3] <- "BC"
  gr2$alt[4] <- "C"
  gr2$ref[9] <- "B"
  gr2$alt[9] <- "A"
  gr2$snv <- rep(TRUE, length(gr2))
  gr2$snv[3] <- FALSE
  gr2$variant.id <- 1:length(gr2)
  
  # match on position only, duplicates alowed
  overlaps <- SeqVarTools:::.matchVariants(gr1, gr2, match.on="position")  
  checkEquals(overlaps$variant.id.1, c(3, 3, 5, 5, 7, 8, 9))
  checkEquals(overlaps$variant.id.2, c(1, 2, 4, 5, 6, 7, 9))

  # match on position, no duplicates
  overlapsNoDups <- SeqVarTools:::.matchVariants(gr1, gr2, match.on="position", allowOverlaps=FALSE)
  checkEquals(overlapsNoDups$variant.id.1, c(3, 5, 7, 8, 9))
  checkEquals(overlapsNoDups$variant.id.2, c(1, 4, 6, 7, 9))
  
  # match on position and alleles  
  overlaps <- SeqVarTools:::.matchVariants(gr1, gr2, match.on="alleles")
  checkEquals(overlaps$variant.id.1, c(3, 3, 5, 7, 8, 9))
  checkEquals(overlaps$variant.id.2, c(1, 2, 5, 6, 7, 9))
  checkEquals(overlaps$recode, c(rep(FALSE, 5), TRUE))

  overlapsNoDups <- SeqVarTools:::.matchVariants(gr1, gr2, match.on="alleles", allowOverlaps=FALSE)
  checkEquals(overlapsNoDups$variant.id.1, c(3, 5, 7, 8, 9))
  checkEquals(overlapsNoDups$variant.id.2, c(1, 5, 6, 7, 9))
}

test_matchSamples <- function() {
  
  # no duplicates at this point
  samp1 <- data.frame(subject.id=c("A", "A", "B", "C", "D"), stringsAsFactors=F)
  samp1$sample.id <- letters[1:nrow(samp1)]                    
  
  samp2 <- data.frame(subject.id=c("A", "B", "D", "D", "E"), stringsAsFactors=F)
  samp2$sample.id <- letters[1:nrow(samp2)]
  
  chk <- SeqVarTools:::.matchSamples(samp1, samp2)
  chk$chk <- paste(chk$subject.id, chk$sample.id.1, chk$sample.id.2)
  checkTrue(setequal(chk$chk, c("A a a", "A b a", "B c b", "D e c", "D e d")))
  checkTrue(setequal(chk$subject.id, intersect(samp1$subject.id, samp2$subject.id)))
  checkEquals(nrow(chk), 5)
  
}

test_duplicateDiscordanceAcrossDatasets <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  
  filename <- seqExampleFileName("gds")
  tmpfile <- tempfile()
  file.copy(filename, tmpfile)
  
  gds1 <- seqOpen(filename)
  gds2 <- seqOpen(tmpfile)
  
  # add sample data
  samples1 <- data.frame(sample.id=seqGetData(gds1, "sample.id"), stringsAsFactors=F)
  samples1$subject.id <- samples1$sample.id
  seqData1 <- SeqVarData(gds1, sampleData=AnnotatedDataFrame(samples1))
  
  samples2 <- data.frame(sample.id=seqGetData(gds2, "sample.id"), stringsAsFactors=F)
  samples2$subject.id <- samples2$sample.id
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))

  seqSetFilter(seqData1)
  seqSetFilter(seqData2)
  
  seqSetFilter(seqData1, sample.id=seqGetData(seqData1, "sample.id")[1:3])
  seqSetFilter(seqData2, sample.id=seqGetData(seqData2, "sample.id")[2:4])
  
  # check filters
  filt.1 <- seqGetFilter(seqData1)
  filt.2 <- seqGetFilter(seqData2)
  
  # makes sure it runs
  res <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles")
  res.var <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", by.variant=TRUE)
  
  # check filters
  checkEquals(seqGetFilter(seqData1), filt.1)
  checkEquals(seqGetFilter(seqData2), filt.2)
  
  # check data for samples
  checkEquals(seqGetData(seqData1, "sample.id")[2:3], res$sample.id.1)
  checkEquals(seqGetData(seqData2, "sample.id")[1:2], res$sample.id.2)
  # and for variants
  checkEquals(seqGetData(seqData1, "variant.id")[isSNV(seqData1)], res.var$variant.id.1)
  checkEquals(seqGetData(seqData2, "variant.id")[isSNV(seqData2)], res.var$variant.id.2)
  
  checkEquals(res$n.concordant, res$n.variants)
  checkEquals(res$n.alt, res$n.alt.conc)

  checkEquals(res.var$n.concordant, res.var$n.samples)
  checkEquals(res.var$n.alt, res.var$n.alt.conc)
  
  seqSetFilter(seqData1)
  seqSetFilter(seqData2)

    # change subjectID for seqData2
  #samples2 <- sampleData(seqData2)
  samples2$subject.id[2:1] <- samples2$subject.id[1:2]
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))
  
  
  seqSetFilter(seqData1, sample.id=samples1$sample.id[1:2], variant.id=seqGetData(seqData1, "variant.id")[1:50])
  seqSetFilter(seqData2, sample.id=samples2$sample.id[1:2], variant.id=seqGetData(seqData2, "variant.id")[25:75])
  
  res <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles") 
  res.var <- duplicateDiscordance(seqData1, seqData2, by.variant=TRUE, match.variants.on="alleles") 
  
  # set filter to only read overlaps
  seqSetFilter(seqData1, variant.id=intersect(seqGetData(seqData1, "variant.id"), seqGetData(gds2, "variant.id")))
  
  # read dosages
  dos <- refDosage(seqData1)
  
  # get the granges object
  gr <- SeqVarTools:::.getGRanges(seqData1)
  
  # snvs only
  dos <- dos[, gr$snv]
  
  # non missing
  nonmiss <- !is.na(colSums(dos))
  checkEquals(res$n.variants[1], sum(nonmiss))
  
  dos <- dos[, nonmiss]
  match <- dos[1 ,] == dos[2, ]
  checkEquals(res$n.concordant[1], sum(match))
  
  # minor allele matches
  minor <- dos[1, ] < 2 | dos[2, ] < 2
  
  checkEquals(res$n.alt.conc[1], sum(minor & match))
  
  seqClose(seqData1)
  seqClose(seqData2)
  
  unlink(tmpfile)
  
}



test_duplicateDiscordanceAcrossDatasets_hethom <- function() {
  gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
  
  filename <- seqExampleFileName("gds")
  tmpfile <- tempfile()
  file.copy(filename, tmpfile)
  
  gds1 <- seqOpen(filename)
  gds2 <- seqOpen(tmpfile)
  
  # add sample data
  samples1 <- data.frame(sample.id=seqGetData(gds1, "sample.id"), stringsAsFactors=F)
  samples1$subject.id <- samples1$sample.id
  seqData1 <- SeqVarData(gds1, sampleData=AnnotatedDataFrame(samples1))
  
  samples2 <- data.frame(sample.id=seqGetData(gds2, "sample.id"), stringsAsFactors=F)
  samples2$subject.id <- samples2$sample.id
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))
  
  seqSetFilter(seqData1)
  seqSetFilter(seqData2)
  
  seqSetFilter(seqData1, sample.id=seqGetData(seqData1, "sample.id")[1:3])
  seqSetFilter(seqData2, sample.id=seqGetData(seqData2, "sample.id")[2:4])
  
  # check filters
  filt.1 <- seqGetFilter(seqData1)
  filt.2 <- seqGetFilter(seqData2)
  
  # makes sure it runs
  res <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", discordance.type = "hethom")
  tmp <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", discordance.type = "genotype")
  checkEquals(res$n.variants, tmp$n.variants)
  checkEquals(res$n.concordant, tmp$n.concordant)
  
  res.var <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", by.variant=TRUE, discordance.type="hethom")
  tmp <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", by.variant=TRUE)
  checkEquals(res.var$n.samples, tmp$n.samples)
  checkEquals(res.var$n.concordant, tmp$n.concordant)
  
  # check filters
  checkEquals(seqGetFilter(seqData1), filt.1)
  checkEquals(seqGetFilter(seqData2), filt.2)
  
  # check data for samples
  checkEquals(seqGetData(seqData1, "sample.id")[2:3], res$sample.id.1)
  checkEquals(seqGetData(seqData2, "sample.id")[1:2], res$sample.id.2)
  # and for variants
  checkEquals(seqGetData(seqData1, "variant.id")[isSNV(seqData1)], res.var$variant.id.1)
  checkEquals(seqGetData(seqData2, "variant.id")[isSNV(seqData2)], res.var$variant.id.2)
  
  checkEquals(res$n.concordant, res$n.variants)
  checkEquals(res$n.alt, res$n.alt.conc)
  
  checkEquals(res.var$n.concordant, res.var$n.samples)
  
  seqSetFilter(seqData1)
  seqSetFilter(seqData2)
  
  # change subjectID for seqData2
  #samples2 <- sampleData(seqData2)
  samples2$subject.id[2:1] <- samples2$subject.id[1:2]
  seqData2 <- SeqVarData(gds2, sampleData=AnnotatedDataFrame(samples2))
  
  
  seqSetFilter(seqData1, sample.id=samples1$sample.id[1:2], variant.id=seqGetData(seqData1, "variant.id")[1:50])
  seqSetFilter(seqData2, sample.id=samples2$sample.id[1:2], variant.id=seqGetData(seqData2, "variant.id")[25:75])
  
  res <- duplicateDiscordance(seqData1, seqData2, match.variants.on="alleles", discordance.type="hethom") 
  res.var <- duplicateDiscordance(seqData1, seqData2, by.variant=TRUE, match.variants.on="alleles", discordance.type="hethom") 
  
  # set filter to only read overlaps
  seqSetFilter(seqData1, variant.id=intersect(seqGetData(seqData1, "variant.id"), seqGetData(gds2, "variant.id")))
  
  # read dosages
  dos <- refDosage(seqData1)
  
  # get the granges object
  gr <- SeqVarTools:::.getGRanges(seqData1)
  
  # snvs only
  dos <- dos[, gr$snv]
  
  # non missing
  nonmiss <- !is.na(colSums(dos))
  checkEquals(res$n.variants[1], sum(nonmiss))
  
  dos <- dos[, nonmiss]
  dos <- abs(dos - 1)
  match <- dos[1 ,] == dos[2, ]
  checkEquals(res$n.concordant[1], sum(match))
  
  seqClose(seqData1)
  seqClose(seqData2)
  
  unlink(tmpfile)
  
}

.f <- function(counts) {
    nhet <- counts$nAa
    ntot <- rowSums(counts)
    afreq <- (2*counts$nAA + counts$nAa) / (2*ntot)
    exp.het <- 2*afreq*(1-afreq)*ntot
    1 - (nhet/exp.het)
}

setMethod("hwe",
          "SeqVarGDSClass",
          function(gdsobj, permute=FALSE, parallel=FALSE) {
              counts <- .countGenotypes(gdsobj, permute, parallel=parallel)

              afreq <- (2*counts$nAA + counts$nAa) / (2*rowSums(counts))
              p <- HWExact(counts)
              f <- .f(counts)

              ## set non-biallelic or monomorphic variants to NA
              sel <- nAlleles(gdsobj) != 2L |
                  counts$nAa + counts$naa == 0L |
                      counts$nAa + counts$nAA == 0L
              p[sel] <- NA
              f[sel] <- NA

              variant.id <- seqGetData(gdsobj, "variant.id")
              cbind(variant.id, counts, afreq, p, f)
          })

test_altDosage_sparse <- function() {
  gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
  gds <- seqOpen(gdsfile)
  d1 <- altDosage(gds, sparse=FALSE)
  d2 <- altDosage(gds, sparse=TRUE)
  checkTrue(is.matrix(d1))
  checkTrue(is(d2, "dgCMatrix"))
  checkTrue(all(is.na(d1) == is.na(d2)))
  checkTrue(all(is.na(d2) | d1 == d2))
  checkEquals(dimnames(d1), dimnames(d2))
  seqClose(gds)
}

test_expandedAltDosage_sparse <- function() {
  gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
  gds <- seqOpen(gdsfile)
  d1 <- expandedAltDosage(gds, sparse=FALSE)
  d2 <- expandedAltDosage(gds, sparse=TRUE)
  checkTrue(is.matrix(d1))
  checkTrue(is(d2, "dgCMatrix"))
  checkTrue(all(is.na(d1) == is.na(d2)))
  checkTrue(all(is.na(d2) | d1 == d2))
  checkEquals(dimnames(d1), dimnames(d2))
  seqClose(gds)
}

test_refFrac_excep <- function() {
    ## no AD field
    gds <- seqOpen(seqExampleFileName("gds"))
    checkException(refFrac(gds))
    seqClose(gds)
}

test_refFrac <- function() {
    gdsfmt::showfile.gds(closeall=TRUE, verbose=FALSE)
    gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
    gds <- seqOpen(gdsfile)
    seqSetFilter(gds, variant.sel=isSNV(gds))
    ad <- getVariableLengthData(gds, "annotation/format/AD")
    rf <- ad[1,,] / (ad[1,,] + ad[2,,])
    checkEquals(rf, refFrac(gds))

    geno <- seqGetData(gds, "genotype")
    het <- geno[1,,] != geno[2,,]
    rf[!het | is.na(het)] <- NA
    checkEquals(colMeans(rf, na.rm=TRUE), refFracOverHets(gds), checkNames=FALSE)

    ## compare with allelic balance from VCF
    ab <- getVariableLengthData(gds, "annotation/format/AB")
    rf[is.na(ab)] <- NA
    checkEquals(rf, ab, tolerance=0.01)
    
    seqClose(gds)
}

---
title: "Using Iterators in SeqVarTools"
author: "Stephanie M. Gogarten"
date: "`r Sys.Date()`"
output: 
    BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Iterators in SeqVarTools}
  %\VignetteDepends{SeqVarTools, GenomicRanges}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Using Iterators in SeqVarTools

## Introduction

Iterators can be used to apply a user function to a `SeqVarData` object. Creating the iterator defines the sets of variants returned on every subsequent call to `iterateFilter()`. `iterateFilter()` returns `TRUE` if there are more variants remaining, and `FALSE` when all variants have been returned.

## Block iterators

The simplest type of iterator, a `SeqVarBlockIterator`, returns variants in consecutive blocks.

```{r}
library(SeqVarTools)
gds <- seqOpen(seqExampleFileName("gds"))
seqData <- SeqVarData(gds)
iterator <- SeqVarBlockIterator(seqData, variantBlock=500)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)
```

A filter can be applied before the iterator is created, and only variants included in the filter will be returned by the iterator.

```{r}
seqSetFilter(seqData, variant.sel=1:100)
iterator <- SeqVarBlockIterator(seqData, variantBlock=500)
var.info <- variantInfo(iterator)
nrow(var.info)
iterateFilter(iterator)
seqResetFilter(seqData)
```

## Range iterators

A `GRanges` object can be used to create a `SeqVarRangeIterator`, where every iteration returns the next range.

```{r}
library(GenomicRanges)
gr <- GRanges(seqnames=rep(1,3), 
              ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
iterator <- SeqVarRangeIterator(seqData, variantRanges=gr)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)
```

## Window iterators

Window iterators (`SeqVarWindowIterator`) are a special class of range iterators. When the object is created, the ranges are generated automatically with a specified width and step size, covering the entire genome.

```{r}
seqSetFilterChrom(seqData, include="22")
iterator <- SeqVarWindowIterator(seqData, windowSize=10000, 
                                 windowShift=5000)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
seqResetFilter(seqData)
```

## List iterators

A `SeqVarListIterator` can be used to specify particular variants to include in each iteration. The input is a `GRangesList`, and each list element defines an iteration set.

```{r}
gr <- GRangesList(
  GRanges(seqnames=rep(22,2), 
          ranges=IRanges(start=c(16e6, 17e6), width=1e6)),
  GRanges(seqnames=rep(22,2), 
          ranges=IRanges(start=c(18e6, 20e6), width=1e6)))
iterator <- SeqVarListIterator(seqData, variantRanges=gr)
var.info <- list(variantInfo(iterator))
i <- 2
while(iterateFilter(iterator)) {
    var.info[[i]] <- variantInfo(iterator)
    i <- i + 1
}
lapply(var.info, head)
```

After the last iteration, any methods used on the iterator object will return 0 variants. The `resetIterator()` method can be used to reset an iterator back to the beginning.

```{r}
variantInfo(iterator)
resetIterator(iterator)
variantInfo(iterator)
```

```{r}
seqClose(gds)
```

---
title: "Introduction to SeqVarTools"
author: "Stephanie M. Gogarten"
date: "`r Sys.Date()`"
output: 
    BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{Introduction to SeqVarTools}
  %\VignetteDepends{SeqVarTools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

`SeqArray` provides an alternative to the Variant Call Format (VCF) for storage of variants called from sequencing data, enabling efficient storage, fast access to subsets of the data, and rapid computation.

`SeqVarTools` provides an interface to the `SeqArray` storage format with tools for many common tasks in variant analysis.

It is highly recommended to read the vignette in `SeqArray` in addition to this document to understand the data structure and full array of features.

## Converting a Variant Call Format (VCF) file

To work with `SeqVarTools`, we must first convert a VCF file into the `SeqArray` GDS format. All information in the VCF file is preserved in the resulting GDS file.

```{r}
library(SeqVarTools)
vcffile <- seqExampleFileName("vcf")
gdsfile <- "tmp.gds"
seqVCF2GDS(vcffile, gdsfile, verbose=FALSE)
gds <- seqOpen(gdsfile)
gds
```

We can look at some basic information in this file, such as the reference and alternate alleles.

```{r}
head(refChar(gds))
head(altChar(gds))
```

How many alleles are there for each variant?

```{r}
table(nAlleles(gds))
```

Two variants have 3 alleles (1 REF and 2 ALT). We can extract the second alternate allele for these variants by using the argument `n=2` to `altChar()`.

```{r}
multi.allelic <- which(nAlleles(gds) > 2)
altChar(gds)[multi.allelic]
altChar(gds, n=1)[multi.allelic]
altChar(gds, n=2)[multi.allelic]
```

These two sites have three alleles, two are each single nucleotides and the third is a dinucleotide, representing an indel.

```{r}
table(isSNV(gds))
isSNV(gds)[multi.allelic]
```

Chromosome and position can be accessed as vectors or as a `GRanges` object.

```{r}
head(seqGetData(gds, "chromosome"))
head(seqGetData(gds, "position"))
granges(gds)
```

We can also find the sample and variant IDs.

```{r}
head(seqGetData(gds, "sample.id"))
head(seqGetData(gds, "variant.id"))
```

The variant IDs are sequential integers created by `seqVCF2GDS`. We may wish to rename them to something more useful. Note the `annotation/` prefix required to retrieve the `id` variable. We need to confirm that the new IDs are unique (which is not always the case for the `annotation/id` field).

```{r}
rsID <- seqGetData(gds, "annotation/id")
head(rsID)
length(unique(rsID)) == length(rsID)
```

Renaming the variant IDs requires modifying the GDS file, so we have to close it first.

```{r}
seqClose(gds)
setVariantID(gdsfile, rsID)
gds <- seqOpen(gdsfile)
head(seqGetData(gds, "variant.id"))
```

Note that using character strings for `variant.id` instead of integers may decrease performance for large datasets.

`getGenotype()` transforms the genotypes from the internal storage format to VCF-like character strings.

```{r}
geno <- getGenotype(gds)
dim(geno)
geno[1:10, 1:5]
```

`getGenotypeAlleles()` returns the nucleotides instead of integers.

```{r}
geno <- getGenotypeAlleles(gds)
geno[1:10, 1:5]
```

## Applying methods to subsets of data

If a dataset is large, we may want to work with subsets of the data at one time. We can use `applyMethod()` to select a subset of variants and/or samples. `applyMethod()` is essentially a wrapper for `seqSetFilter` that enables us to apply a method or function to a data subset in one line. If it is desired to use the same filter multiple times, it may be more efficient to set the filter once instead of using `applyMethod()`.

```{r}
samp.id <- seqGetData(gds, "sample.id")[1:10]
var.id <- seqGetData(gds, "variant.id")[1:5]
applyMethod(gds, getGenotype, variant=var.id, sample=samp.id)
```

As an alternative to specifying variant ids, we can use a `GRanges` object to select a range on chromosome 22.

```{r}
library(GenomicRanges)
gr <- GRanges(seqnames="22", IRanges(start=1, end=250000000))
geno <- applyMethod(gds, getGenotype, variant=gr)
dim(geno)
```

## Examples

### Transition/transversion ratio

The transition/transversion ratio (TiTv) is frequently used as a quality metric. We can calculate TiTv over the entire dataset or by sample.

```{r}
titv(gds)
head(titv(gds, by.sample=TRUE))
```

Alternatively, we can plot TiTv binned by various metrics (allele frequency, missing rate, depth) to assess variant quality. We need the ids of the variants that fall in each bin.

```{r}
binVar <- function(var, names, breaks) {
  names(var) <- names
  var <- sort(var)
  mids <- breaks[1:length(breaks)-1] + 
    (breaks[2:length(breaks)] - breaks[1:length(breaks)-1])/2
  bins <- cut(var, breaks, labels=mids, right=FALSE)
  split(names(var), bins)
}
```

```{r}
variant.id <- seqGetData(gds, "variant.id")
afreq <- alleleFrequency(gds)
maf <- pmin(afreq, 1-afreq)
maf.bins <- binVar(maf, variant.id, seq(0,0.5,0.02))
nbins <- length(maf.bins)
titv.maf <- rep(NA, nbins)
for (i in 1:nbins) {
  capture.output(titv.maf[i] <- applyMethod(gds, titv, variant=maf.bins[[i]]))
}
plot(as.numeric(names(maf.bins)), titv.maf, xlab="MAF", ylab="TiTv")
```

```{r}
miss <- missingGenotypeRate(gds)
miss.bins <- binVar(miss, variant.id, c(seq(0,0.5,0.05),1))
nbins <- length(miss.bins)
titv.miss <- rep(NA, nbins)
for (i in 1:nbins) {
  capture.output(titv.miss[i] <- applyMethod(gds, titv, variant=miss.bins[[i]]))
}
plot(as.numeric(names(miss.bins)), titv.miss, xlab="missing rate", ylab="TiTv")
```

```{r}
depth <- seqApply(gds, "annotation/format/DP", mean, margin="by.variant", 
                  as.is="double", na.rm=TRUE)
depth.bins <- binVar(depth, variant.id, seq(0,200,20))
nbins <- length(depth.bins)
titv.depth <- rep(NA, nbins)
for (i in 1:nbins) {
  capture.output(titv.depth[i] <- applyMethod(gds, titv, variant=depth.bins[[i]]))
}
plot(as.numeric(names(depth.bins)), titv.depth, xlab="mean depth", ylab="TiTv")
```

### Heterozygosity

We will calculate the ratio of heterozygotes to non-reference homozygotes by sample. First, we filter the data to exclude any variants with missing rate $<0.1$ or heterozygosity $>0.6$.

```{r}
miss.var <- missingGenotypeRate(gds, margin="by.variant")
het.var <- heterozygosity(gds, margin="by.variant")
filt <- seqGetData(gds, "variant.id")[miss.var <= 0.1 & het.var <= 0.6]
```

We calculate the heterozygosity and homozyogity by sample, using only the variants selected above.

```{r}
seqSetFilter(gds, variant.id=filt)
hethom <- hethom(gds)
hist(hethom, main="", xlab="Het/Hom Non-Ref")
seqSetFilter(gds)
```

### Principal Component Analysis

We can do Principal Component Analysis (PCA) to separate subjects by ancestry. All the samples in the example file are CEU, so we expect to see only one cluster.

```{r}
pc <- pca(gds)
names(pc)
plot(pc$eigenvect[,1], pc$eigenvect[,2])
```

See also the packages `SNPRelate` and `GENESIS` for determining relatedness and population structure.

### Hardy-Weinberg Equilibrium

We can test for deviations from Hardy-Weinberg Equilibrium (HWE), which can reveal variants of low quality. A test with permuted genotypes gives expected values under the null hypothesis of HWE.

```{r}
hw <- hwe(gds)
pval <- -log10(sort(hw$p))
hw.perm <- hwe(gds, permute=TRUE)
x <- -log10(sort(hw.perm$p))
plot(x, pval, xlab="-log10(expected P)", ylab="-log10(observed P)")
abline(0,1,col="red")
```

The inbreeding coefficient can also be used as a quality metric. For variants, this is 1 - observed heterozygosity / expected heterozygosity.

```{r}
hist(hw$f)
```

We can also calculate the inbreeding coefficient by sample.

```{r}
ic <- inbreedCoeff(gds, margin="by.sample")
range(ic)
```

### Mendelian errors

Checking for Mendelian errors is another way to assess variant quality. The example data contains a trio (child, mother, and father).

```{r}
data(pedigree)
pedigree[pedigree$family == 1463,]
err <- mendelErr(gds, pedigree, verbose=FALSE)
table(err$by.variant)
err$by.trio
```

The example data do not have any Mendelian errors.

### Association tests

We can run single-variant association tests for continuous or binary traits. We use a `SeqVarData` object to combine the genotypes with sample annotation. We use the pedigree data from the previous section and add simulated phenotypes.

```{r}
library(Biobase)
sample.id <- seqGetData(gds, "sample.id")
pedigree <- pedigree[match(sample.id, pedigree$sample.id),]
n <- length(sample.id)
pedigree$phenotype <- rnorm(n, mean=10)
pedigree$case.status <- rbinom(n, 1, 0.3)
sample.data <- AnnotatedDataFrame(pedigree)

seqData <- SeqVarData(gds, sample.data)

## continuous phenotype
assoc <- regression(seqData, outcome="phenotype", covar="sex",
                    model.type="linear")
head(assoc)
pval <- -log10(sort(assoc$Wald.Pval))
n <- length(pval)
x <- -log10((1:n)/n)
plot(x, pval, xlab="-log10(expected P)", ylab="-log10(observed P)")
abline(0, 1, col = "red")
```

For binary phenotypes, there are two options, "logistic" and "firth". "logistic" uses `glm` and performs a Wald test. "firth" uses `logistf`. We recommend using the Firth test for rare variants ([Wang 2014](https://pubmed.ncbi.nlm.nih.gov/24995013/))]

```{r}
assoc <- regression(seqData, outcome="case.status", covar="sex",
                    model.type="firth")
head(assoc)
pval <- -log10(sort(assoc$PPL.Pval))
n <- length(pval)
x <- -log10((1:n)/n)
plot(x, pval, xlab="-log10(expected P)", ylab="-log10(observed P)")
abline(0, 1, col = "red")
```

For aggregate tests, see the package `GENESIS`.

```{r}
seqClose(gds)
```

```{r, echo=FALSE, results='hide'}
unlink(gdsfile)
```

\name{meanBySample}
\alias{meanBySample}
\alias{meanBySample,SeqVarGDSClass-method}

\title{Mean value by sample}

\description{Calculate the mean value of a variable by sample over all variants}

\usage{
\S4method{meanBySample}{SeqVarGDSClass}(gdsobj, var.name, use.names=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{var.name}{Character string with name of the variable, most
    likely "annotation/format/VARIABLE_NAME".}
  \item{use.names}{A logical indicating whether to assign sample IDs
    as names of the output vector.}
}

\details{
  Mean values by variant can be calculated using \code{seqApply(gdsobj, var.name,
    mean, na.rm=TRUE)}.  Currently \code{seqApply} can only be used with
  the option \code{margin="by.variant"}.
  This method provides a way to calculate mean values by sample.
}

\value{
  A numeric vector of mean values.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{seqApply}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
head(meanBySample(gds, "annotation/format/DP", use.names=TRUE))
seqClose(gds)
}

\keyword{manip}

\name{countSingletons}
\alias{countSingletons}
\alias{countSingletons,SeqVarGDSClass-method}

\title{Count singletons}

\description{Count singleton variants for each sample}

\usage{
\S4method{countSingletons}{SeqVarGDSClass}(gdsobj, use.names=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{use.names}{A logical indicating whether to assign variant IDs
    as names of the output vector.}
}

\details{
  A singleton variant is a variant in which only one sample has a non-reference allele.  For each sample, \code{countSingletons} finds the number of variants for which that sample has the only non-reference allele.
}

\value{
  A vector of the number of singleton variants per sample.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{alleleFrequency}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
head(countSingletons(gds))
seqClose(gds)
}

\keyword{manip}

\name{mendelErr}
\alias{mendelErr}
\alias{mendelErr,SeqVarGDSClass-method}

\title{Mendelian errors}

\description{Detect Mendelian errors}

\usage{
\S4method{mendelErr}{SeqVarGDSClass}(gdsobj, pedigree, use.names=FALSE,
autosomes=1:22, xchrom="X", ychrom="Y", verbose=TRUE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{pedigree}{A data.frame with columns (family, individ, father,
  mother, sex, sample.id).  "sex" column should have values "M"/"F".
  "sample.id" values should correspond to "sample.id" in \code{gdsobj}.}
  \item{use.names}{A logical indicating whether to assign variant IDs
    as names of the output vector.}
  \item{autosomes}{A vector with chromosome values in \code{gdsobj} corresponding to autosomes.}
  \item{xchrom}{The chromosome value in \code{gdsobj} corresponding to
    the X chromosome.}
  \item{ychrom}{The chromosome value in \code{gdsobj} corresponding to
    the Y chromosome.}
  \item{verbose}{A logical indicating whether to print the number of
  samples selected for each trio.}
}

\details{
  Mendelian errors are detected for each trio in \code{pedigree}.  Duos
  (mother or father missing) are included.  The pedigree must have only
  one sample per individual.
}

\value{
  A list with the following elements:
  \item{by.variant}{An integer vector with the number of mendelian errors detected for
    each variant.  If \code{use.names=TRUE}, the vector will be named
  with variant IDs.}
  \item{by.trio}{An integer vector with the number of mendelian errors
  detected for each trio.  The vector will be named with the sample ID
  of the child in each trio.}
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
data(pedigree)
err <- mendelErr(gds, pedigree)
table(err$by.variant)
err$by.trio
seqClose(gds)
}

\keyword{manip}

\name{isSNV}
\alias{isSNV}
\alias{isSNV,SeqVarGDSClass-method}

\title{Flag single nucleotide variants}

\description{Flag single nucleotide variants}

\usage{
\S4method{isSNV}{SeqVarGDSClass}(gdsobj, biallelic=TRUE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{biallelic}{A logical indicating whether only biallelic SNVs are considered.}
}

\details{
  If \code{biallelic=TRUE}, a variant is considered a single nucleotide variant (SNV) if there is
  one reference allele and one alternate allele, each one base in
  length.  If \code{biallelic=FALSE}, there may be multiple alternate
  alleles, each one base in length.

  Setting \code{biallelic=TRUE} is considerably faster for large data sets.
}

\value{
  A logical vector indicating which variants are SNVs.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{allele-methods}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
table(isSNV(gds))
seqClose(gds)
}

\name{alternateAlleleDetection}
\alias{alternateAlleleDetection}
\alias{alternateAlleleDetection,SeqVarData,SeqVarData-method}

\title{alternateAlleleDetection}

\description{Calculate rates of detecting minor alleles given a ``gold standard'' dataset}

\usage{
\S4method{alternateAlleleDetection}{SeqVarData,SeqVarData}(gdsobj, gdsobj2,
    match.samples.on=c("subject.id", "subject.id"), verbose=TRUE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarData}} object with VCF data.}
  \item{gdsobj2}{A \code{\link{SeqVarData}} object with VCF data to be used as the ``gold standard''.}
  \item{match.samples.on}{A length-2 character vector indicating the column to be used for matching in each dataset's \code{sampleData} annotation}
  \item{verbose}{A logical indicating whether to print progress messages.}
}

\details{

  Calculates the accuracy of detecting alternate alleles in one dataset (\code{gdsobj}) given a ``gold standard'' dataset (\code{gdsobj2}).
  Samples are matched using the \code{match.samples.on} argument. The first element of \code{match.samples.on} indicates the column to be used as the subject identifier for the first dataset, and the second element is the column to be used for the second dataset.
  Variants are matched on position and alleles using bi-allelic SNVs only.
  Genotype dosages are recoded to count the same allele if the reference allele in one dataset is the alternate allele in the other dataset.
  If a variant in one dataset matches to multiple variants in the second dataset, then only the first match will be used.
  If a variant is missing in either dataset for a given sample pair, that sample pair is ignored for that variant.
  To exclude certain variants or samples from the calculate, use \code{\link{seqSetFilter}} to set appropriate filters on each gds object.
  
  This test is positive if an alternate allele was been detected.
  Results are returned on an allele level, such that:

  \code{TP}, \code{TN}, \code{FP}, and \code{FN} are calculated as follows:
  \tabular{lllll}{
              \tab    \tab           \tab genoData2 \tab \cr
              \tab    \tab aa        \tab Ra        \tab RR\cr
              \tab aa \tab 2TP       \tab 1TP + 1FP \tab 2FP\cr
    genoData1 \tab Ra \tab 1TP + 1FN \tab 1TN + 1TP \tab 1TN + 1FP\cr
	      \tab RR \tab 2FN       \tab 1FN + 1TN \tab 2TN\cr
	    }
  where ``R'' indicates a reference allele and ``a'' indicates an alternate allele.
}

\value{
  A data frame with the following columns:
\item{variant.id.1}{variant id from the first dataset}
\item{variant.id.2}{matched variant id from the second dataset}
\item{n.samples}{the number of samples with non-missing data for this variant}
\item{true.pos}{the number of alleles that are true positives for this variant}
\item{true.neg}{the number of alleles that are true negatives for this variant}
\item{false.pos}{the number of alleles that are false positives for this variant}
\item{false.neg}{the number of alleles that are false negatives for this variant}
}

\author{Adrienne Stilp}

\seealso{
  \code{\link{SeqVarGDSClass}}
}

\examples{
\dontrun{
gds1 <- seqOpen(gdsfile.1) # dataset to test, e.g. sequencing
sample1 <- data.frame(subject.id=c("a", "b", "c"), sample.id=c("A", "B", "C"), stringsAsFactors=F)
seqData1 <- SeqVarData(gds1, sampleData=sample1)

gds2 <- seqOpen(gdsfile.2) # gold standard dataset, e.g. array genotyping
sample2 <- data.frame(subject.id=c("b", "c", "d"), sample.id=c("B", "C", "D"), stringsAsFactors=F)
seqData2 <- SeqVarData(gds2, sampleData=sample2)

res <- alleleDetectionAccuracy(seqData1, seqData2)
}
}

\keyword{manip}

\name{refFrac}
\alias{refFrac}
\alias{refFrac,SeqVarGDSClass-method}
\alias{refFracOverHets}
\alias{refFracOverHets,SeqVarGDSClass-method}
\alias{refFracPlot}
\alias{refFracPlot,SeqVarGDSClass-method}

\title{Reference allele fraction}

\description{
  Calculate fraction of reference allele reads
}

\usage{
\S4method{refFrac}{SeqVarGDSClass}(gdsobj, use.names=TRUE, parallel=FALSE)
\S4method{refFracOverHets}{SeqVarGDSClass}(gdsobj, FUN=mean, use.names=TRUE, parallel=FALSE)
\S4method{refFracPlot}{SeqVarGDSClass}(gdsobj, variant.id, highlight=NULL, ...)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{FUN}{The function to apply over heterozygote calls (mean or median).}
  \item{use.names}{A logical indicating whether to assign variant or
    samples IDs as names of the output vector.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
  \item{variant.id}{A vector of variant.ids to plot.}
  \item{highlight}{A list of sample.ids to highlight with sequential integers on each plot}
  \item{...}{Additional arguments passed to \code{\link{plot}}.}
}

\details{
  The variable "annotation/format/AD" (allelic depth) is required to
  compute the reference allele fraction.

  \code{refFracPlot} generates plots of total unfiltered depth (sum over "AD" for
  all alleles) versus reference allele fraction.  Points are color-coded
  by called genotype: teal = reference homozygote, orange = heterozygote
  including the reference allele, fuschia = heterozygote with two
  alternate alleles, purple = alternate homozygote, black = missing. 
  Darker colors indicate a higher density of points.
  Vertical black line
  is at 0.5, vertical orange line is the median reference allele fraction
  for ref/alt heterozygotes.  Values significantly different from 0.5
  (after applying a Bonferroni correction) are plotted with triangles.
}

\value{
  \code{refFrac} returns a sample by variant array of the reference allele
  fraction, defined as ref_depth / total_depth.

  \code{refFracOverHets} returns the mean (or other function,
  e.g. median) of reference allele
  depth (per variant) over all samples called as heterozygotes. 
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gdsfile <- system.file("extdata", "hapmap_exome_chr22.gds", package="SeqVarTools")
gds <- seqOpen(gdsfile)
RF <- refFrac(gds)
dim(RF)
samples <- seqGetData(gds, "sample.id")
refFracPlot(gds, variant.id=5:6, 
            highlight=list(samples[2:3], samples[4:5]))
seqClose(gds)
}

\keyword{manip}

\name{chromWithPAR}
\alias{chromWithPAR}
\alias{chromWithPAR,SeqVarGDSClass-method}

\title{Identify pseudoautosomal region}

\description{Flag single nucleotide variants}

\usage{
\S4method{chromWithPAR}{SeqVarGDSClass}(gdsobj, genome.build=c("hg19", "hg38"))
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{genome.build}{A character sting indicating genome build.}
}

\details{
  The pseudoautosomal region (PAR) should be treated like the autosomes for purposes of calculating allele frequency. This method returns a vector where sex chromosome variants are labeled wither "X", "Y", or "PAR".
}

\value{
  A character vector of chromosome, with values "PAR" for the pseudoautosomal region.
}

\references{
\url{https://www.ncbi.nlm.nih.gov/grc/human}
}

\author{Stephanie Gogarten}

\name{applyMethod}
\alias{applyMethod}
\alias{applyMethod,SeqVarGDSClass,function,character-method}
\alias{applyMethod,SeqVarGDSClass,function,numeric-method}
\alias{applyMethod,SeqVarGDSClass,function,GRanges-method}
\alias{applyMethod,SeqVarGDSClass,function,missing-method}

\title{Apply method to GDS object}

\description{Apply a method to a subset of variants and/or samples in a GDS object}

\usage{
\S4method{applyMethod}{SeqVarGDSClass,function,character}(gdsobj, FUN, variant, sample=NULL, ...)
\S4method{applyMethod}{SeqVarGDSClass,function,numeric}(gdsobj, FUN, variant, sample=NULL, ...)
\S4method{applyMethod}{SeqVarGDSClass,function,GRanges}(gdsobj, FUN, variant, sample=NULL, ...)
\S4method{applyMethod}{SeqVarGDSClass,function,missing}(gdsobj, FUN, variant, sample=NULL, ...)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{FUN}{A method or function to be applied to \code{gdsobj}.}
  \item{variant}{A vector of variant.id values or a \code{GRanges}
    object defining the variants to be included in the call to
    \code{FUN}.}
  \item{sample}{A vector of sample.id values defining the samples to be
    included in the call to \code{FUN}.}
  \item{\dots}{Additional arguments, passed to \code{FUN}.}
}

\details{
  \code{applyMethod} applies a method or function \code{FUN} to the
  subset of variants defined by \code{variant} and samples defined by
  \code{sample} in a GDS object.

  If a filter was previously set with \code{\link{seqSetFilter}}, it
  will be saved and reset after the call to \code{applyMethod}.
}

\value{
  The result of the call to \code{FUN}.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
variant.id <- seqGetData(gds, "variant.id")
sample.id <- seqGetData(gds, "sample.id")
applyMethod(gds, getGenotype, variant.id[1:5], sample.id[1:10])

library(GenomicRanges)
chrom <- seqGetData(gds, "chromosome")
pos22 <- seqGetData(gds, "position")[chrom == 22]
ranges <- GRanges(seqnames="22", IRanges(min(pos22), max(pos22)))
applyMethod(gds, heterozygosity, ranges, margin="by.sample")
applyMethod(gds, heterozygosity, ranges, margin="by.variant")

seqClose(gds)
}

\keyword{manip}

\name{heterozygosity}
\alias{heterozygosity}
\alias{heterozygosity,SeqVarGDSClass-method}
\alias{homozygosity}
\alias{homozygosity,SeqVarGDSClass-method}
\alias{hethom}
\alias{hethom,SeqVarGDSClass-method}

\title{Heterozygosity and Homozygosity}

\description{
  Calculate heterozygosity and homozygosity by variant or by sample
}

\usage{
\S4method{heterozygosity}{SeqVarGDSClass}(gdsobj, margin=c("by.variant", "by.sample"),
    use.names=FALSE, parallel=FALSE)
\S4method{homozygosity}{SeqVarGDSClass}(gdsobj, allele=c("any", "ref", "alt"), margin=c("by.variant", "by.sample"),
    use.names=FALSE, parallel=FALSE)
\S4method{hethom}{SeqVarGDSClass}(gdsobj, use.names=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{margin}{Possible values are "by.variant" or "by.sample,"
    indicating whether the calculation should be done over all samples
    for each variant, or over all variants for each sample.}
  \item{use.names}{A logical indicating whether to assign variant or
    samples IDs as names of the output vector.}
  \item{allele}{Possible values are "any", "ref," or "alt," indicating
    which alleles to consider when calculating homozygosity.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details. Only applies
  if \code{margin="as.variant"}.}
}

\details{
  \code{heterozyogosity} calulates the fraction of heterozygous genotypes
  in a GDS object, either by variant or by sample.

  \code{homozygosity} calculates the rate of homozygous genotypes in a
  GDS object, either by sample or by variant.  If \code{allele="any"},
  all homozygous genotypes are considered (reference or any alternate
  allele).  If \code{allele="ref"}, only reference homozygotes are
  considered.  If \code{allele="alt"}, any alternate allele homozygote
  is considered.  For example, "ref" will count "0/0" genotypes only,
  "alt" will count "1/1", "2/2", etc. (but not "0/0"), and "any" will
  count all of the above.

  \code{hethom} calculates the ratio of heterozygous genotypes to alternate homozygous genotypes by sample.
}

\value{
  A numeric vector of heterozyogity or homozygosity rates.  If
  \code{margin="by.variant"}, the vector will have length equal to the
  number of variants in the GDS object.  If \code{margin="by.sample"},
  the vector will have length equal to the number of samples.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{alleleFrequency}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
head(heterozygosity(gds, margin="by.variant"))
head(homozygosity(gds, allele="any", margin="by.variant"))
head(homozygosity(gds, allele="ref", margin="by.variant"))
head(homozygosity(gds, allele="alt", margin="by.variant"))

## Het/Hom Non-Ref by sample
head(hethom(gds))

seqClose(gds)
}

\keyword{manip}

\name{SeqVarData}
\docType{class}

\alias{SeqVarData-class}
\alias{SeqVarData}
\alias{show,SeqVarData-method}
\alias{sampleData}
\alias{sampleData,SeqVarData-method}
\alias{sampleData<-}
\alias{sampleData<-,SeqVarData,AnnotatedDataFrame-method}
\alias{variantData}
\alias{variantData,SeqVarData-method}
\alias{variantData<-}
\alias{variantData<-,SeqVarData,AnnotatedDataFrame-method}
\alias{granges}
\alias{granges,SeqVarData-method}
\alias{validateSex}
\alias{validateSex,SeqVarData-method}

\title{SeqVarData}

\description{
	Extends \code{\link{SeqVarGDSClass}} to include annotation for samples and variants.
}

\details{
  A \code{SeqVarData} object adds an \code{\link{AnnotatedDataFrame}} for both samples and variants to a \code{\link{SeqVarGDSClass}} object.

  Note that a \code{SeqVarData} object must be created using an \emph{unfiltered}  \code{\link{SeqVarGDSClass}} object. The \code{sample.id} column in the \code{sampleData} \code{\link{AnnotatedDataFrame}} must exactly match the \code{sample.id} node in the GDS file (and similarly for \code{variant.id} in \code{variantData}). This enables all subsequent filters set on the \code{SeqVarData} object to apply to the GDS and the annotation simultaneously.
}

\section{Constructor}{
  \itemize{
    \item{
      \code{SeqVarData(gds, sampleData, variantData)}: Returns a \code{SeqVarData} object.

      \code{gds} can be either the filename of a sequencing GDS file or an existing \code{\link{SeqVarGDSClass}} object.

      \code{sampleData} must be an \code{\link{AnnotatedDataFrame}} with a column \code{sample.id} matching \code{sample.id} in the GDS file. If this argument is missing, a data frame with 0 columns will be created.

      \code{variantData} must be an \code{\link{AnnotatedDataFrame}} with a column \code{variant.id} matching \code{variant.id} in the GDS file. If this argument is missing, a data frame with 0 columns will be created.
    }
  }
}


\section{Accessors}{
	\itemize{	
		\item{
			\code{sampleData(x)}, \code{sampleData(x)<- value}:
				Get or set the \code{\link{AnnotatedDataFrame}} with sample data. If a sample filter has been applied with \code{\link{seqSetFilter}}, only selected samples will be returned. \code{value} must include all samples.
		}	
		\item{
			\code{variantData(x)}, \code{variantData(x)<- value}:
				Get or set the \code{\link{AnnotatedDataFrame}} with variant data. If a variant filter has been applied with \code{\link{seqSetFilter}}, only selected variants will be returned. \code{value} must include all variants.
		}	
		\item{
			\code{granges(x)}:
				Return a \code{\link{GRanges}} object with the columns of \code{variantData} as metadata columns.
		}	
		\item{
			\code{validateSex(x)}:
				Return the contents of a column named "sex" in \code{sampleData(x)}, provided the contents are valid (values either "M"/"F" or 1/2, or NA). If the column is missing or invalid, return \code{NULL}.
		}	
	}

	See \code{\link{SeqVarGDSClass}} for additional methods.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{seqVCF2GDS}},
  \code{\link{seqOpen}},
  \code{\link{seqGetData}},
  \code{\link{seqSetFilter}},
  \code{\link{seqApply}},
  \code{\link{seqClose}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))

## create sample annotation
library(Biobase)
sample.id <- seqGetData(gds, "sample.id")
sex <- sample(c("M","F"), length(sample.id), replace=TRUE)
phenotype <- rnorm(length(sample.id), mean=10)
samp <- data.frame(sample.id, sex, phenotype, stringsAsFactors=FALSE)
meta <- data.frame(labelDescription=c("unique sample identifier",
     "sex (M=male, f=female)", "example phenotype"), 
      row.names=names(samp), stringsAsFactors=FALSE)
sample.data <- AnnotatedDataFrame(samp, meta)

seqData <- SeqVarData(gds, sample.data)

head(validateSex(seqData))

## add another annotation column
sample.data$site <- sample(letters, length(sample.id), replace=TRUE)
varMetadata(sample.data)["site", "labelDescription"] <- "study site"
sampleData(seqData) <- sample.data

## set a filter
seqSetFilter(seqData, sample.id=sample.id[1:10])
nrow(sampleData(seqData))

seqClose(seqData)
}

\keyword{methods}
\keyword{classes}

\name{inbreedCoeff}
\alias{inbreedCoeff}
\alias{inbreedCoeff,SeqVarGDSClass-method}

\title{Inbreeding coefficient}

\description{Calculates the inbreeding coefficient by variant or by sample}

\usage{
\S4method{inbreedCoeff}{SeqVarGDSClass}(gdsobj, margin=c("by.variant", "by.sample"), use.names=FALSE,
    parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{margin}{Possible values are "by.variant" or "by.sample,"
    indicating whether the calculation should be done over all samples
    for each variant, or over all variants for each sample.}
  \item{use.names}{A logical indicating whether to assign variant or
    sample IDs as names of the output vector.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details. Only applies
  if \code{margin="as.variant"}.}
}

\details{
  For inbreeding coefficients by variant, calculates 1 - observed
  heterozygosity / expected heterozygosity.
  
  For individual inbreeding coefficients (\code{margin="by.sample"}),
  calculates Visscher's estimator described in Yang et al. (2010).
}

\value{
  Values for the inbreeding coefficient.
}

\author{Xiuwen Zheng, Stephanie Gogarten}

\references{
  Yang J, Benyamin B, McEvoy BP, Gordon S, Henders AK, Nyholt DR, Madden
  PA, Heath AC, Martin NG, Montgomery GW, Goddard ME, Visscher
  PM. 2010. Common SNPs explain a large proportion of the heritability
  for human height. Nat Genet. 42(7):565-9. Epub 2010 Jun 20.
}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
f <- inbreedCoeff(gds, margin="by.variant")
range(f, na.rm=TRUE)

ic <- inbreedCoeff(gds, margin="by.sample")
range(ic)
seqClose(gds)
}

\keyword{manip}

\name{pca}
\alias{pca}
\alias{pca,SeqVarGDSClass-method}

\title{Principal Component Analysis}

\description{Calculates the eigenvalues and eignevectors of a
  SeqVarGDSClass object with Principal Component Analysis}

\usage{
\S4method{pca}{SeqVarGDSClass}(gdsobj, eigen.cnt=32)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{eigen.cnt}{An integer indicating how many eigenvalues and
    eignvectors to return.}
}

\details{
  Calculates the genetic covariance matrix and finds the eigen decomposition.
}

\value{
  A list with two elements:
  \item{eigenval}{A vector of length \code{eigen.cnt} with eigenvalues}
  \item{eigenvect}{A matrix of dimension ("selected samples", \code{eigen.cnt}).}
}

\author{Xiuwen Zheng, Stephanie Gogarten}

\references{
  Patterson N, Price AL, Reich D (2006) Population structure and
  eigenanalysis. PLoS Genetics 2:e190.
}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
pca <- pca(gds)
pca$eigenval
head(pca$eigenvect)
seqClose(gds)
}

\keyword{manip}

\name{regression}
\alias{regression}
\alias{regression,SeqVarData-method}

\title{Linear or logistic regression}

\description{Run linear or logistic regression on variants}

\usage{
\S4method{regression}{SeqVarData}(gdsobj, outcome, covar=NULL,
    model.type=c("linear", "logistic", "firth"),
    parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarData}} object}
  \item{outcome}{A character string with the name of the column in \code{sampleData(gdsobj)} containing the outcome variable}
  \item{covar}{A character vector with the name of the column(s) in \code{sampleData(gdsobj)} containing the covariates}
  \item{model.type}{the type of model to be run.  "linear" uses \code{\link{lm}}, "logistic" uses \code{\link{glm}} with \code{family=binomial()}, and "firth" uses \code{\link{logistf}}.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}

}

\details{
\code{regression} tests the additive effect of the reference allele.
}

\value{
a data.frame with the following columns (if applicable):
  \item{variant.id}{variant identifier}
  \item{n}{number of samples with non-missing data}
  \item{n0}{number of controls (outcome=0) with non-missing data}
  \item{n1}{number of cases (outcome=1) with non-missing data}
  \item{freq}{reference allele frequency}
  \item{freq0}{reference allele frequency in controls}
  \item{freq1}{reference allele frequency in cases}
  \item{Est}{beta estimate for genotype}
  \item{SE}{standard error of beta estimate for the genotype}
  \item{Wald.Stat}{chi-squared test statistic for association}
  \item{Wald.pval}{p-value for association}
  \item{PPL.Stat}{firth only: profile penalized likelihood test statistic for association}
  \item{PPL.pval}{firth only: p-value for association}
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarData}},
  \code{\link{seqSetFilter}},
  \code{\link{lm}},
  \code{\link{glm}},
  \code{\link{logistf}}	
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))

## create some phenotype data
library(Biobase)
sample.id <- seqGetData(gds, "sample.id")
n <- length(sample.id)
df <- data.frame(sample.id,
   sex=sample(c("M", "F"), n, replace=TRUE),
   age=sample(18:70, n, replace=TRUE),
   phen=rnorm(n),
   stringsAsFactors=FALSE)
meta <- data.frame(labelDescription=c("sample identifier",
   "sex", "age", "phenotype"), row.names=names(df))
sample.data <- AnnotatedDataFrame(df, meta)
seqData <- SeqVarData(gds, sample.data)

## select samples and variants
seqSetFilter(gds, sample.id=sample.id[1:50], variant.id=1:10)

res <- regression(seqData, outcome="phen", covar=c("sex", "age"))
res
seqClose(gds)
}

\keyword{manip}

\name{alleleFrequency}
\alias{alleleFrequency}
\alias{alleleFrequency,SeqVarGDSClass-method}
\alias{alleleFrequency,SeqVarData-method}
\alias{alleleCount}
\alias{alleleCount,SeqVarGDSClass-method}
\alias{alleleCount,SeqVarData-method}
\alias{minorAlleleCount}
\alias{minorAlleleCount,SeqVarGDSClass-method}
\alias{minorAlleleCount,SeqVarData-method}

\title{Allele frequency}

\description{Calculate allele frequency for each variant}

\usage{
\S4method{alleleFrequency}{SeqVarGDSClass}(gdsobj, n=0, use.names=FALSE, parallel=FALSE)
\S4method{alleleFrequency}{SeqVarData}(gdsobj, n=0, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE,
    genome.build=c("hg19", "hg38"), parallel=FALSE)
\S4method{alleleCount}{SeqVarGDSClass}(gdsobj, n=0, use.names=FALSE, parallel=FALSE)
\S4method{alleleCount}{SeqVarData}(gdsobj, n=0, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE,
    genome.build=c("hg19", "hg38"), parallel=FALSE)
\S4method{minorAlleleCount}{SeqVarData}(gdsobj, use.names=FALSE, sex.adjust=TRUE, male.diploid=TRUE,
    genome.build=c("hg19", "hg38"), parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{n}{An integer indicating which allele to calculate the frequency
  of.  \code{n=0} is the reference allele, \code{n=1} is the first
  alternate allele, and so on.}
  \item{use.names}{A logical indicating whether to assign variant IDs
    as names of the output vector.}
  \item{sex.adjust}{Logical for whether to adjust frequency calculations based on sex. If \code{TRUE}, X chromosome frequency (excluding the PAR) will be calculated assuming the dosage of the specifed allele for males is half that for females. Y chromosome frequency will be calculated using males only.}
  \item{male.diploid}{Logical for whether males on sex chromosomes are coded as diploid.}
  \item{genome.build}{A character sting indicating genome build; used to identify pseudoautosomal regions on the X and Y chromosomes.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
}

\details{
  Frequency or count can be calculated over any allele, specified by the argument
  \code{n}.  Default is the reference allele frequency (\code{n=0}).

  The \code{\link{SeqVarData}} method will calculate frequency and count correctly for X and Y chromosomes, provided a column "sex" is included in the sampleData slot with values "M"/"F" or 1/2. Arguments given to this method are passed to the parent method for \code{\link{SeqVarGDSClass}}. If the ploidy of the "genotype" node in the GDS file is 2, the default assumption is that genotypes for males on sex chromosomes are coded as diploid, "0/0" or "1/1". If this is not the case, use \code{male.diploid=FALSE}.

  For multiallelic variants, the minor allele count will be the smaller of the reference allele count or the sum of all alternate allele counts. 
}

\value{
  A numeric vector of allele frequencies.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{chromWithPAR}},
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{heterozygosity}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
head(alleleFrequency(gds))
head(alleleFrequency(gds, n=1))
head(alleleFrequency(gds, n=2))
seqClose(gds)
}

\keyword{manip}

\name{missingGenotypeRate}
\alias{missingGenotypeRate}
\alias{missingGenotypeRate,SeqVarGDSClass-method}

\title{Missing genotype rate}

\description{
  Calculate missing genotype rate by variant or by sample
}

\usage{
\S4method{missingGenotypeRate}{SeqVarGDSClass}(gdsobj, margin=c("by.variant", "by.sample"), use.names=FALSE,
    parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{margin}{Possible values are "by.variant" or "by.sample,"
    indicating whether the calculation should be done over all samples
    for each variant, or over all variants for each sample.}
  \item{use.names}{A logical indicating whether to assign variant IDs
    as names of the output vector.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
}

\details{
  Calculates the fraction of missing genotypes in a GDS object, either
  by variant or by sample.
}

\value{
  A numeric vector of missing genotype rates.  If
  \code{margin="by.variant"}, the vector will have length equal to the
  number of variants in the GDS object.  If \code{margin="by.sample"},
  the vector will have length equal to the number of samples.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{getGenotype}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
head(missingGenotypeRate(gds, margin="by.variant"))
head(missingGenotypeRate(gds, margin="by.sample"))
seqClose(gds)
}

\keyword{manip}

\name{duplicateDiscordance}
\alias{duplicateDiscordance}
\alias{duplicateDiscordance,SeqVarData,missing-method}
\alias{duplicateDiscordance,SeqVarData,SeqVarData-method}
\alias{duplicateDiscordance,SeqVarIterator,missing-method}

\title{Duplicate discordance}

\description{Find discordance rate for duplicate sample pairs}

\usage{
%\S4method{duplicateDiscordance}{SeqVarData,missing}(gdsobj, match.samples.on="subject.id", check.phase=FALSE, verbose=TRUE)
\S4method{duplicateDiscordance}{SeqVarData,missing}(gdsobj, match.samples.on="subject.id", by.variant=FALSE,
    all.pairs=TRUE, verbose=TRUE)
\S4method{duplicateDiscordance}{SeqVarIterator,missing}(gdsobj, match.samples.on="subject.id", by.variant=FALSE,
    all.pairs=TRUE, verbose=TRUE)
\S4method{duplicateDiscordance}{SeqVarData,SeqVarData}(gdsobj, obj2, match.samples.on=c("subject.id", "subject.id"),
    match.variants.on=c("alleles", "position"),
    discordance.type=c("genotype", "hethom"),
    by.variant=FALSE, verbose=TRUE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarData}} object with VCF data.}
  \item{obj2}{A \code{\link{SeqVarData}} object with VCF data.}
  \item{match.samples.on}{Character string or vector of strings indicating which column should be used for matching samples. See details.}
  \item{match.variants.on}{Character string of length one indicating how to match variants. See details.}
  \item{discordance.type}{Character string describing how discordances should be calculated. See details.}
  %\item{check.phase}{A logical indicating whether phase should be considered when calculating discordance.}
  \item{by.variant}{Calculate discordance by variant, otherwise by sample}
  \item{all.pairs}{Logical for whether to include all possible pairs of samples (\code{all.pairs=TRUE}) or only the first pair per subject (\code{all.pairs=FALSE}).}
  \item{verbose}{A logical indicating whether to print progress messages.}
}

\details{
  %For calls that involve only one gds file, duplicate discordance is calculated by sample pair and by variant.  If there are more than two samples per subject in \code{samples}, only the first two samples are used and a warning message is printed.
  %If \code{check.phase=TRUE}, variants with mismatched phase are considered discordant.  If \code{check.phase=FALSE}, phase is ignored.
  For calls that involve only one gds file, duplicate discordance is calculated by matching samples on common values of a column in \code{\link{sampleData}}. If \code{all.pairs=TRUE}, every possible pair of samples is included, so there may be multiple pairs per subject. If \code{all.pairs=FALSE}, only the first pair for each subject is used.
 
  For calls that involve two gds files, 
  duplicate discordance is calculated by matching sample pairs and variants between the two data sets. 
  Only biallelic SNVs are considered in the comparison.
  Variants can be matched using chromosome and position only (\code{match.variants.on="position"}) or by using chromosome, position, and alleles (\code{match.variants.on="alleles"}).
  If matching on alleles and the reference allele in the first dataset is the alternate allele in the second dataset, the genotype dosage will be recoded so the same allele is counted before making the comparison.
  If a variant in one dataset maps to multiple variants in the other dataset, only the first pair is considered for the comparison.
  Discordances can be calculated using either genotypes (\code{discordance.type = "genotype"}) or heterozygote/homozygote status (\code{discordance.type = "hethom"}).
  The latter is a method to calculate discordance that does not require alleles to be measured on the same strand in both datasets, so it is probably best to also set \code{match.variants.on = "position"} if using the \code{"hethom"} option.

  The argument \code{match.samples.on} can be used to select which column in the \code{\link{sampleData}} of the input \code{\link{SeqVarData}} object should be used for matching samples.
  For one gds file, \code{match.samples.on} should be a single string.
  For two gds files, \code{match.samples.on} should be a length-2 vector of character strings, where the first element is the column to use for the first gds object and the second element is the column to use for the second gds file.

  To exclude certain variants or samples from the calculate, use \code{\link{seqSetFilter}} to set appropriate filters on each gds object.

}

\value{
  %% For calls involving one gds file, a list with the following elements:
  %% \item{by.variant}{A data.frame with the number of discordances for
  %% each variant, the
  %% number of sample pairs with non-missing data, and the discordance rate
  %% (num.discord / num.pair).  Row names are variant ids.}
  %% \item{by.subject}{A data.frame with the sample ids for each pair, the
  %% number of discordances, the number of non-missing variants, and the
  %% discordance rate (num.discord / num.var).  Row.names are subject.id
  %% (as given in \code{samples}).}
  
  %% For calls involving two gds files,
  A data frame with the following columns, depending on whether \code{by.variant=TRUE} or \code{FALSE}:
\item{subject.id}{currently, this is the sample ID (\code{by.variant=FALSE} only)}
\item{sample.id.1/variant.id.1}{sample id or variant id in the first gds file}
\item{sample.id.2/variant.id.2}{sample id or variant id in the second gds file}
\item{n.variants/n.samples}{the number of non-missing variants or samples that were compared}
\item{n.concordant}{the number of concordant variants}
\item{n.alt}{the number of variants involving the alternate allele in either sample}
\item{n.alt.conc}{the number of concordant variants invovling the alternate allele in either sample}
\item{n.het.ref}{the number of mismatches where one call is a heterozygote and the other is a reference homozygote}
\item{n.het.alt}{the number of mismatches where one call is a heterozygote and the other is an alternate homozygote}
\item{n.ref.alt}{the number of mismatches where the calls are opposite homozygotes}

}

\author{Stephanie Gogarten, Adrienne Stilp}

\seealso{
  \code{\link{SeqVarData}},
  \code{\link{SeqVarIterator}}
}

\examples{
require(Biobase)

gds <- seqOpen(seqExampleFileName("gds"))

## the example file has one sample per subject, but we
## will match the first four samples into pairs as an example
sample.id <- seqGetData(gds, "sample.id")
samples <- AnnotatedDataFrame(data.frame(data.frame(subject.id=rep(c("subj1", "subj2"), times=45),
                      sample.id=sample.id,
                      stringsAsFactors=FALSE)))
seqData <- SeqVarData(gds, sampleData=samples)

## set a filter on the first four samples
seqSetFilter(seqData, sample.id=sample.id[1:4])

disc <- duplicateDiscordance(seqData, by.variant=FALSE)
disc
disc <- duplicateDiscordance(seqData, by.variant=TRUE)
head(disc)

## recommended to use an iterator object for large datasets
iterator <- SeqVarBlockIterator(seqData)
disc <- duplicateDiscordance(iterator, by.variant=FALSE)
disc

seqClose(gds)
}

\keyword{manip}

\name{setVariantID}
\alias{setVariantID}

\title{Change the variant ID of a GDS file}

\description{
  Replace the variable "variant.id" in a GDS file with a user-supplied
  unique vector of the same length.
}

\usage{
setVariantID(gdsfile, variant.id)
}

\arguments{
  \item{gdsfile}{A character string with the file path of a GDS file.}
  \item{variant.id}{A vector with the new variant IDs.}
}

\details{
  A VCF file created by \code{\link{seqVCF2GDS}} creates a variable
  "variant.id" containing sequential integers to identify each variant.
  \code{setVariantID} allows the user to replace these values with
  something more meaningful.  The replacement values in
  \code{variant.id} must be unique and have the same length as the
  original "variant.id" vector.

  Using character values for \code{variant.id} may affect performance
  for large datasets.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{seqVCF2GDS}}
}

\examples{
oldfile <- system.file("extdata", "gl_chr1.gds", package="SeqVarTools")
newfile <- tempfile()
file.copy(oldfile, newfile)

gds <- seqOpen(newfile)
rsID <- seqGetData(gds, "annotation/id")
seqClose(gds)

setVariantID(newfile, rsID)
gds <- seqOpen(newfile)
seqGetData(gds, "variant.id")
head(getGenotype(gds))
seqClose(gds)
 
unlink(newfile)
}

\keyword{manip}

\name{titv}
\alias{titv}
\alias{titv,SeqVarGDSClass-method}

\title{Transition/Transversion Ratio}

\description{
  Calculate transition/transversion ratio overall or by sample
}

\usage{
\S4method{titv}{SeqVarGDSClass}(gdsobj, by.sample=FALSE, use.names=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{by.sample}{A logical indicating whether TiTv should be
    calculated by sample or overall for the entire GDS object.}
  \item{use.names}{A logical indicating whether to assign
    sample IDs as names of the output vector (if \code{by.sample=TRUE}).}
}

\details{
  If \code{by.sample=FALSE} (the default), \code{titv} calulates the
  transition/transversion ratio (TiTv) over all samples.

  If \code{by.sample=TRUE}, \code{titv} calculates TiTv over all
  variant genotypes (heterozygous or homozygous non-reference) for each
  sample.
}

\value{
  A single value for TiTv if \code{by.sample=FALSE}.  If \code{by.sample=TRUE},
  a numeric vector containing TiTv for each sample.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{isVariant}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
titv(gds)
titv(gds, by.sample=TRUE)

## apply to a subset of variants
library(GenomicRanges)
chrom <- seqGetData(gds, "chromosome")
pos22 <- seqGetData(gds, "position")[chrom == 22]
ranges <- GRanges(seqnames="22", IRanges(min(pos22), max(pos22)))
applyMethod(gds, titv, ranges)

seqClose(gds)
}

\keyword{manip}

\name{pedigree}
\alias{pedigree}
\docType{data}
\title{Pedigree for example data}
\description{
  Pedigree for example data files in SeqArray.
}
\usage{
pedigree
}
\format{
  A data.frame with the following columns.
  \describe{
    \item{\code{family}}{Family ID}
    \item{\code{individ}}{Individual ID}
    \item{\code{father}}{Father ID}
    \item{\code{mother}}{Mother ID}
    \item{\code{sex}}{Sex}
    \item{\code{sample.id}}{sample.id in VCF/GDS files}
  }
}
\details{
  There is one trio in the pedigree.
}
\source{
HapMap
}
\examples{
data(pedigree)
head(pedigree)
gds <- seqOpen(seqExampleFileName("gds"))
setdiff(seqGetData(gds, "sample.id"), pedigree$sample.id)
seqClose(gds)
}
\keyword{datasets}

\name{Iterator}
\docType{class}

\alias{SeqVarIterator}
\alias{SeqVarIterator-class}
\alias{SeqVarBlockIterator}
\alias{SeqVarBlockIterator-class}
\alias{SeqVarRangeIterator}
\alias{SeqVarRangeIterator-class}
\alias{SeqVarWindowIterator}
\alias{SeqVarWindowIterator-class}
\alias{SeqVarListIterator}
\alias{SeqVarListIterator-class}
\alias{show,SeqVarIterator-method}
\alias{show,SeqVarRangeIterator-method}
\alias{show,SeqVarListIterator-method}
\alias{iterateFilter}
\alias{iterateFilter,SeqVarIterator-method}
\alias{resetIterator}
\alias{resetIterator,SeqVarIterator-method}
\alias{lastFilter}
\alias{lastFilter,SeqVarIterator-method}
\alias{lastFilter<-}
\alias{lastFilter<-,SeqVarIterator,numeric-method}
\alias{variantFilter}
\alias{variantFilter,SeqVarIterator-method}
\alias{variantBlock}
\alias{variantBlock,SeqVarBlockIterator-method}
\alias{variantRanges}
\alias{variantRanges,SeqVarRangeIterator-method}
\alias{variantRanges,SeqVarListIterator-method}
\alias{currentRanges}
\alias{currentRanges,SeqVarBlockIterator-method}
\alias{currentRanges,SeqVarRangeIterator-method}
\alias{currentRanges,SeqVarListIterator-method}
\alias{currentVariants}
\alias{currentVariants,SeqVarIterator-method}
\alias{currentVariants,SeqVarBlockIterator-method}

\title{Iterators}

\description{
	Extends \code{\link{SeqVarData}} to provide iterators over variants.
}

\details{
	Iterator classes allow for iterating filters over blocks of variants, ranges, or sliding windows.

	For \code{SeqVarBlockIterator}, each call to \code{iterateFilter} will select the next unit of \code{variantBlock} variants.

	For \code{SeqVarRangeIterator}, each call to \code{iterateFilter} will select the next range in \code{variantRanges}.

	\code{SeqVarWindowIterator} is an extension of \code{SeqVarRangeIterator} where the ranges are determined automatically by sliding a window of size \code{windowSize} base pairs by steps of \code{windowShift} across the genome. Only windows containing unique sets of variants are kept.

	For \code{SeqVarListIterator}, each call to \code{iterateFilter} will select the next set of ranges in \code{variantRanges}.

	Any filter set on the object previously will be applied in addition to the selected blocks or ranges.
}


\section{Constructors}{
  \itemize{
    \item{
      \code{SeqVarBlockIterator(seqData, variantBlock=10000, verbose=TRUE)}: Returns a \code{SeqVarBlockIterator} object with the filter set to the first block.

      \code{seqData} is a \code{\link{SeqVarData}} object.

      \code{variantBlock} is an integer specifying the number of variants in an iteration block.

      \code{verbose} is a logical indicator for verbose output.
    }
    \item{
      \code{SeqVarRangeIterator(seqData, variantRanges=GRanges(), verbose=TRUE)}: Returns a \code{SeqVarRangeIterator} object with the filter set to the first range.

      \code{seqData} is a \code{\link{SeqVarData}} object.

      \code{variantRanges} is a \code{\link{GRanges}} object specifying the ranges for iteration.

      \code{verbose} is a logical indicator for verbose output.
    }
    \item{
      \code{SeqVarWindowIterator(seqData,  windowSize=10000, windowShift=5000, verbose=TRUE)}: Returns a \code{SeqVarWindowIterator} object with the filter set to the first window.

      \code{seqData} is a \code{\link{SeqVarData}} object.

      \code{windowSize} is the size in base pairs of the sliding window.

      \code{windowShift} is the size in base pairs of the step for each consecutive window.

      \code{verbose} is a logical indicator for verbose output.
    }
    \item{
      \code{SeqVarListIterator(seqData, variantRanges, verbose=TRUE)}: Returns a \code{SeqVarRangeIterator} object with the filter set to the first range.

      \code{seqData} is a \code{\link{SeqVarData}} object.

      \code{variantRanges} is a \code{\link{GRangesList}} object specifying the ranges for iteration.

      \code{verbose} is a logical indicator for verbose output.
    }
  }
}


\section{Accessors}{
  \itemize{
    \item{\code{iterateFilter(x)}: Advance the filter to the next block, range, or set of ranges. Returns \code{TRUE} while there are still variants left to be read, \code{FALSE} if the end of iteration is reached.
    }
    \item{\code{lastFilter(x)}, \code{lastFilter(x)<- value}: Get or set the last filter index from the previous call to \code{iterateFilter}.
    }
    \item{\code{variantBlock(x)}: Get the size of the variant block.
    }
    \item{\code{variantFilter(x)}: Get the list of variant indices.
    }
    \item{\code{variantRanges(x)}: Get the variant ranges.
    }
    \item{\code{currentRanges(x)}: Get the ranges selected in the current iteration.
    }
    \item{\code{currentVariants(x)}: Get the variants selected in the current iteration. Returns a \code{\link{DataFrame}} where the row name is the variant.id, "variant" is the variant position as a \code{link{GRanges}}, "range" is the range the variant is from, and any columns in either \code{\link{variantData}} or the metadata columns of \code{currentRanges} are included.
    }
    \item{\code{resetIterator(x)}: Set the filter to the first block, range, or set of ranges (the same variants that are selected when the iterator object is created).
    }
  }
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{SeqVarData}},
  \code{\link{seqSetFilter}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
seqData <- SeqVarData(gds)

# iterate by blocks
seqSetFilter(seqData, variant.sel=seq(1,1000,2))
iterator <- SeqVarBlockIterator(seqData, variantBlock=10)
seqGetData(iterator, "variant.id")
iterateFilter(iterator)
seqGetData(iterator, "variant.id")
seqResetFilter(iterator)

# iterate by ranges
library(GenomicRanges)
gr <- GRanges(seqnames=rep(1,3), ranges=IRanges(start=c(1e6, 2e6, 3e6), width=1e6))
iterator <- SeqVarRangeIterator(seqData, variantRanges=gr)
granges(iterator)
iterateFilter(iterator) # no variants in the second range
granges(iterator)
iterateFilter(iterator)
granges(iterator)
iterateFilter(iterator)
seqResetFilter(iterator)

# iterate by windows
seqSetFilterChrom(seqData, include="22")
iterator <- SeqVarWindowIterator(seqData)
seqGetData(iterator, "variant.id")
while (iterateFilter(iterator)) {
    print(seqGetData(iterator, "variant.id"))
}
seqResetFilter(iterator)

# iterate by list of ranges
gr <- GRangesList(
  GRanges(seqnames=rep(22,2), ranges=IRanges(start=c(16e6, 17e6), width=1e6)),
  GRanges(seqnames=rep(22,2), ranges=IRanges(start=c(18e6, 20e6), width=1e6)))
iterator <- SeqVarListIterator(seqData, variantRanges=gr)
granges(iterator)
iterateFilter(iterator)
granges(iterator)
iterateFilter(iterator)
resetIterator(iterator)

seqClose(iterator)
}

\keyword{methods}
\keyword{classes}

\name{getGenotype}
\alias{getGenotype}
\alias{getGenotypeAlleles}
\alias{refDosage}
\alias{altDosage}
\alias{expandedAltDosage}
\alias{alleleDosage}
\alias{getGenotype,SeqVarGDSClass-method}
\alias{getGenotypeAlleles,SeqVarGDSClass-method}
\alias{refDosage,SeqVarGDSClass-method}
\alias{altDosage,SeqVarGDSClass-method}
\alias{expandedAltDosage,SeqVarGDSClass-method}
\alias{alleleDosage,SeqVarGDSClass,numeric-method}
\alias{alleleDosage,SeqVarGDSClass,list-method}

\title{Get genotype data}

\description{
  Get matrix of genotype values from a GDS object
}

\usage{
\S4method{getGenotype}{SeqVarGDSClass}(gdsobj, use.names=TRUE, parallel=FALSE)
\S4method{getGenotypeAlleles}{SeqVarGDSClass}(gdsobj, use.names=TRUE, sort=FALSE, parallel=FALSE)
\S4method{refDosage}{SeqVarGDSClass}(gdsobj, use.names=TRUE, ...)
\S4method{altDosage}{SeqVarGDSClass}(gdsobj, use.names=TRUE, sparse=FALSE, parallel=FALSE, ...)
\S4method{expandedAltDosage}{SeqVarGDSClass}(gdsobj, use.names=TRUE, sparse=FALSE, parallel=FALSE)
\S4method{alleleDosage}{SeqVarGDSClass,numeric}(gdsobj, n=0, use.names=TRUE, parallel=FALSE)
\S4method{alleleDosage}{SeqVarGDSClass,list}(gdsobj, n, use.names=TRUE, parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{use.names}{A logical indicating whether to assign sample and
    variant IDs as dimnames of the resulting matrix.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
  \item{sort}{Logical for whether to sort alleles lexographically ("G/T"
    instead of "T/G").}
  \item{sparse}{Logical for whether to return the alterate allele dosage as a sparse matrix using the \pkg{\link{Matrix}} package. In most cases, setting \code{sparse=TRUE} will dramatically reduce the size of the returned object.}
  \item{n}{An integer, vector, or list indicating which allele(s) to return dosage
  for.  \code{n=0} is the reference allele, \code{n=1} is the first
  alternate allele, and so on.}
  \item{...}{Arguments to pass to \code{\link{seqBlockApply}}, e.g. \code{bsize} to set the block size.}
}

\details{
  In \code{getGenotype}, genotypes are coded as in the VCF file, where "0/0" is homozygous
  reference, "0/1" is heterozygous for the first alternate allele, "0/2"
  is heterozygous for the second alternate allele, etc.
  Separators are
  "/" for unphased and "|" for phased.  If \code{sort=TRUE}, all
  returned genotypes will be unphased.
  Missing genotypes are coded as \code{NA}.
  Only haploid or diploid genotypes (the first two alleles at a given site) are returned.

  If the argument \code{n} to\code{alleleDosage} is a single integer, the same allele is counted for all variants. If \code{n} is a vector with length=number of variants in the current filter, a different allele is counted for each variant. If \code{n} is a list, more than one allele can be counted for each variant. For example, if \code{n[[1]]=c(1,3)}, genotypes "0/1" and "0/3" will each have a dosage of 1 and genotype "1/3" will have a dosage of 2.
}

\value{
  \code{getGenotype} and \code{getGenotypeAlleles} return a character
  matrix with dimensions [sample,variant] containing haploid or diploid
  genotypes.

  \code{getGenotype} returns alleles as "0", "1", "2", etc. indicating
  reference and alternate alleles.

  \code{getGenotypeAlleles} returns alleles as "A", "C", "G", "T".
  \code{sort=TRUE} sorts lexographically, which may be useful for
  comparing genotypes with data generated using a different reference
  sequence.

  \code{refDosage} returns an integer matrix with the dosage of the
  reference allele: 2 for two copies of the reference allele ("0/0"), 1
  for one copy of the reference allele, and 0 for two alternate alleles.

  \code{altDosage} returns an integer matrix with the dosage of any
  alternate allele: 2 for two alternate alleles ("1/1", "1/2", etc.), 1
  for one alternate allele, and 0 for no alternate allele (homozygous reference).

  \code{expandedAltDosage} returns an integer matrix with the dosage of each
  alternate allele as a separate column. A variant with 2 possible alternate alleles will have 2 columns of output, etc.

  \code{alleleDosage} with an integer argument returns an integer matrix with the dosage of the
  specified allele only: 2 for two copies of the allele ("0/0" if \code{n=0}, "1/1" if \code{n=1}, etc.), 1
  for one copy of the specified allele, and 0 for no copies of the allele.

  \code{alleleDosage} with a list argument returns a list of sample x allele matrices with the dosage of each specified allele for each variant.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{seqGetData}},
  \code{\link{seqSetFilter}},
  \code{\link{alleleFrequency}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
seqSetFilter(gds, variant.sel=1323:1327, sample.sel=1:10)
nAlleles(gds)
getGenotype(gds)
getGenotypeAlleles(gds)
refDosage(gds)
altDosage(gds)
expandedAltDosage(gds)
alleleDosage(gds, n=0)
alleleDosage(gds, n=1)
alleleDosage(gds, n=c(0,1,0,1,0))
alleleDosage(gds, n=list(0,c(0,1),0,c(0,1),1))
seqClose(gds)
}

\keyword{manip}

\name{imputedDosage}
\alias{imputedDosage}
\alias{imputedDosage,SeqVarGDSClass-method}

\title{Get imputed dosage}

\description{
  Get matrix of imputed dosage values from a GDS object
}

\usage{
\S4method{imputedDosage}{SeqVarGDSClass}(gdsobj, dosage.field="DS", use.names=TRUE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{dosage.field}{The name of the dosage field in the GDS object (will be prepended with \code{"annotation/format"}).}
  \item{use.names}{A logical indicating whether to assign sample and
    variant IDs as dimnames of the resulting matrix.}
}

\details{
  Reads dosage from the dosage-specific field in the GDS object, rather than counting alleles from called genotypes.

  Only one dosage value per variant is allowed; the method will return an error if multiple dosages are present for a single variant.
}

\value{
  A numeric matrix of dosage values with dimensions [sample,variant].
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{refDosage}},
  \code{\link{altDosage}}
}

\examples{
# convert VCF to GDS keeping dosage field
vcffile <- system.file("extdata", "gl_chr1.vcf", package="SeqVarTools")
gdsfile <- tempfile()
seqVCF2GDS(vcffile, gdsfile, fmt.import="DS", storage.option="ZIP_RA",
           verbose=FALSE)

gds <- seqOpen(gdsfile)
dos <- imputedDosage(gds)
head(dos)
seqClose(gds)
unlink(gdsfile)
}

\keyword{manip}

\name{isVariant}
\alias{isVariant}
\alias{isVariant,SeqVarGDSClass-method}

\title{Locate variant samples across sites}

\description{
  Locate which samples are variant for each site in a GDS object
}

\usage{
\S4method{isVariant}{SeqVarGDSClass}(gdsobj, use.names=FALSE, parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{use.names}{A logical indicating whether to assign sample and
    variant IDs as dimnames of the resulting matrix.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
}

\details{
  Each sample/site cell of the resulting matrix is \code{TRUE} if the
  genotype at that location for that sample contains an alternate
  allele.  A genotype of "0/0" is not variant, while genotypes
  "0/1", "1/0", "0/2", etc. are variant.
}

\value{
  A logical matrix with dimensions [sample,site] which is \code{TRUE}
  for cells where the genotype contains an alternate allele.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{getGenotype}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
variant.id <- seqGetData(gds, "variant.id")
sample.id <- seqGetData(gds, "sample.id")
applyMethod(gds, isVariant, variant.id[1:5], sample.id[1:10])
applyMethod(gds, isVariant, variant.id[1:5], sample.id[1:10], use.names=TRUE)
seqClose(gds)
}

\keyword{manip}

\name{hwe}
\alias{hwe}
\alias{hwe,SeqVarGDSClass-method}

\title{Exact test for Hardy-Weinberg equilibrium}

\description{Performs an exact test for Hardy-Weinberg equilibrium on
  Single-Nucleotide Variants}

\usage{
\S4method{hwe}{SeqVarGDSClass}(gdsobj, permute=FALSE, parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{permute}{A logical indicating whether to permute the genotypes to get a set of p-values under the null hypothesis.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
}

\details{
  HWE calculations are performed with the \code{\link{HWExact}}
  function in the \pkg{\link{GWASExactHW}} package.

  \code{permute=TRUE} will permute the genotypes prior to running the test. This can be useful for obtaining a set of expected values under the null hypothesis to compare to the observed values.

  P values are set to \code{NA} for all multiallelic
  and monomorphic variants.
}

\value{
  A data.frame with the following columns:
\item{variant.id}{The unique identifier for the variant}
\item{nAA}{The number of reference homozygotes}
\item{nAa}{The number of heterozygotes}
\item{naa}{The number of alternate homozygotes}
\item{afreq}{The reference allele frequency}
\item{p}{p values for the exact test}
\item{f}{The inbreeding coefficient, 1 - observed
  heterozygosity / expected heterozygosity}
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
## autosomal variants only
auto <- seqGetData(gds, "chromosome") \%in\% 1:22
var.auto <- seqGetData(gds, "variant.id")[auto]
hw <- applyMethod(gds, hwe, variant=var.auto)
head(hw)
sum(is.na(hw$p))
range(hw$p, na.rm=TRUE)
seqClose(gds)
}

\keyword{manip}

\name{allele-methods}
\alias{allele-methods}
\alias{refChar}
\alias{refChar,SeqVarGDSClass-method}
\alias{altChar}
\alias{altChar,SeqVarGDSClass-method}
\alias{nAlleles}
\alias{nAlleles,SeqVarGDSClass-method}

\title{Extract allele information from a GDS object}

\description{
  Extract reference and alternate alleles and allele counts
  from a GDS object.
}

\usage{
\S4method{refChar}{SeqVarGDSClass}(gdsobj)
\S4method{altChar}{SeqVarGDSClass}(gdsobj, n=0)
\S4method{nAlleles}{SeqVarGDSClass}(gdsobj)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{n}{An integer indicating which alternate allele to return.  \code{n=0} returns a
    comma-separated string of all alternate alleles.}
}

\details{
  These methods parse the "allele" field of a GDS object.
}

\value{
  \code{refChar} returns a character vector of reference alleles.

  \code{altChar} returns a character vector of alternate alleles.  If
  \code{n=0}, multiple alternate alleles are represented as a
  comma-separated string.  If \code{n>0}, only the \code{n}th alternate
  allele is returned.

  \code{nAlleles} returns an integer vector of the number of alleles
  (reference and alternate) for each variant.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
table(refChar(gds))
table(altChar(gds))
table(altChar(gds, n=1))
table(altChar(gds, n=2), useNA="ifany")
table(nAlleles(gds))
seqClose(gds)
}

\keyword{manip}

\name{variantInfo}
\alias{variantInfo}
\alias{variantInfo,SeqVarGDSClass-method}
\alias{expandedVariantIndex}
\alias{expandedVariantIndex,SeqVarGDSClass-method}

\title{Variant info}

\description{Return basic variant info as a data.frame.}

\usage{
\S4method{variantInfo}{SeqVarGDSClass}(gdsobj, alleles=TRUE, expanded=FALSE)
\S4method{expandedVariantIndex}{SeqVarGDSClass}(gdsobj)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{alleles}{A logical value for whether to include ref and alt alleles}
  \item{expanded}{A logical value for whether to expand multi-allelic variants with one row per alternate allele.}
}

\details{
Variants can be represented in collapsed form, with one row per variant, or in expanded form, with one row per alternate allele for multiallelic variants.
}

\value{
\code{variantInfo} returns a data.frame with variant.id, chromosome, and position for each variant. If \code{alleles=TRUE}, the data.frame includes ref and alt. If \code{expanded=TRUE}, the data.frame includes allele.index, which is 1 for the first alternate allele, 2 for the second alternate, etc.

\code{expandedVariantIndex} returns an index to transform a vector or matrix from collapsed to expanded form.
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}}
}

\examples{
gds <- seqOpen(seqExampleFileName("gds"))
seqSetFilter(gds, variant.sel=1323:1327)
variantInfo(gds, alleles=TRUE)
variantInfo(gds, alleles=TRUE, expanded=TRUE)
expandedVariantIndex(gds)
seqClose(gds)
}

\name{getVariableLengthData}
\alias{getVariableLengthData}
\alias{getVariableLengthData,SeqVarGDSClass,character-method}

\title{Get variable-length data}

\description{
  Get data with multiple values per sample from a GDS object and return
  as an array
}

\usage{
\S4method{getVariableLengthData}{SeqVarGDSClass,character}(gdsobj, var.name, use.names=TRUE, parallel=FALSE)
}

\arguments{
  \item{gdsobj}{A \code{\link{SeqVarGDSClass}} object with VCF data.}
  \item{var.name}{Character string with name of the variable, most
    likely "annotation/format/VARIABLE_NAME".}
  \item{use.names}{A logical indicating whether to assign sample and
    variant IDs as dimnames of the resulting matrix.}
  \item{parallel}{Logical, numeric, or other value to control parallel
    processing; see \code{\link{seqParallel}} for details.}
}

\details{
  Data which are indicated as having variable length (possibly different
  numbers of values for each variant) in the VCF header are stored as
  variable-length data in the GDS file.  Each such data object has two
  components, "length" and "data."  "length" indicates how many values
  there are for each variant, while "data" is a matrix with one row per
  sample and columns defined as all values for variant 1, followed by
  all values for variant 2, etc.

  \code{getVariableLengthData} converts this format to a 3-dimensional
  array, where the length of the first dimension is the maximum number
  of values in "length," and the remaining dimensions are sample and
  variant.  Missing values are given as NA.  If the first dimension of
  this array would have length 1, the result is converted to a matrix.
}

\value{
  An array with dimensions [n, sample, variant] where n is the maximum
  number of values possible for a given sample/variant cell.  If n=1, a
  matrix with dimensions [sample,variant].
}

\author{Stephanie Gogarten}

\seealso{
  \code{\link{SeqVarGDSClass}},
  \code{\link{applyMethod}},
  \code{\link{seqGetData}}
}

\examples{
file <- system.file("extdata", "gl_chr1.gds", package="SeqVarTools")
gds <- seqOpen(file)
## genotype likelihood 
gl <- seqGetData(gds, "annotation/format/GL")
names(gl)
gl$length
## 3 values per variant - likelihood of RR,RA,AA genotypes
dim(gl$data)
## 85 samples (rows) and 9 variants with 3 values each - 27 columns

gl.array <- getVariableLengthData(gds, "annotation/format/GL")
dim(gl.array)
## 3 genotypes x 85 samples x 9 variants
head(gl.array[1,,])
head(gl.array[2,,])
head(gl.array[3,,])

seqClose(gds)
}

\keyword{manip}

\name{SeqVarTools-package}
\alias{SeqVarTools-package}
\alias{SeqVarTools}
\docType{package}
\title{Tools for Variant Analysis}

\description{
  This package provides tools for data exploration and analysis of
  variants, extending the functionality of the package
  \pkg{\link{SeqArray}}.
}

\details{
  \pkg{\link{SeqArray}} provides an alternative to the Variant Call
  Format (VCF) for storage of variants called from sequencing data,
  enabling efficient storage, fast access to subsets of the data, and
  rapid computation.

  SeqVarTools provides an interface to the \pkg{\link{SeqArray}} storage
  format with tools for many common tasks in variant analysis and
  integration with basic S4 classes in Bioconductor.
}

\author{
  Stephanie M. Gogarten, Xiuwen Zheng
  
  Maintainer: Stephanie M. Gogarten \email{sdmorris@u.washington.edu}
}

\keyword{package}
